#include "Atoms.h"
#include "stdlib.h"

Atoms::Atoms() : rbt("")
{
  tablesize = MINTABLESIZE;
  a = new SimpleString[tablesize];
  if (a == 0)
    {
      cerr << "Atoms : Atoms : no more memory \n";
      exit(-1);
    }
  nextfree = 1;
  a[0] = "";
}

Atoms::~Atoms()
{
  delete [] a;
}

Atoms::Atoms(Atoms& atoms) : rbt("")
{
  tablesize = atoms.tablesize;
  a = new SimpleString[tablesize];
  if (a == 0)
    {
      cerr << "Atoms : Atoms : no more memory \n";
      exit(-1);
    }
  nextfree = atoms.nextfree;
  for (int i=0;i < tablesize;i++)
    {
      a[i] = atoms.a[i];  
    }
}

Atoms& Atoms::operator=(Atoms& atoms)
{
  delete [] a;
  tablesize = atoms.tablesize;
  a = new SimpleString[tablesize];
  if (a == 0)
    {
      cerr << "Atoms : Atoms : no more memory \n";
      exit(-1);
    }
  nextfree = atoms.nextfree;
  for (int i=0;i < tablesize;i++)
    {
      a[i] = atoms.a[i];  
    }
  return *this;
}

void Atoms::clear()
{
  nextfree = 1;
  a[0] = "";
  rbt.clear();
}

int Atoms::inserted(SimpleString& ss)
{
  if (ss == "")
    return 0;
  int* key = rbt.lookup(ss);
  if (key != 0) // it is already there
    return 1;
  else
    return 0;
}

int Atoms::inserted(const char* s)
{
  if (*s == 0)
    return 0;
  SimpleString ss=s;
  int* key = rbt.lookup(ss);
  if (key != 0) // its already there
    return 1;
  else
    return 0;
}

int Atoms::operator()(SimpleString& ss)
{
  if (ss == "")
    return 0;
  int* key = rbt.lookup(ss);
  if (key != 0) // it is already there, fine
    return *key;
  // we have to insert it
  rbt.insert(ss,nextfree);
  ensure_tablelen(nextfree);
  a[nextfree] = ss;
  return(nextfree++);
}

int Atoms::operator()(const char* s)
{
  if (*s == 0)
    return 0;
  SimpleString ss=s;
  return operator()(ss);
}

SimpleString Atoms::operator()(int i)
{
  if ((i <= 0) || (i >= nextfree))
    return "";
  else
    return a[i];
}

int Atoms::size() 
{ 
  return nextfree-1; 
}

void Atoms::ensure_tablelen(int len)
{
  if (len < tablesize)
    return;
  // remalloc
  tablesize *= 2;
  SimpleString* h = new SimpleString[tablesize];
  if (h == 0)
    {
      cerr << "Atoms : Atoms : no more memory \n";
      exit(-1);
    }
  for (int i=0;i<nextfree;i++)
    h[i] = a[i];
  delete [] a;
  a = h;
}


