#ifndef __Coefficient_h
#define __Coefficient_h

// You might want to use a different datatype for the representation 
// of the coefficients and the computation with them.
// In that case please select the datatype of your choice 
// by selecting one of the following datatypes.or defining the
// respective Name in the Makefile (preferred).
// #define COEFFICIENT_int
// #define COEFFICIENT_ledaint
// #define COEFFICIENT_gnuint
// #define COEFFICIENT__int64
// COEFFICIENT_int is the preferred datatype providing best performance
// COEFFICIENT_ledaint requires LEDA (works with 3.5.2 -> freely available)
// COEFFICIENT_gnuint requires the GNU C++ Library
// COEFFICIENT_int64 for using the native 64-bit datatype on Alpha/Windows NT
//                   provided by Nick

// If you want to implement your own datatype make sure that
//   * the two functions coeffcient_to_int and coefficient_to_double 
//     doing what their name stands for are properly implemented
//   * the datatype behaves otherwise very much like int
//   * that you link the necessary libraries

// Standard int datatype for coefficients -> fastest
#ifdef COEFFICIENT_int
#define Coefficient int
inline int    coefficient_to_int   (Coefficient coefficient) { return coefficient; }
inline double coefficient_to_double(Coefficient coefficient) { return (double) coefficient; }
#endif

#ifdef COEFFICIENT_gnuint
// The Integer datatype which is available in libg++
#include "Integer.h"
#define Coefficient Integer
inline int    coefficient_to_int   (Coefficient coefficient) 
   { return (int) coefficient.as_long(); }
inline double coefficient_to_double(Coefficient coefficient) 
   { return coefficient.as_double(); }
#endif

#ifdef COEFFICIENT_ledaint
// LEDA 3.5.2 (last version which is freely available with source code)
#define __NO_LOCAL_TYPEDEF__
#include "LEDA/basic.h"
#include "LEDA/integer.h"
#define Coefficient integer
inline int    coefficient_to_int   (Coefficient coefficient) 
   { return (int) coefficient.to_long(); }
inline double coefficient_to_double(Coefficient coefficient) 
   { return coefficient.to_double(); }
#endif

#ifdef COEFFICIENT__int64
// works with native 64bit datatype int on Alpha bases Windows NT sytems
// provided by Nick 
#define Coefficient __int64
inline int    coefficient_to_int   (Coefficient coefficient) 
   { return (int) coefficient; }
inline double coefficient_to_double(Coefficient coefficient) 
   { return coefficient.to_double(); }
#ifdef WIN32
extern "C"{
#include <stdlib.h>
#define abs(i) ((i) >= 0? (i): (-(i)))
}
inline ostream& operator<<(ostream& out,Coefficient i) 
{ 
        char buffer[20];
        _i64toa(i, buffer, 10);
        out << buffer;
        return out;
}
#endif 
#endif


#endif
