#include "Enumerate.h"
#include "timing.h"

Enumerate::Enumerate()
{
  // ToDo
  // initialize random seed
  srand(gmt_secs());
  select_to_del_lit_option = 0;
}

Enumerate::Enumerate(const Enumerate& )
{
  // ToDo
}

Enumerate& Enumerate::operator=(const Enumerate& )
{
  // ToDo
  return *this;
}

Enumerate::~Enumerate()
{
  // ToDo
}

int Enumerate::init_enumeration()
{
  last_solution.clear();
  if (lpbiref.max() == -1)   // handle empty constraint sets!
    {
      return SAT; // then it is satisfied
    }
  int greatest_variable = litref.greatest_variable();
  // the sizes of the stacks
  FixedStack.init_size(0,greatest_variable);
  FixedStackIndex = -1;
  FixedStackChoicepoints.init_size(0,greatest_variable);
  FixedStackChoicepointsIndex = -1;
  
  ToFixStack.init_size(0,greatest_variable);
  ToFixStackIndex = -1;
  LitValue.init_size(-greatest_variable,greatest_variable);
  LitHeuValue.init_size(-greatest_variable,greatest_variable);
  int i;
  for(i=0; i <= greatest_variable; i++)
    {
     if (!litref.var_occurs(i))
	{
	  // literal does not occur
	  LitValue[i] = LitValue[-i] = 2;
	  // should not be needed
	}
      else
	LitValue[i] = LitValue[-i] = 0; 
    }
  
  EnumTableStack.init_size(-1,greatest_variable); // -1 for unchanging copy of initial state
  EnumTableStackIndex = 0;
  int greatest_iid = lpbiref.max();
  if (greatest_variable <= 10) // a stack of 100 fits always!
    largestenlargedETSindex = greatest_variable;
  else
    {
      largestenlargedETSindex = 10;
    }
  for(i=-1; i <= largestenlargedETSindex; i++)
    {
      EnumTableStack[i].enlarge(greatest_iid); // most of the memory!!
    }
  RArrayList<LPBIState>& state = EnumTableStack[0];
  lpbiref.first();
  while(!lpbiref.last())
    {
      int iid = lpbiref.next();
      state[iid].sumci = lpbiref.get_lpbi(iid).get_sumcoeffs();
      state[iid].rhs = lpbiref.get_lpbi(iid).get_rhs();
      state[iid].maxci = lpbiref.max_product(iid).coefficient;
      state[iid].maxliteral = lpbiref.max_product(iid).literal;
      // if initial inequality is unsatisfiable then abort
      if (state[iid].sumci < state[iid].rhs)
	return UNSAT;
    }
  state.make_list(lpbiref.min(),lpbiref.max());
  // unlink not existant iids
  int lpbiref_max = lpbiref.max();
  for(i=lpbiref.min(); i <= lpbiref_max; i++)
    {      
      if (lpbiref.empty(i))
	state.unlink(i);
    }
  // copy table in unchanging location -1
  EnumTableStack[-1] = EnumTableStack[0];
  return OPEN;
}

int Enumerate::reset_enumeration()
{
  // assumes that init_enumeration has been called and that
  // just a running enumeration must be reset without changing
  // the datastructure
  int greatest_variable = litref.greatest_variable();
  FixedStackIndex = -1;
  FixedStackChoicepointsIndex = -1;
  ToFixStackIndex = -1;

  for(int i=0; i <= greatest_variable; i++)
    {
     if (!litref.var_occurs(i))
	{
	  // literal does not occur
	  LitValue[i] = LitValue[-i] = 2;
	  // should not be needed
	}
      else
	LitValue[i] = LitValue[-i] = 0; 
    }
  EnumTableStackIndex = 0;
  EnumTableStack[0] = EnumTableStack[-1]; // copy from unchanging initial configuration
  return OPEN;
}

int Enumerate::add(int iid,LPBIneq& lpbi)
{
  lpbiref.add(iid,lpbi);
  litref.add(iid,lpbi);
  return 0;
}

int Enumerate::del(int iid)
{
  litref.del(iid,lpbiref[iid]);
  lpbiref.del(iid);
  return 0;
}

int Enumerate::replace(int iid,LPBIneq& lpbi)
{
  litref.del(iid,lpbiref[iid]);
  lpbiref.get_lpbi(iid) = lpbi; // copy lpbi on old place
  litref.add(iid,lpbi);
  return 0;
}

void Enumerate::clear()
{
  lpbiref.clear();
  litref.clear();
}

int Enumerate::solve()
{
  int res = OPEN;   
  RArray<int> select_stack(0,LitValue.upper());
  int ss_index = 0;
  
  while ((res == OPEN) && (ss_index >= 0))
    {
      EnumTablesave();
      select_stack[ss_index++] = -select_literal();
      nodes++;
      res = fix_lit(-select_stack[ss_index-1]);
      if (res == SAT)
	{
	  update_solution();
	  return SAT;
	}
      else if (res == UNSAT)
	{
	  while (res == UNSAT)
	    {
	      EnumTablerestore();
	      nodes++;
	      res = fix_lit(select_stack[--ss_index]);
	    }	  
	}
      // else if (res == OPEN) ; continue with selection
    }
  return res;
}

int Enumerate::solve_limit(int limit)
{
//  cout << "solve_limit\n";cout.flush();
  int res = OPEN;   
  limit += nodes;
  RArray<int> select_stack(0,LitValue.upper());
  int ss_index = 0;
  int saved = 0;

//  cout << "ETSIbefore : " << EnumTableStackIndex << "\n";
  while ((res == OPEN) && (nodes < limit) && (ss_index >= 0))
    {
      EnumTablesave(); saved++;
      select_stack[ss_index++] = -select_literal();
      nodes++;
      res = fix_lit(-select_stack[ss_index-1]);
      if (res == SAT)
	{
	  update_solution();
	}
      else if (res == UNSAT)
	{
	  while ((res == UNSAT) && (ss_index > 0))
	    {
	      EnumTablerestore(); saved--;
	      nodes++;
	      res = fix_lit(select_stack[--ss_index]);
	    }  
	}
      // else if (res == OPEN) ; continue with selection
    }
//  cout << "ETSIafter : " << EnumTableStackIndex << "\n";
//  cout << "restoring Table " << saved << " times \n";cout.flush();
  while(saved--)
    EnumTablerestore();
//  cout << "ETSIrestored : " << EnumTableStackIndex << "\n";cout.flush();
//  cout << "solve_limit done\n";cout.flush();
  return res;
}

void Enumerate::EnumTablesave()
{
  int i;
  // work on copy of EnumTable
//  EnumTableStack[EnumTableStackIndex+1] = EnumTableStack[EnumTableStackIndex];
  // for copying we step through the initial table and copy all active ones
  // and copy all inactive information
  RArrayList<LPBIState>& orig_table = EnumTableStack[-1];
  RArrayList<LPBIState>& to_copy_table = EnumTableStack[EnumTableStackIndex];
  // ensure that the used EnumTablestackindex is enlarged
  if (largestenlargedETSindex < EnumTableStackIndex + 1)
    {
      // enlarge it 
// cout << "ENLARGE SINCE NEEDED: Depth " << EnumTableStackIndex + 1 << "\n";
      int greatest_iid = lpbiref.max();
      EnumTableStack[EnumTableStackIndex+1].enlarge(greatest_iid);
      // allocate memory when it is needed
      largestenlargedETSindex++;
    }
  RArrayList<LPBIState>& copy_table = EnumTableStack[EnumTableStackIndex+1];
  for(i = orig_table.first();orig_table.active_index(i);i = orig_table.next(i))
    {
      if (to_copy_table.active_elem(i))
	{
	  copy_table.copy_node(i,to_copy_table); // copy complete information
	}
      else
	{
	  copy_table.mark_inactive(i); // just assure that elem is marked inactive
	}
    }
  copy_table.copy_first_index(to_copy_table);
  // now table should be correctly copied
#ifdef CHECK
  for(i = to_copy_table.first();to_copy_table.active_index(i);i = to_copy_table.next(i))
    {
      if (!copy_table.active_elem(i))
	{
	  cerr << "LPBIneqSet : EnumTablesave : Fatal Error EnumTableCopy corrupted (1)\n";
	  exit(-1);
	}
    }
  for(i = orig_table.first();orig_table.active_index(i);i = orig_table.next(i))
    {
      if (to_copy_table.active_elem(i) != copy_table.active_elem(i))
	{
	  cerr << "LPBIneqSet : EnumTablesave : Fatal Error EnumTableCopy corrupted (2)\n";
	  exit(-1);
	}
    }
#endif
  EnumTableStackIndex++;
  FixedStackChoicepoints[++FixedStackChoicepointsIndex] = FixedStackIndex;
}

void Enumerate::EnumTablerestore()
{
  // restore last state including state of literals
  EnumTableStackIndex--;
  // undo variable bindings
  int undountil = FixedStackChoicepoints[FixedStackChoicepointsIndex--];
  int undolit;
  while (FixedStackIndex > undountil)
    {
      undolit = FixedStack[FixedStackIndex--];
      LitValue[undolit] = 0;
      LitValue[-undolit] = 0;
    }
}

int Enumerate::GenerateCover(OrdInt& cover)
{
  // enumeration must be initialized
  int res,fixlit;
  int i=0;

  cover.clear();
  EnumTablesave();
  do 
    {
      fixlit = select_literal();
      if (i <= last_cover.upper())
	last_cover.ensure(2*(i+1));
      last_cover[i++] = fixlit;
      cover += fixlit;
      res = fix_lit(fixlit);
    } while (res == OPEN);
  if (res == SAT)
    {
      update_solution();
    }
  EnumTablerestore();
  return res;
}

int Enumerate::select_to_del_lit(OrdInt& to_del)
{
  int i;
  switch (select_to_del_lit_option) {
  case 0 :
    return to_del.pop();
  case 1 :
    // earlier chosen, earlier eliminated 
    // (not very fast but not critical
    for (i=0; i <= last_cover.upper(); i++)
      {
	if (to_del.member(last_cover[i]))
	  {
	    to_del -= last_cover[i];
	    return last_cover[i];
	  }
      }
    break;
  case 2 :
    // earlier chosen, later eliminated 
    // (not very fast but not critical
    for (i=last_cover.upper(); i >= 0; i--)
      {
	if (to_del.member(last_cover[i]))
	  {
	    to_del -= last_cover[i];
	    return last_cover[i];
	  }
      }
    break;
  case 4 :
    return to_del.pop_random(); // put some random feature in
  default :
    return to_del.pop();
  }
  cerr << "Enumerate::select_to_del_lit : no literal found ?!?!\n";
  return 0;
}

int Enumerate::MinimizeCover(OrdInt& cover, int max_cut_nodes, int lb)
{
  // enumeration must be initialized

  // first, the set of fixed literals is reduced by
  // just eliminating a literal adn checking whether
  // fixing the set yields still unsatisfiability
  OrdInt to_del = cover;
  while(!to_del.empty())
    {
      int del_lit = select_to_del_lit(to_del);
      cover -= del_lit;
      EnumTablesave();
      int res = fix_lits(cover);
      if (res == SAT)
	{
	  update_solution();
	}
      EnumTablerestore();
      if (res == OPEN) 
	cover += del_lit;
      else if (res == SAT)
	{
	  cover += del_lit; // is ok when calculating lower bound!
	  if (!lb)
	    return SAT;
	}
      // else (res == UNSAT); -> literal is not needed for unsatisfiability
    }
  // next, we reduce the set of literals by eliminating the 
  // literal and checking whether fixing the set *plus the negation
  // of the fixed literal* yields unsatisfiability (soundness diss)
  // Of course, we could have done this from the very beginning, but
  // unfortunately then, we typically can eliminate less literals.
  // It is a mess with these NP-complete problems....
  to_del = cover;
  while(!to_del.empty())
    {
      int del_lit = select_to_del_lit(to_del);
      cover -= del_lit;
      EnumTablesave();
      cover += -del_lit; // both pos and neg fixed yield bot, then erasable
      int res = fix_lits(cover); 
      if (res == SAT)
	{
	  update_solution();
	}
      cover -= -del_lit;
      EnumTablerestore();
      if (res == OPEN) 
	cover += del_lit;
      else if (res == SAT)
	{
	  cover += del_lit; // is ok when calculating lower bound!
	  if (!lb)
	    return SAT;
	}
      // else (res == UNSAT); -> literal is not needed for unsatisfiability
    }
  if (max_cut_nodes == 0)
    return UNSAT;
  // tries to eliminate each literal while spending at most max_cut_nodes
  to_del = cover;
  while(!to_del.empty())
    {
      int del_lit = select_to_del_lit(to_del);
      cover -= del_lit;
      EnumTablesave();
      cover += -del_lit; // both pos and neg fixed yield bot, then erasable
      int res = fix_lits(cover); // cannot yield unsat!
      res = solve_limit(max_cut_nodes);
      cover -= -del_lit;
      EnumTablerestore();
      if (res == OPEN) 
	cover += del_lit;
      else if (res == SAT)
	{
	  // solution is updated by solve_limit
	  cover += del_lit; // is ok when calculating lower bound!
	  if (!lb)
	    return SAT;
	}
      // else (res == UNSAT); -> literal is not needed for unsatisfiability
    }  
  return UNSAT;
}

Coefficient Enumerate::LowerBound(Products term)
{
  // enumeration must be initialized
  // some literals can be fixed
  int res,fixlit;
  OrdInt cover;
  Coefficient lb_add;
  Coefficient lower_bound = term.constant;
  term.constant = 0;

  // first check whether there are fixed literals
  // add pos fixed to lower_bound and eliminate all fixed lits
  cover.clear();
  term.first();
  while(!term.last())
    {
      Product& p = term.next();
      if (LitValue[p.literal] < 0)
	cover += p.literal;
      else if (LitValue[p.literal] > 0)
	{
	  cover += p.literal;
	  lower_bound += p.coefficient;
	}
    }
  cover.first();
  while(!cover.last())
    term.del(cover.next());
  cover.clear();
  // now search for covers
  while (!term.empty())
    {
      EnumTablesave();
      cover.clear();
      res = OPEN;
      int i=0;
      term.first();
      while(!term.last())
	{
	  Product& p=term.next();
	  fixlit = -p.literal;
	  if (LitValue[fixlit] > 0) // literal already fixed by pbur
	    {
	      // just ignore
	    }
	  else if (LitValue[fixlit] < 0) // ~literal already fixed by pbur
	    {
	      cover += fixlit;
	      if (i <= last_cover.upper())
		last_cover.ensure(2*(i+1));
	      last_cover[i++] = fixlit;
	      res = UNSAT;
	      break;
	    }
	  else // (LitValue[fixlit] == 0) 
	    {
	      cover += fixlit;
	      if (i <= last_cover.upper())
		last_cover.ensure(2*(i+1));
	      last_cover[i++] = fixlit;
	      res = fix_lit(fixlit);
	      if ((res == SAT) || (res == UNSAT))
		break;
	    }
	}
      EnumTablerestore();
      if ((res == SAT) || (res == OPEN)) // best bound reached
	break;
      // now res == UNSAT
      if (cover.empty())
	{
	  cerr << "Enumerate : LowerBound : empty cover generated \n";
	  exit(-1);
	}
      MinimizeCover(cover);
      // get minimal coefficient in cover (lb_add)
      // delete products of literals in cover
      if (cover.empty()) // Hey, fixing is possible!
	{
	  // but we don't fix; (not here in an internal routine)
	  // we abort coeffreduction for this ineq and let
	  // fixingcheck do the job
	  term.clear();
	}
      else
	{
	  lb_add = 0;
	  cover.first();
	  while(!cover.last())
	    {
	      int nclit = -cover.next();
	      if (lb_add == 0)
		lb_add = term.member(nclit);
	      else
		{
		  Coefficient m = term.member(nclit);
		  lb_add = lb_add < m ? lb_add : m;
		}
	      term.del(nclit);
	    }
	  lower_bound += lb_add;
	}
//      cout << " + " << lb_add << " = " << lower_bound;
//      cout.flush();
    }
  return lower_bound;
}

OrdInt Enumerate::current_fixed()
{
  OrdInt fixed_lits;
  int undountil = FixedStackChoicepoints[FixedStackChoicepointsIndex];
  int FSI = FixedStackIndex;
  while (FSI > undountil)
    fixed_lits += FixedStack[FSI--];
  return fixed_lits;
}

OrdInt Enumerate::all_fixed()
{
  OrdInt fixed_lits;
  int FSI = FixedStackIndex;
  while (FSI > -1)
    fixed_lits += FixedStack[FSI--];
  return fixed_lits;
}

////////////////////////////////////////////////////////////////////////////////
// the core of the enumeration
void Enumerate::maxliteral_update(LPBIState& state, int iid)
{
  // must! find an unbound literal
#ifdef CHECK
  ProductNode* pptr = lpbiref[iid].head->next;
  while ((LitValue[pptr->product.literal] == 1) ||
	 (LitValue[pptr->product.literal] == -1))
    {
      pptr = pptr->next;
      if (pptr == 0)
	{
	  cerr << "Enumerate : enum_maxliteral_update : have not found an unbound literal for" << iid << "\n";	  
	  exit(-1);
	}
    }
#else
  ProductNode* pptr = lpbiref[iid].head->next;
  while ((LitValue[pptr->product.literal] == 1) ||
	 (LitValue[pptr->product.literal] == -1))
    pptr = pptr->next;
#endif
  state.maxci = pptr->product.coefficient;
  state.maxliteral = pptr->product.literal;
}

int Enumerate::push_literals_tofix(Coefficient sumci, Coefficient rhs, int iid)
{
  ProductNode* pptr = lpbiref[iid].head->next;
  while(pptr) // through all products
    {
      switch (LitValue[pptr->product.literal]) {
      case 2  :
	if (sumci- pptr->product.coefficient < rhs) // we need to fix it
	  {
	    // but it is already on ToFixStack; just update
	    sumci -= pptr->product.coefficient;
	    rhs -= pptr->product.coefficient;
	    if (rhs <= 0)
	      return UNLINK; // unlink ineq, since it is satisfied
	  }
	else // no further fixing
	  return OPEN;
	break;
      case -2 :
	if (sumci- pptr->product.coefficient < rhs) // we need to fix it
	  {
	    // but its negation needs to be fixed 
	    return UNSAT;
	  }
	// else no further fixing
	return OPEN;
      case 0  :
	if (sumci- pptr->product.coefficient < rhs) // we need to fix it
	  {
	    ENUM_VERB(18,"push ToFiXStack " << pptr->product.literal)
	    // mark it as fixed immediately 
	    LitValue[pptr->product.literal] = 2;
	    LitValue[-pptr->product.literal] = -2;
	    FixedStack[++FixedStackIndex] = pptr->product.literal;
	    ToFixStack[++ToFixStackIndex] = pptr->product.literal;
	    sumci -= pptr->product.coefficient;
	    rhs -= pptr->product.coefficient;
	    if (rhs <= 0)
	      return UNLINK; // unlink it ineq since it is satisfied
	  }
	else // no further fixing
	  return OPEN;
	break;
      }
      pptr = pptr->next;
    }
  return OPEN;  // shouldn't happen
}

int Enumerate::fix_lit(int lit)
{
  // push literal to fix on ToFixStack and work on Stack;
  ToFixStack[0] = lit;
  ToFixStackIndex = 0;
  // all values are already assigned 2*Value if they
  // are pushed on the FixedStack
  // must not be called with an already fixed value
  LitValue[lit] = 2;
  LitValue[-lit] = -2;
  FixedStack[++FixedStackIndex] = lit;
  return fix_stack();
}

int Enumerate::fix_lits(OrdInt& lits)
{
  int lit;
  ToFixStackIndex = -1;
  lits.first();
  while(!lits.last())
    {
      lit = lits.next();
      ToFixStack[++ToFixStackIndex] = lit;
      LitValue[lit] = 2;
      LitValue[-lit] = -2;
      FixedStack[++FixedStackIndex] = lit;
    }
  return fix_stack();
}

int Enumerate::fix_stack()
{
  int i;
  RArrayList<LPBIState>& table = EnumTableStack[EnumTableStackIndex];
  
  while(ToFixStackIndex >= 0)
    {
      fixedlits++;
      int fixlit = ToFixStack[ToFixStackIndex--];
      LitValue[fixlit] = 1;
      LitValue[-fixlit] = -1;
      // already on FixedStack
      ENUM_VERB(10,"fix " << fixlit)
      
      // positive fixing
      i = 1;
      while (i <= litref[fixlit].operator[](0).iid)
	{
	  ENUM_VERB(17,"pos check " << litref[fixlit].operator[](i).iid)
	  if (table.active_elem(litref[fixlit].operator[](i).iid))
	    {
	      // an active inequality
	      LPBIState& state = table[litref[fixlit].operator[](i).iid];
	      ENUM_VERB(15,"pos in iid " << litref[fixlit].operator[](i).iid)
	      if ((state.rhs -= litref[fixlit].operator[](i).coefficient) <= 0)
		{
		  ENUM_VERB(13,"unlink " << litref[fixlit].operator[](i).iid)
		  table.unlink(litref[fixlit].operator[](i).iid);
		}
	      else
		{
		  state.sumci -= litref[fixlit].operator[](i).coefficient;
		  if (fixlit == state.maxliteral)
		    maxliteral_update(state,litref[fixlit].operator[](i).iid);
		}
	    }
	  i++;
	}
      // negative fixing
      i = 1;
      while (i <= litref[-fixlit].operator[](0).iid)
	{
	  ENUM_VERB(17,"neg check " << litref[fixlit].operator[](i).iid)
	  if (table.active_elem(litref[-fixlit].operator[](i).iid))
	    {
	      // an active inequality
	      ENUM_VERB(15,"neg in iid " << litref[-fixlit].operator[](i).iid)
	      LPBIState& state = table[litref[-fixlit].operator[](i).iid];
	      if ((state.sumci -= litref[-fixlit].operator[](i).coefficient) < state.rhs)
		{
		  ENUM_VERB(10," fathomed")
		  return UNSAT; // unsatisfiable
		}
	      else 
		{
		  if (state.sumci-state.maxci < state.rhs)
		    { // further fixing
		      int res = push_literals_tofix(state.sumci,state.rhs,litref[-fixlit].operator[](i).iid);
		      if (res == UNLINK) // will satisfy
			{
			  ENUM_VERB(13,"unlink " << litref[-fixlit].operator[](i).iid)
			  table.unlink(litref[-fixlit].operator[](i).iid);
			}
		      else if (res == UNSAT)
			{
			  ENUM_VERB(10," fathomed")
			  return UNSAT; // unsatisfiable
			}
		    }
		  if (-fixlit == state.maxliteral)
		    maxliteral_update(state,litref[-fixlit].operator[](i).iid);
		}
	    }
	  i++;
	}      
    }
  if (table.empty())
    {
      update_solution();
      return SAT;
    }
  else
    return OPEN;
}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// trace functions
int Enumerate::compute_sizes()
{
  int iids_range = lpbiref.max()-lpbiref.min()+1;
  int greatest_variable = litref.greatest_variable();
  int litstacks_bytes = 3*sizeof(int)*(greatest_variable+1);
  litstacks_bytes += sizeof(int)*2*(greatest_variable);
  litstacks_bytes += sizeof(double)*2*(greatest_variable);
  int enumtable_bytes = (greatest_variable+2)*iids_range*sizeof(LPBIState);
  
  int lpbiref_entries = 0;
  lpbiref.first();
  while(!lpbiref.last())
    lpbiref_entries += lpbiref[lpbiref.next()].length() + 1;
  int lpbiref_bytes = lpbiref_entries * sizeof(ProductNode);

  int sum = lpbiref_bytes + enumtable_bytes;

  cout << "Sizes : \n";
  cout << "iids_rangs : " << iids_range;
  cout << " [" << lpbiref.min() << "-" << lpbiref.max() << "]\n";
  cout << "Stacks     : " << litstacks_bytes << "\n";
  cout << "Table      : " << enumtable_bytes << "\n";
  cout << "Sum        : " << sum << "\n";
  cout.flush();
  return sum;
}

int Enumerate::show_enum_table()
{
  if (verbosity < 9)
    return 0;
  RArrayList<LPBIState>& table = EnumTableStack[EnumTableStackIndex];

  cout << "EnumTable[" << EnumTableStackIndex << "]\n";
  for(int i = table.first();table.active_index(i);i = table.next(i))
    {
      LPBIState& state = table[i];
      cout << "iid : " << i;
      cout << " sumci : " << state.sumci;
      cout << " rhs : " << state.rhs;
      cout << "\n";
    }
  return 0;
}

int Enumerate::check_data_structure()
{
  // each ineq correctly in litref
  lpbiref.first();
  while(!lpbiref.last())
    {
      int iid = lpbiref.next();
      Products& ps = lpbiref[iid];
      ps.first();
      while(!ps.last())
	{
	  Product& p = ps.next();
	  RArray<LitRef>& ref = litref[p.literal];
	  int i;
	  for (i=1; i <= ref[0].iid; i++)
	    {
	      if (ref[i].iid == iid)
		{
		  if (ref[i].coefficient != p.coefficient)
		    {
		      cout << "Enum:Check: " << p.coefficient << "*";
		      plit(p.literal);
		      cout << "in " << iid << " but ";
		      cout << ref[i].coefficient << "*";
		      plit(p.literal);
		      cout << " in index \n";
		    }
		  break;
		}
	    }
	  if (i > ref[0].iid)
	    {
	      cout << "Enum:Check: " << p.coefficient << "*";
	      plit(p.literal);
	      cout << " not in index for " << iid << "\n";
	    }
	}
    }
  // each litref corresponds to ineq
  for (int i=litref.lower(); i <= litref.upper(); i++)
    {
      RArray<LitRef>& ref = litref[i];
      for (int j=1; j <= ref[0].iid; j++)
	{
	  Coefficient coefficient = lpbiref[ref[j].iid].member(i);
	  if (coefficient != 0)
	    {
	      if (ref[j].coefficient != coefficient)
		{
		  cout << "Enum:Check: index disagrees on coefficient for Literal ";
		  plit(i);
		  cout << "; it is " << coefficient << " index says ";
		  cout << ref[j].coefficient << "\n";
		}
	      break;
	    }
	  else 
	    {
	      cout << "Enum:Check: Literal ";
	      plit(i);
	      cout << "not in " << ref[j].iid;
	      cout << "as claimed by index with coefficient " << ref[j].coefficient << "\n";
	    }
	}
    }
  return 0;
}

int Enumerate::show_data_structure(ostream& out)
{
  litref.Print(out,atoms);
  lpbiref.Print(out,atoms);
  return 0;
}

void Enumerate::update_solution()
{ 
  OrdInt oi;
  oi = all_fixed();
  last_solution.clear(); 
  last_solution.from_OrdInt(oi);
}

int coefficient_show_fixed(Coefficient i, int length)
{
  Coefficient j=abs(i);
  int len = 0;
  while ((j = j / 10) > 0)
    len++;
  if (i < 0)
    len++;
  while(length-len > 0)
    {
      length--;
      cout << " ";
    }
  cout << i;
  return 0;
}

/////////////////////////////////////////////////////////////////////////////


