#include "Enumerate.h"
extern "C" {
#include "stdlib.h"
}

// the simplest heuristic, namely none
int Enumerate::heuristic0()
{
  int lit = 1;
  while (LitValue[lit] != 0)
    lit++;
  return lit;
}

// first a tabled version of 2^-i
#include "two_pow_table.C"

inline double two_exp_minus(Coefficient i)
{
  if (i <= 1000)
    return two_powtable[coefficient_to_int(i)];
  else 
    return 0.0;
}

// the values for the heuristics1
inline double value_heuristic1(Coefficient coefficient, Coefficient sumci, Coefficient rhs)
{
  if (coefficient < (sumci - rhs))
    return (coefficient_to_double(coefficient)/
            coefficient_to_double(sumci - rhs)) * 
           two_exp_minus((sumci/rhs) +1);
  else
    return two_exp_minus(sumci/rhs + 1);   
}

// the values for the heuristics2
// the (simple) idea is as follows:
// ci/sumci represents the percentage of the coefficient ci of available 
//    coefficients. If it is high, it is like that it must be set to true 
//    (indiciation of literal)
// rhs/sumci represents the percentage of the needed coefficients of sumci 
//    in order to reach rhs. If it is high, it is like that literals in that 
//    ineq must be fixed.
// Both are indicators for selecting a literal in that ineq; resp. selecting 
// that literal.
// just multiply them and you get
//    ci*rhs/sumci^2
// which is what I compute in that heuristic.
inline double value_heuristic2(Coefficient coefficient, Coefficient sumci, Coefficient rhs)
{
  if (coefficient < (sumci - rhs))
    {
      return coefficient_to_double(coefficient)*coefficient_to_double(rhs)/
            (coefficient_to_double(sumci)*coefficient_to_double(sumci));
    }
  else
    {
      return coefficient_to_double(sumci - rhs)*coefficient_to_double(rhs)/
             (coefficient_to_double(sumci)*coefficient_to_double(sumci));
    }
}

// generic procedure for calculating the weights
// given the coefficent of the literal, the current sum of
// the coefficients, and the current rhs.
int Enumerate::calc_lit_values(double(* val)(Coefficient,Coefficient,Coefficient))
{
  int i;
  for(i=1;i <= LitHeuValue.upper();i++)
    if (LitValue[i] == 0)
      LitHeuValue[-i] = LitHeuValue[i] = 0.0;
  RArrayList<LPBIState>& table = EnumTableStack[EnumTableStackIndex];
  // for all active LPBIneqs
  for(i = table.first();table.active_index(i);i = table.next(i))
    {
      LPBIState& state = table[i];
      ProductNode* pptr = lpbiref[i].head;
      while ((pptr = pptr->next)) // for all products
	if (LitValue[pptr->product.literal] == 0) // active literal
	  LitHeuValue[pptr->product.literal] += 
	    (* val)(pptr->product.coefficient,state.sumci,state.rhs);
    }
  return 0;
}

// inlined computation of calc_lit_values(value_heuristic1)
int Enumerate::calc_lit_values1()
{
  int i;
  for(i=1;i <= LitHeuValue.upper();i++)
    if (LitValue[i] == 0)
      LitHeuValue[-i] = LitHeuValue[i] = 0.0;
  RArrayList<LPBIState>& table = EnumTableStack[EnumTableStackIndex];
  // for all active LPBIneqs
  for(i = table.first();table.active_index(i);i = table.next(i))
    {
      LPBIState& state = table[i];
// Why do C++ compilers not automatically extract nonchanging subexpressions in a loop%$%^!
      double expo = two_exp_minus((state.sumci/state.rhs) + 1);
      Coefficient sc_rhs_diff = state.sumci - state.rhs;
      double expodiv = expo/coefficient_to_double(sc_rhs_diff);
      ProductNode* pptr = lpbiref[i].head;
      while ((pptr = pptr->next)) // for all products
	if (LitValue[pptr->product.literal] == 0) // active literal
	  LitHeuValue[pptr->product.literal] += 
	    (pptr->product.coefficient < (sc_rhs_diff)) ?
	    coefficient_to_double(pptr->product.coefficient)*expodiv : expo;   
    }
  return 0;
}

// inlined computation of calc_lit_values(value_heuristic1)
int Enumerate::calc_lit_values2()
{
  int i;
  for(i=1;i <= LitHeuValue.upper();i++)
    if (LitValue[i] == 0)
      LitHeuValue[-i] = LitHeuValue[i] = 0.0;
  RArrayList<LPBIState>& table = EnumTableStack[EnumTableStackIndex];
  // for all active LPBIneqs
  for(i = table.first();table.active_index(i);i = table.next(i))
    {
      LPBIState& state = table[i];
      Coefficient sc_rhs_diff = state.sumci - state.rhs;
      double expomult = coefficient_to_double(state.rhs)/
                        (coefficient_to_double(state.sumci)*
                        coefficient_to_double(state.sumci));
      double diffexpomult = coefficient_to_double(sc_rhs_diff)*expomult;
      ProductNode* pptr = lpbiref[i].head;
      while ((pptr = pptr->next)) // for all products
	if (LitValue[pptr->product.literal] == 0) // active literal
	  LitHeuValue[pptr->product.literal] += 
	    pptr->product.coefficient < (sc_rhs_diff) ? 
	    coefficient_to_double(pptr->product.coefficient)*expomult : diffexpomult;
    }
  return 0;
}

// generic literal selection routine parametrized by a
// weighting function
int Enumerate::val_heuristic_sum(double (* val)(Coefficient,Coefficient,Coefficient))
{
  calc_lit_values(val);
  return select_maximal_literal_sum();
}

int Enumerate::val_heuristic_min(double (* val)(Coefficient,Coefficient,Coefficient))
{
  calc_lit_values(val);
  return select_maximal_literal_min();
}

// selects a literal Li that maximizes H[Li] + H[-Li] and H[Li] >= H[-Li] 
// for the computed heuristic values H.
int Enumerate::select_maximal_literal_sum()
{
  // Assume that we have J(L_i) for all unfixed literals
  // now search the best Literal
  double maxval = -0.1;
  int maxvar=0;
  int i;
  for(i=1;i <= LitHeuValue.upper();i++) // unbound literal
    {
      if (LitValue[i] == 0)
	{
	  if (LitHeuValue[i] + LitHeuValue[-i] > maxval)
	    {
	      maxvar = i;
	      maxval = LitHeuValue[i] + LitHeuValue[-i];
	    }
	}
    }  
  return (LitHeuValue[maxvar] >= LitHeuValue[-maxvar]) ? maxvar : -maxvar;
}

// selects a literal Li that maximizes min(H[Li] + H[-Li]) and H[Li] >= H[-Li]
// for the computed heuristic values H.
int Enumerate::select_maximal_literal_min()
{
  // Assume that we have J(L_i) for all unfixed literals
  // now search the best Literal
  double maxval = -0.1;
  int maxvar=0;
  int i;
  for(i=1;i <= LitHeuValue.upper();i++)
    {
      if (LitValue[i] == 0) // unbound literal
	{
	  if ((LitHeuValue[i] > maxval) && (LitHeuValue[-i] > maxval))
	    {
	      maxvar = i;
	      maxval = LitHeuValue[i] > LitHeuValue[-i] ?  LitHeuValue[-i] : LitHeuValue[i];
	    }
	}
    }  
  return (LitHeuValue[maxvar] >= LitHeuValue[-maxvar]) ? maxvar : -maxvar;
}

// gives a point to a literal if fixing the negation of the literal forces
// at least one literal in an inequality to be fixed
// either counts how many literals need to be fixed, or just 1 if at least
// one literal needs to be fixed.
inline int Enumerate::count_fixed_literals(ProductNode* pptr, ProductNode* fixed_pptr, 
                                           Coefficient sumci, Coefficient rhs)
{
  Coefficient fixsumci = sumci-fixed_pptr->product.coefficient;
  int count = 0;

  while (pptr) // for all products
    {     
      if (LitValue[pptr->product.literal] == 0) // active literal
        {
          if (pptr != fixed_pptr) // the literal for which we test whether fixing implies fixings
            {
              if (fixsumci - pptr->product.coefficient < rhs)
                {
//                return 1; // if one fixing is sufficient as indicator
                  count++;
                }
              else
                {
                  return count; // first time nothing to fix, no further fixings possible
                }
            }
        }
      pptr = pptr->next;
    }
  return count;
}

int Enumerate::calc_lit_values3(void )
{
  int i;
  int found=0;
  int count;

  for(i=1;i <= LitHeuValue.upper();i++)
    {
      if (LitValue[i] == 0)
	{
	  LitHeuValue[-i] = LitHeuValue[i] = 0.0;
	}
    }
  RArrayList<LPBIState>& table = EnumTableStack[EnumTableStackIndex];
  // for all active LPBIneqs
  for(i = table.first();table.active_index(i);i = table.next(i))
    {
      LPBIState& state = table[i];
      ProductNode* pptr = lpbiref[i].head->next;
      while (pptr) // for all products
        {         
          if (LitValue[pptr->product.literal] == 0) // active literal
            {
              if ((count = count_fixed_literals(lpbiref[i].head->next, pptr, state.sumci, state.rhs)) > 0)
                {
                  LitHeuValue[pptr->product.literal] += double(count);
                  found = 1;
                }
              else
                break;
            }
          pptr = pptr->next;
        }
    }
  if (found) // something will fix -> take it
    {
      return 0;
    }
  else
    {
      calc_lit_values(value_heuristic1);
      return 0;
    }
}


// only for optimization useful
// literal with the greatest coefficient in the objective function
// if none available returns 0, then other heuristic must be chosen
int Enumerate::heuristic_max_obj()
{
  if (objective_function_iid == -1) // not optimizing
    return 0;
  int iid = objective_function_iid;
  ProductNode* pptr = lpbiref[iid].head->next;
  while((pptr) && (LitValue[pptr->product.literal] != 0))
    pptr = pptr->next;
  if (pptr == 0) // in objf all lits fixed
    {
      return 0;
    }
  else
    return pptr->product.literal;
}

// select a random literal
int Enumerate::heuristic4()
{
  int i;
  int no_lits = 0;
  for(i=1;i <= LitHeuValue.upper();i++)
    {
      if (LitValue[i] == 0)
	no_lits++;
    }
  int which=rand()%(no_lits*2);
  for(i=1;i <= LitHeuValue.upper();i++)
    {
      if (LitValue[i] == 0)
	{
	  if (which <= 1)
	    return which == 1 ? i : -i;
	  else
	    which -= 2;
	}
    }
  cerr << "Enumerate : heuristic4 : no literal selected\n";
  exit(-1);
  return 0; // should not happen
}


int Enumerate::select_literal()
{
  switch (heuristic) {
  case 0 : return heuristic0();
//   case 1 : return val_heuristic_sum(value_heuristic1);
  case 1 : calc_lit_values1();return select_maximal_literal_sum();
//  case 2 : return val_heuristic_sum(value_heuristic2);
  case 2 : calc_lit_values2();return select_maximal_literal_sum();
  case 3 : calc_lit_values3();return select_maximal_literal_sum();
  case 4 : return heuristic4();

  case 10 : return heuristic0(); // nothing to change
//  case 11 : return val_heuristic_min(value_heuristic1);
  case 11 : calc_lit_values1();return select_maximal_literal_min();
//  case 12 : return val_heuristic_min(value_heuristic2);
  case 12 : calc_lit_values2();return select_maximal_literal_min();
  case 13 : calc_lit_values3();return select_maximal_literal_min();
  case 14 : return heuristic4(); // random; hence no difference

  case 100 : 
    { 
      int lit = heuristic_max_obj(); 
      if (lit != 0) return lit; else return heuristic0();
    }
  case 101 : 
    { 
      int lit = heuristic_max_obj(); 
      if (lit != 0) 
	return lit; 
      else 
	{ calc_lit_values1();return select_maximal_literal_sum(); }
    }
  case 102 : 
    { 
      int lit = heuristic_max_obj(); 
      if (lit != 0) 
	return lit; 
      else 
	{ calc_lit_values2();return select_maximal_literal_sum(); } 
    }
  case 103 : 
    { 
      int lit = heuristic_max_obj(); 
      if (lit != 0) 
	return lit; 
      else 
	{ calc_lit_values3();return select_maximal_literal_sum(); }
    }
  case 104 :
    { 
      int lit = heuristic_max_obj(); 
      if (lit != 0) return lit; else return heuristic4();
    }
  case 110 : 
    { 
      int lit = heuristic_max_obj(); 
      if (lit != 0) return lit; else return heuristic0();
    }
  case 111 : 
    { 
      int lit = heuristic_max_obj(); 
      if (lit != 0) 
	return lit; 
      else 
	{ calc_lit_values1();return select_maximal_literal_min(); }
    }
  case 112 : 
    { 
      int lit = heuristic_max_obj(); 
      if (lit != 0) 
	return lit; 
      else 
	{ calc_lit_values2();return select_maximal_literal_min(); }
    }
  case 113 : 
    { 
      int lit = heuristic_max_obj(); 
      if (lit != 0) 
	return lit; 
      else 
	{ calc_lit_values3();return select_maximal_literal_min(); }
    }
  case 114 :
    { 
      int lit = heuristic_max_obj(); 
      if (lit != 0) return lit; else return heuristic4();
    }
  default : return val_heuristic_sum(value_heuristic1);
    }
}
