#include "Equations.h"
#include "stdlib.h"

Equations::Equations()
{
  f.resize(0,32);
  for(int i=0; i<= 32; i++)
    f[i] = i;  
  current = f.upper();
}

Equations::Equations(const Equations& eq)
{
  f = eq.f;
  current = eq.current;  
}

Equations& Equations::operator=(const Equations& eq)
{
  f = eq.f;
  current = eq.current;
  return *this;
}

Equations::~Equations()
{
  // Nothing special to do
}

void Equations::clear()
{
  for(int i=0; i<=f.upper(); i++)
    f[i] = i;  
  current = f.upper();
}

int Equations::operator()(int i)
{
  if (i > 0)
    {
      if (f[i] == i)
	return i;
      else
	return operator()(f[i]);
    }
  else
    {
      if (f[-i] == -i)
	return i;
      else
	return operator()(-f[-i]);
    }
}

int Equations::equal(int i,int j)
{
  return operator()(i) == operator()(j);
}

int Equations::add(int left,int right)
{
  // ensure sizes
  if ((abs(left) > f.upper()) || (abs(right) > f.upper()))
    {
      int oldupper = f.upper();
      int max = abs(left) > abs(right) ? abs(left) : abs(right);
      f.ensure(max);
      for (int i=oldupper+1;i <= max;i++)
	f[i] = i;
    }
  // add link
  int p0 = left;
  int q0 = right;
  int p1 = operator[](left);
  int q1 = operator[](right);
  while (p1 != q1)
    {
      if (abs(p1) == abs(q1))
	{
	  cerr << "Equations : add : inconsistent; neg = pos\n";
	  exit(-1);
	}
      if (abs(p1) < abs(q1))
	{
	  if (q0 > 0)
	    {
	      f[q0] = p1;
	      q0 = q1;
	      q1 = operator[](q1);
	    }
	  else
	    {
	      f[-q0] = -p1;
	      q0 = q1;
	      q1 = operator[](q1);
	    }
	}
      else
	{
	  if (p0 > 0)
	    f[p0] = q1;
	  else
	    f[-p0] = -q1;
	  p0 = p1;
	  p1 = operator[](p1);
	}
    }
  return f[left];
}

int Equations::first()
{
  current=f.upper();
  while ((current > 0) && (f[current] == current))
    current--;
  return current;
}

int Equations::last()
{
  return current==0;
}

int Equations::next()
{
  int ret = current;
  current--;
  while ((current > 0) && (f[current] == current))
    current--;
  return ret;
}

void Equations::Print(ostream& out)
{
  for(int i=0;i <= f.upper();i++)
    {
      if (f[i] != i)
	out << i << "=" << operator()(i) << " ";
    }
}

