#include "Implications.h"

// always store implications A->B such that abs(A) < abs(B)

void Implications::add(int literal,OrdInt& implied)
{
  implied.first();
  while(!implied.last())
    {
      add(literal,implied.next());
    }
}

void Implications::add(int literal,int implied)
{
  int A,B;
  if (abs(literal) != abs(implied))
    {
      if (abs(literal) < abs(implied))
	{
	  A = literal;
	  B = implied;
	}
      else
	{
	  A = -implied;
	  B = -literal;
	}
      implications.ensure(-abs(A),abs(A));
      implications[A] += B;
    }
}

void Implications::del(int literal)
{
  if (implications.upper() >= abs(literal))
    {
      implications[literal].clear();
      implications[-literal].clear();
    }
}

OrdInt Implications::get_fixed()
{
  OrdInt fixed_lits;
  OrdInt new_fixed_lits;
  for(int i=1;i <= implications.upper();i++)
    {
      if (implications[i].Intersect(implications[-i]))
	{
	  new_fixed_lits.set_Intersection(implications[i],implications[-i]);
	  fixed_lits += new_fixed_lits;
	}
      new_fixed_lits = implications[i];
      new_fixed_lits.negate_elems();
      if (implications[i].Intersect(new_fixed_lits))
	fixed_lits += -i;
      new_fixed_lits = implications[-i];
      new_fixed_lits.negate_elems();
      if (implications[-i].Intersect(new_fixed_lits))
	fixed_lits += i;      
    }
  return fixed_lits;
}

int Implications::calc_equations()
{
  if (eq == 0)
    return 0;
  OrdInt negated;
  for(int i=1;i <= implications.upper();i++)
    {
      if (!implications[i].empty() && !implications[-i].empty())
	{
	  negated = implications[i];
	  negated.negate_elems();
	  if (negated.Intersect(implications[-i]))
	    {
	      negated.Intersection(implications[-i]);
	      negated.first();
	      while(!negated.last())
		add_equation(i,-negated.next());
	    }
	}
    }
  if (eq)
    eq->first();
  return 0;
}

int Implications::equations_empty()
{
  if (eq)
    return eq->last();
  else
    return 1;
}

int Implications::pop_equation(int& left, int& right)
{
  if (!eq)
    return 0;
  if (eq->last())
    {
      cerr << "Implications : pop_equation : no equations left\n";
      exit(-1);
    }
  left = eq->next();
  right = (*eq)(left);
  return 0;
}

void Implications::add_equation(int left, int right)
{
  if (eq)
    eq->add(left,right);
}

