#ifndef _LPBIneq_h
#define _LPBIneq_h

#include "Products.h"

class LPBIneq {
public:
  LPBIneq();
  ~LPBIneq();
  LPBIneq(const LPBIneq& );
  LPBIneq(Products& p, Coefficient rhs);
  LPBIneq(SimpleString& );
  LPBIneq(char *);
  void operator=(const LPBIneq& );
  void add(Coefficient coefficient, int literal);
  void add(LPBIneq& );
  void mult(Coefficient coefficient);
  void fix(int literal);
  void equate(int lit1, int lit2);
  void clear();
  Coefficient member(int literal); // returns coefficient of literal; neg. if negated; 0 otherwise
  int no_products();
  int empty();
  Coefficient get_rhs();
  void set_rhs(Coefficient );
  Coefficient get_sumcoeffs();
  Product& max();
  Product& min();

  void normalize();
  void gcd_divide();
  int fixed_literal(); // returns a literal if it can be fixed; 0 otherwise
  int fix_is_tautology(int literal); // returns 1 if fixing literal makes ineq a tautology
  int is_tautology();     
  int is_contradiction(); 
  int simply_implies(LPBIneq& lpbi); // lpbi is "obviously" implied by ineq

  // Iterator for products
  void first();
  Product& next();
  int last();

  int check_data_structure();
  
  int initialize(SimpleString& s, Atoms* atoms = 0);
  int initialize(char* s, Atoms* atoms = 0) { SimpleString ss(s); return initialize(ss,atoms); }
  int Read(istream& in=cin,Atoms* atoms = 0);
  void Print(ostream& out=cout, Atoms* atoms = 0);
  friend ostream& operator<<(ostream& ,LPBIneq& );
  friend istream& operator>>(istream& ,LPBIneq& );

private:
  LPBIneq(Products& p); // doesn't normalize -> internal use only
  Products products;
  Coefficient rhs;
  Coefficient gcd_coefficient(ProductNode* ,Coefficient );
friend class LPBIneqSet;
friend class Enumerate;
friend class PBCS;
friend class LPBIRefs;
};

inline LPBIneq::LPBIneq() {rhs = 0;}
inline LPBIneq::~LPBIneq() {}
inline LPBIneq::LPBIneq(const LPBIneq& lpbi) : products(lpbi.products)
{ rhs = lpbi.rhs; }
inline LPBIneq::LPBIneq(SimpleString& s) : products(s)
{ rhs = -products.constant;
  products.constant = 0; }
inline LPBIneq::LPBIneq(char* s) : products(s)
{ rhs = -products.constant;
  products.constant = 0; }
inline void LPBIneq::operator=(const LPBIneq& lpbi)
{ products = lpbi.products;
  rhs = lpbi.rhs; }
inline void LPBIneq::add(Coefficient coefficient, int literal)
{
  products.add(coefficient,literal);
  rhs -= products.constant;
  products.constant = 0;
  // no normalization!
}

inline void LPBIneq::mult(Coefficient coefficient)
{
  if (coefficient < 0)
    coefficient *= -1;
  products.mult(coefficient);
  rhs *= coefficient;
}

inline void LPBIneq::clear()
{ products.clear(); rhs = 0; }
inline Coefficient LPBIneq::member(int literal) { return products.member(literal); }
inline int LPBIneq::no_products() { return products.length(); }
inline int LPBIneq::empty() { return products.empty(); }
inline Coefficient LPBIneq::get_rhs() { return rhs; }
inline void LPBIneq::set_rhs(Coefficient _rhs)
{ rhs = _rhs; normalize(); }
inline Coefficient LPBIneq::get_sumcoeffs() { return products.sumcoeffs; }
inline Product& LPBIneq::max() { return products.max(); }
inline Product& LPBIneq::min() { return products.min(); }
inline void     LPBIneq::first() { products.first(); }
inline Product& LPBIneq::next()  { return products.next(); }
inline int      LPBIneq::last()  { return products.last(); }

inline istream& operator>>(istream& in ,LPBIneq& lpbi) 
{ lpbi.Read(in); return in; }
inline ostream& operator<<(ostream& out,LPBIneq& lpbi) 
{ lpbi.Print(out); return out; }


#endif

