#include "LPBIneqSet.h"

LPBIneqSet::LPBIneqSet()
{
  greatest_iid = 0;
}

LPBIneqSet::~LPBIneqSet()
{
}

LPBIneqSet::LPBIneqSet(const LPBIneqSet& ) 
{
  cerr << "DON'T COPY CONSTRUCT LPBIneqSet";
}

LPBIneqSet& LPBIneqSet::operator=(const LPBIneqSet& )
{
  cerr << "DON'T ASSIGN LPBIneqSet";
  return *this;
}

void LPBIneqSet::clear()
{
  while(!alliids.empty())
    {
      del(alliids.min());
    }
}

int LPBIneqSet::card()
{
  return alliids.length();
}

int LPBIneqSet::add(LPBIneq& lpbi)
{
  int iid = Pool.get_new_id();
  Pool[iid] = lpbi;
  alliids += iid;
  if (iid > greatest_iid)
    greatest_iid = iid;
  return iid;
}

void LPBIneqSet::del(int iid)
{
  alliids -= iid;
  if (iid == greatest_iid)
    {
      if (alliids.empty())
	greatest_iid = alliids.empty() ? 0 : alliids.max();
    }
  Pool.free_id(iid);  
}

LPBIneq LPBIneqSet::get_lpbi(int iid)
{
  if (alliids.member(iid))
    return Pool[iid];
  LPBIneq lpbi;
  return lpbi;
}

int LPBIneqSet::fixed_literal()
{
  alliids.first();
  while(!alliids.last())
    {
      int lit = Pool[alliids.next()].fixed_literal();
      if (lit != 0)
	return lit;
    }
  return 0;
}

int LPBIneqSet::simply_implies(LPBIneq& lpbi)
{
  alliids.first();
  while(!alliids.last())
    {
      if (Pool[alliids.next()].simply_implies(lpbi))
	return 1;
    }
  return 0;
}

int LPBIneqSet::delete_implied_by(LPBIneq& lpbi)
{
  OrdInt todel;
  alliids.first();
  while(!alliids.last())
    {
      int iid = alliids.next();
      if (lpbi.simply_implies(Pool[iid]))
	todel += iid;
    }
  todel.first();
  while(!todel.last())
    del(todel.next());
  return todel.length();
}

int LPBIneqSet::delete_implied()
{
  OrdInt todel;
  OrdInt checkiids = alliids;
  checkiids.first();
  while(!checkiids.last())
    {
      int check_iid = checkiids.next();
      LPBIneq& check_lpbi = Pool[check_iid];
      alliids.first();
      while(!alliids.last())
	{
	  int iid = alliids.next();	  
	  if ((iid != check_iid) && (!todel.member(iid)))
	    { 
	      if (Pool[iid].simply_implies(check_lpbi))
		todel += check_iid;
	    }
	}
    }
  todel.first();
  while(!todel.last())
    del(todel.next());
  return todel.length();
}

OrdInt LPBIneqSet::all_vars()
{
  OrdInt av;

  alliids.first();
  while(!alliids.last())
    {
      LPBIneq& lpbi = Pool[alliids.next()];
      lpbi.first();
      while(!lpbi.last())
	av += abs(lpbi.next().literal);
    }
  return av;
}

TPool<LPBIneq> LPBIneqSet::Pool;
