#include "Linearize.h"

int linearize_fortet(PolPBTerm& t, Products& lt, LPBIneqSet& S, Atoms* atoms)
{
  lt.clear(); // the linearized term
  if (atoms == 0)
    {
      cout << "linearize_fortet : atoms must be nonzero since new variables are created\n";
      return -1;
    }
  PolPBTermNode* h = t.products;
  while (h)
    {
      LPBIneq I1,I2; 
      SimpleString s("__NL_"); // holds the new varname
      if (h->literals.length() == 1)
	{
	  h->literals.first();
	  lt.add(h->coefficient,h->literals.next());
	}
      else
	{
	  SimpleString hs;
	  h->literals.first();
	  while (!h->literals.last())
	    {
	      int lit = h->literals.next();
	      s += "_";
	      if (lit > 0)
		{
		  hs = (* atoms)(lit);
		  s += hs;
		}
	      else
		{
		  s += "~";
		  hs = (* atoms)(-lit);		
		  s += hs;
		}
	      I1.add(1,-lit);
	      I2.add(1,lit);
	    }
	  if (!(atoms->inserted(s)))
	    {
	      int newlit = (* atoms)(s);
	      I1.add(1,newlit);
	      I1.set_rhs(1);
	      I2.add(h->literals.length(),-newlit);
	      I2.set_rhs(h->literals.length());
	      if (!I1.is_tautology())
		S.add(I1);
	      if (!I2.is_tautology())
		S.add(I2);
	    }
	  lt.add(h->coefficient,(* atoms)(s));
	}
      h = h->next;
    }  
  lt.set_constant(lt.get_constant() + t.constant);
  return 0;
}

int linearize_fortet(PolPBIneq& pbi, LPBIneqSet& S, Atoms* atoms)
{
  Products p;
  linearize_fortet(pbi.t,p,S,atoms);
  LPBIneq lpbi(p,pbi.rhs);
  if (!lpbi.is_tautology())
    S.add(lpbi);
  return 0;
}

// linearization method without introducing new variables
// see P.Barth "Logic Based 0-1 Constraint Programming" Kluwer 1996
// implementation leaves much place for improvement!!!
// this one just works (hopefully :-)

int lin_split(PolPBTerm& t, int literal, PolPBTerm& pos, PolPBTerm& neg)
{
  PolPBTermNode* h = t.products;
  OrdInt lits;
  while(h)
    {
      if (!h->literals.member(literal))
	{
	  lits = h->literals;
	  lits -= -literal;
	  neg.add(h->coefficient,lits);
	}
      if (!h->literals.member(-literal)) // not else if!!!
	{
	  lits = h->literals;
	  lits -= literal;
	  pos.add(h->coefficient,lits);
	}
      h = h->next;
    }
  return 0;
}

RArray<int> occs;

int lin_splitlit(PolPBTerm& pbt)
{
  PolPBTermNode* h = pbt.products;  
#if 0
  // heuristic: first nonlinear occurence, i.e. none
  while(h)
    {
      if (h->literals.length() > 1)
	{
	  h->literals.first();
	  return h->literals.next();
	}
      h = h->next;
    }
#else
  // heuristic: max no of occurences
  OrdInt vars;
  h = pbt.products;  
  while(h) // count occurences
    {
      h->literals.first();
      while(!h->literals.last())
	{
	  int var = abs(h->literals.next());
	  occs.ensure(var);
	  if (vars.member(var))
	    occs[var] = occs[var] + 1;
	  else
	    {
	      vars += var;
	      occs[var] = 1;
	    }
	}
      h = h->next;
    }  
  h = pbt.products;  
  int splitvar = 0;
  while(h) // select best
    {
      if (h->literals.length() > 1)
	{
	  h->literals.first();
	  while(!h->literals.last())
	    {
	      if (splitvar ==  0)
		splitvar = abs(h->literals.next());
	      else
		{
		  int v = abs(h->literals.next());
		  if (occs[v] > occs[splitvar])
		    splitvar = v;
		}
	    }
	}
      h = h->next;
    }
  if (splitvar)
    return splitvar;
#endif
  cerr << "lin_splitlit : fatal error : no nonlinear occurence found \n";
  exit(-1);
  return 0;
}

// #define __Lin__Show__

Coefficient nl_lowerbound(PolPBTerm& t, LPBIneqSet& S, int fixlit, PolPBIneq& pbi)
{
  // very simple implementation!
  // a lot of possibilities to get better bounds
  // Best thing: use PBCS and enumeration (coeffred) for the bounds

  Coefficient lb = t.constant;
  PolPBTermNode* h = t.products;
  while(h)
    {
      if (h->coefficient < 0)
	lb += h->coefficient;
      h = h->next;
    }
#ifdef __Lin__Show__
  cout << "Lower Bound of " << t << " and fixing " << fixlit << " is " << lb << "\n";
#endif
  return lb;   
}

int lin_addcase(PolPBIneq& pbi, LPBIneqSet& S, Coefficient rhs, int fixlit, PolPBIneq& old)
{
  Coefficient lb = nl_lowerbound(pbi.t,S,fixlit,old);  
  pbi.rhs = rhs - pbi.t.constant;
  pbi.t.constant = 0;
  OrdInt lits_add;
  if (rhs-lb < 0 && 1) // just add positive coefficients
    {
      pbi.add(rhs-lb,lits_add); // constant
      lits_add += -fixlit;
      pbi.add(lb-rhs,lits_add);
    }
  else
    {
      lits_add += fixlit;
      pbi.add(rhs-lb,lits_add);
    }
  return linearize(pbi,S);
}

int linearize(PolPBIneq& pbi, LPBIneqSet& S)
{
#ifdef __Lin__Show__
  cout << pbi << "\n";
#endif
  if (pbi.is_linear())
    {
#ifdef __Lin__Show__
      cout << "Lin\n";
#endif
      LPBIneq I;
      int literal;
      Coefficient coefficient;
      PolPBTermNode* h = pbi.t.products;
      while(h)
	{
	  h->literals.first();
	  literal = h->literals.next();
	  coefficient = h->coefficient;
	  I.add(coefficient,literal);
	  h = h->next;
	}
      I.set_rhs(I.get_rhs() + pbi.rhs);
      if (!I.is_tautology())
	{
#ifdef __Lin__Show__
	  cout << "adding : " << I << "\n";
#endif	  	  
	  S.add(I);
	}
      return 0; 
    }
  // really nonlinear! (currently don't check for compact form)
  // split ineq
  PolPBIneq pos;
  PolPBIneq neg;
  int splitlit = lin_splitlit(pbi.t);
#ifdef __Lin__Show__
  cout << "splitting on " << splitlit << "\n";
#endif
  lin_split(pbi.t,splitlit,pos.t,neg.t);
#ifdef __Lin__Show__
  cout << "pos : " << pos.t << "\n";
  cout << "neg : " << neg.t << "\n";
#endif
  int ret;
  if ((ret = lin_addcase(pos,S,pbi.rhs,-splitlit,pbi)))
    return ret;
  return lin_addcase(neg,S,pbi.rhs,splitlit,pbi);
}

#ifdef __GNUC__
template class RArray<int>; 
#endif
