#include "LitRefs.h"

LitRefs::LitRefs(int size)
{
  litref.ensure(-size,size);
  for (int i = 0 ;i <= litref.upper();i++)
    litref[i][0].iid = litref[-i][0].iid = 0;
}

LitRefs::LitRefs(const LitRefs& LR)
{
  litref = LR.litref;
}

LitRefs& LitRefs::operator=(const LitRefs& LR)
{
  litref = LR.litref;
  return *this;
}

LitRefs::~LitRefs()
{
}

int LitRefs::greatest_variable()
{
  int gv = 0;
  int i;
  for (i=litref.lower(); i < 0; i++)
    {
      if (litref[i][0].iid != 0)
	{
	  gv = -i;
	  break;
	}
    }
  for (i=litref.upper(); i > 0; i--)
    {
      if (litref[i][0].iid != 0)
	{
	  gv = gv > i ? gv : i;
	  break;
	}
    }
  return gv;
}

int LitRefs::add(int iid, LPBIneq& lpbi)
{
  lpbi.first();
  while(!lpbi.last())
    {
      Product& p = lpbi.next();
      add(iid,p.literal,p.coefficient);
    }
  return 0;
}

int LitRefs::del(int iid, LPBIneq& lpbi)
{
  lpbi.first();
  while(!lpbi.last())
    {
      Product& p = lpbi.next();
      del(iid,p.literal,p.coefficient);
    }
  return 0;
}

int LitRefs::del(int iid, Products& ps)
{
  ps.first();
  while(!ps.last())
    {
      Product& p = ps.next();
      del(iid,p.literal,p.coefficient);      
    }
  return 0;
}

void LitRefs::clear()
{
  for (int i = litref.lower() ; i <= litref.upper(); i++)
    litref[i][0].iid = 0;
}

int LitRefs::add(int iid, int literal, Coefficient coefficient)
{
// cout << "adding " << coefficient << "*" << literal << "\n";
  // ensure size in litref
  if (abs(literal) > litref.upper())
    {
      int greatestvariable = abs(literal);
      int to = litref.upper();
      litref.ensure(-2*greatestvariable,2*greatestvariable);
      for(int i=to+1;i<=litref.upper();i++)
	litref[i][0].iid = litref[-i][0].iid = 0;
    }
  RArray<LitRef>& refs = litref[literal];
  // ensure size in refs
  if (refs[0].iid >= refs.upper())
    refs.ensure(2*refs[0].iid);
  refs[0].iid++;
  refs[refs[0].iid].iid = iid;
  refs[refs[0].iid].coefficient = coefficient;
  return refs[0].iid; // return number of references
}

int LitRefs::del(int iid, int literal, Coefficient )
{
  RArray<LitRef>& refs = litref[literal];
  int i=1;
  while ((i <= refs[0].iid) && (refs[i].iid != iid))
    i++;
  if (i > refs[0].iid)
    {
      cout << "LitRefs : del : datastructure corrupted \n";
//      cout << "deleting " << iid << " " << coefficient << " " << literal;
    }
  // copy last to deleted entry and update number of entries
  refs[i] = refs[refs[0].iid--];
  return refs[0].iid;
}

int LitRefs::Print_One(int i,ostream& out, Atoms* atoms)
{
  RArray<LitRef>& refs = litref[i];
  int j=0;
  if (refs[0].iid != 0)
    {
      plit(i,out,atoms);
      out << " : ";
      j = 1;
      while (j <= refs[0].iid)
	{
	  out << refs[j].iid << "->" << refs[j].coefficient << " ";
	  j++;
	}
      out << "\n";
    }
  return j;
}

int LitRefs::Print(ostream& out,Atoms* atoms)
{
  for (int i=0;i <= litref.upper();i++)
    {
      Print_One(i,out,atoms);
      Print_One(-i,out,atoms);
    }
  return 0;
}




