#ifndef LitRefs_h
#define LitRefs_h

#include "LPBIneq.h"
#include "RArray.h"
extern "C" {
#include "stdlib.h"
}

class LitRef {
public:
  int iid;
  Coefficient coefficient;
  LitRef() {};
  LitRef(const LitRef& l) : iid(l.iid), coefficient(l.coefficient) {}
  void operator=(const LitRef& l) { iid = l.iid; coefficient = l.coefficient; }
  ~LitRef() {};
};

class LitRefs {
public:
  LitRefs(int size=128);
  LitRefs(const LitRefs& );
  LitRefs& operator=(const LitRefs&);
  ~LitRefs();

  RArray<LitRef>& operator[](int i) { return litref[i]; }

  int var_occurs(int i)    { return ((litref[i][0].iid != 0) || (litref[-i][0].iid != 0)); }
  int var_no_occurs(int i) { return litref[i][0].iid + litref[-i][0].iid; }
  int lit_occurs(int i)    { return (litref[i][0].iid != 0); }
  int lit_no_occurs(int i) { return litref[i][0].iid; }  
  int greatest_variable();

  // steps through all variables that occur in an lpbi
  int first_var();
  int last_var();
  int next_var();
  
  int upper() { return litref.upper(); }
  int lower() { return litref.lower(); }

  int add(int iid, LPBIneq& lpbi);
  int del(int iid, LPBIneq& lpbi);
  int del(int iid, Products& ps);
  void clear();
  // returns the number of references of this literal
  int add(int iid, int literal, Coefficient coefficient);
  int del(int iid, int literal, Coefficient coefficient);
  
  int Print_One(int i,ostream& out=cout, Atoms* atoms=0);
  int Print(ostream& out=cout, Atoms* atoms=0);

private:
  // The index datastructure by literal reference
  // litref[i][0].iid == the number of entries in the litref
  RArray< RArray<LitRef> > litref;
  int current_lit;

  int LitRefs::plit(int lit,ostream& out = cout,Atoms* atoms=0);
};

inline int LitRefs::first_var()
{
  current_lit = 1;
  while ((current_lit <= litref.upper()) && (!var_occurs(current_lit)))
    current_lit++;
  return current_lit;
}

inline int LitRefs::last_var()
{
  return current_lit > litref.upper();
}

inline int LitRefs::next_var()
{
  int ret = current_lit++;
  while ((current_lit <= litref.upper()) && (!var_occurs(current_lit)))
    current_lit++;  
  return ret;
}

inline int LitRefs::plit(int lit,ostream& out,Atoms* atoms)
{
  if (atoms == 0)
    out << lit;
  else
    {
      if (lit < 0)
	out << "~" << (*atoms)(-lit);
      else
	out << (*atoms)(lit);
    }
  return 0;
}

#endif
