#include "OrdInt.h"
extern "C" {
#include "stdlib.h"
}

OrdInt::OrdInt()
{
  head = new OrdIntNode;
  head->next = NULL;
  head->key = 0;
}

OrdInt::~OrdInt()
{
#ifdef OWN_MEMORY_OrdIntNode
  // insert the whole chain at once at the beginning of the freelist
  OrdIntNode* oi = head;
  
  while (oi->next) 
    oi = oi->next;
  oi->next = OrdIntNode::oifree;
  OrdIntNode::oifree = head;
#else
  OrdIntNode* oi = head->next;
  OrdIntNode* oinext;

  while(oi)
    {
      oinext = oi->next;
      delete oi;
      oi = oinext;
    }
  delete head;
#endif
}

inline void OrdInt::free_list(OrdIntNode* oi)
{
#ifdef OWN_MEMORY_OrdIntNode
  // insert the whole chain at once at the beginning of the freelist
  if (oi == NULL)
    return;
  OrdIntNode* first = oi;
  
  while (oi->next) 
    oi = oi->next;
  oi->next = OrdIntNode::oifree;
  OrdIntNode::oifree = first;
#else
  OrdIntNode* oinext;

  while(oi)
    {
      oinext = oi->next;
      delete oi;
      oi = oinext;
    }
#endif  
}

OrdInt::OrdInt(const OrdInt& oi)
{
#ifdef OWN_MEMORY_OrdIntNode
  OrdIntNode* tocopy = oi.head;

  head = new OrdIntNode;
  head->next = OrdIntNode::oifree;
  head->key = 0;
  current = head;
  while (tocopy->next)
    {
      if (OrdIntNode::oifree) // fine, we have some left
	{
	  OrdIntNode::oifree->key = tocopy->next->key;
	  tocopy = tocopy->next;
          current->next = OrdIntNode::oifree;  // neccesary
	  current = OrdIntNode::oifree;
	  OrdIntNode::oifree = OrdIntNode::oifree->next;
	}
      else // standard allocation
        {
	  OrdIntNode* newnode = new OrdIntNode;
	  
	  newnode->key = tocopy->next->key;
	  current->next = newnode;
	  current = newnode;
	  tocopy = tocopy->next;	  
	}
    }
  current->next = NULL;
#else
  OrdIntNode* tocopy = oi.head;
  OrdIntNode* current;

  head = new OrdIntNode;
  head->next = NULL;
  head->key = 0;
  current = head;
  while (tocopy->next)
    {
      OrdIntNode* newnode = new OrdIntNode;
      
      newnode->key = tocopy->next->key;
      current->next = newnode;
      current = newnode;
      tocopy = tocopy->next;
    }      
  current->next = NULL;
#endif
}

OrdInt::OrdInt(SimpleString& s)
{
  head = new OrdIntNode;
  head->next = NULL;
  head->key = 0;
  this->initialize(s);
}

void OrdInt::operator=(const OrdInt& oi)
{
  OrdIntNode *tocopy;

  if (head == oi.head) // nothing to do, the same OrdInt
    return;
  tocopy = oi.head;
  current = head;
  while (tocopy->next && current->next) // Just copy the information as long as possible
    {
      current->next->key = tocopy->next->key;
      tocopy = tocopy->next;
      current = current->next;
    }
  if (tocopy->next == NULL)  // All copied, just free the remaining nodes of current
    {
      OrdIntNode* oin = current->next;
      OrdIntNode* oinext;

      while(oin)
	{
	  oinext = oin->next;
	  delete oin;
	  oin = oinext;
	}
      current->next = NULL;
    }
  else // if (current->next == NULL) // Construct New ones for the rest
    {
      while (tocopy->next)
	{
	  OrdIntNode* newnode = new OrdIntNode;

	  newnode->key = tocopy->next->key;
	  current->next = newnode;
	  current = newnode;
	  tocopy = tocopy->next;
	}      
      current->next = NULL;
    }	  
}

void OrdInt::add(int i)
{
  OrdIntNode* newnode = new OrdIntNode;
  OrdIntNode* h = head;

  if (newnode == NULL)
    cerr << "SHIT \n";
  newnode->key = i;
  while (h->next && (h->next->key < i))
    h = h->next;
  if (h->next == NULL)
    {
      newnode->next = NULL;
      h->next = newnode;
    }
  else if (h->next->key == i)
    delete newnode;
  else
    {
      newnode->next = h->next;
      h->next = newnode;
    }
}

void OrdInt::del(int i)
{
  OrdIntNode* h = head;

  while (h->next && (h->next->key < i))
    h = h->next;
  if (h->next != NULL)
    {
      if (h->next->key == i)
	{
	  OrdIntNode* hh = h->next;
	  h->next = h->next->next;
	  delete hh;
	}
    }
}

void OrdInt::clear()
{
  OrdIntNode* h = head->next;
  OrdIntNode* hh;

  while (h)
    {
      hh = h->next;
      delete h;
      h = hh;
    }
  head->next = NULL;
}

int OrdInt::pop()
{
  if (head->next == NULL)
    return 0;
  else
    {
      int info = head->next->key;
      OrdIntNode* h = head->next;

      head->next = h->next;
      delete h;
      return info;
    }
}

int OrdInt::pop_random()
{
  int len=length();
  int info;
  if (len == 0)
    return 0;
  else if (len == 1)
    return pop();
  else // get_random_position
    {
      int which = rand()%len;
      OrdIntNode* h = head;
      while (which--)
	h = h->next;
      info = h->next->key;
      OrdIntNode* hh = h->next;
      h->next = h->next->next;
      delete hh;     
    }
  return info;
}

int OrdInt::member(int i)
{
  OrdIntNode* actual = head->next;
  
  while (actual)
    {
      if (actual->key < i)
	actual = actual->next;
      else 
	return (actual->key == i);
    }
  return 0; 
}


int OrdInt::length()
{
  int len = 0;
  OrdIntNode* oi = head->next;

  while(oi)
    {
      len++;
      oi = oi->next;
    }
  return len;
}

int OrdInt::min()
{
  OrdIntNode* actual = head->next;
  
  if(actual)
    return actual->key;
  else
    {
      cerr << "OrdInt : min of emptyset; returning 0";
      return 0;
    }
}

int OrdInt::max()
{
  OrdIntNode* actual = head;
  
  if (!actual->next)
    {
      cerr << "OrdInt : max of emptyset; returning 0";
      return 0;
    }
  else
    {
      while(actual->next)
	actual = actual->next;
      return actual->key;
    }
}

// lexicographical ordering
int OrdInt::compare(OrdInt& oi)
{
  OrdIntNode* cur = head->next;
  OrdIntNode* other = oi.head->next;

  while (cur && other)
    {
      if (cur->key < other->key)
	return -1;
      if (cur->key > other->key)
	return 1;
      else // if (cur->key == other->key)
	{
	  cur = cur->next;
	  other = other->next;
	}
    }
  if (other)
    return -1;
  else if (cur)
    return 1;
  else
    return 0;
}

// special function for set_... members
inline OrdIntNode* OrdInt::destructive_copy(OrdIntNode* cur,int key)
{
  if (cur->next != NULL) // copy in remaining nodes
    {
      cur->next->key = key;
    }
  else // get new node first
    {
      OrdIntNode* newnode = new OrdIntNode;
      if (newnode == NULL)
	cerr << "SHIT \n";
      newnode->key = key;
      newnode->next = NULL;
      cur->next = newnode;
    }
  return cur->next;
}

void OrdInt::Subtract(OrdInt& oi)
{
  OrdIntNode* s1=head;
  OrdIntNode* s2=oi.head;

  if (!s1->next || !s2->next)
    return;
  for(;;)
    {
      if (s1->next->key < s2->next->key)
	{
	  s1 = s1->next;
	  if (!s1->next)
	    return;
	}
      else if (s1->next->key > s2->next->key)
	{
	  s2 = s2->next;
	  if (!s2->next)
	    return;
	}
      else /* if (s1->next->key == s2->next->key) */
	{	
	  OrdIntNode* h = s1->next;

	  s1->next = s1->next->next;
	  delete h;
	  if (!s1->next)
	    return;
	}
    }
}

void OrdInt::set_Subtract(OrdInt& oi1, OrdInt& oi2)
{
  OrdIntNode* s1 = oi1.head->next;
  OrdIntNode* s2 = oi2.head->next;
  current = head;

  if (s1 && s2)
    {
      for(;;) // while (s1 && s2) moved in the loop
	{
	  if (s1->key < s2->key)
	    {
	      current = destructive_copy(current,s1->key);
	      if ((s1 = s1->next) == NULL)
		break;
	    }
	  else if (s1->key > s2->key)
	    {
	      if ((s2 = s2->next) == NULL)
		break;
	    }
	  else /* if (s1->key == s2->key) */
	    {
	      if ((s1 = s1->next) == NULL)
		break;
	      if ((s2 = s2->next) == NULL)
		break;
	    }
	}
    }
  while (s1) // copy rest
    {
      current = destructive_copy(current,s1->key);
      s1 = s1->next;
    }
  free_list(current->next);
  current->next = NULL;
}


// this = oi1\setminus oi2, i3 = |oi1\cap oi2|, i4 = |oi2\setminus oi1|
void OrdInt::set_Subtract_Intersectionlen_Subtractlen(
	      OrdInt& oi1,OrdInt& oi2, int& i3, int& i4)
{
  OrdIntNode* s1 = oi1.head->next;
  OrdIntNode* s2 = oi2.head->next;
  current = head;
  i3 = i4 = 0;

  if (s1 && s2)
    {
      for(;;) // while (s1 && s2) moved in the loop
	{
	  if (s1->key < s2->key)
	    {
	      current = destructive_copy(current,s1->key);
	      if ((s1 = s1->next) == NULL)
		break;
	    }
	  else if (s1->key > s2->key)
	    {
	      i4++;
	      if ((s2 = s2->next) == NULL)
		break;
	    }
	  else /* if (s1->key == s2->key) */
	    {
	      i3++;
	      s1 = s1->next;
	      s2 = s2->next;
	      if ((s1  == NULL) || (s2 == NULL))
		break;
	    }
	}
    }
  while (s1) // copy rest
    {
      current = destructive_copy(current,s1->key);
      s1 = s1->next;
    }
  while (s2) // count rest
    {
      i4++;
      s2 = s2->next;
    }
  free_list(current->next);
  current->next = NULL;  
}


int OrdInt::Subtract_length(OrdInt& oi)
{
  OrdIntNode* s1 = head->next;
  OrdIntNode* s2 = oi.head->next;
  int s_length = 0;

  if (s1 && s2)    
    for(;;) // while (s1 && s2) moved inline
      {
	if (s1->key < s2->key)
	  {
	    s_length++;
	    if ((s1 = s1->next) == NULL)
	      break;
	  }
       else if (s1->key > s2->key)
	 {
	   if ((s2 = s2->next) == NULL)
	     break;
	 }
       else // if (s1->next->key == s2->next->key)
	 {
	   if ((s1 = s1->next) == NULL)
	     break;
	   if ((s2 = s2->next) == NULL)
	     break;
	 }
      }
  while (s1)
    {
      s_length++;
      s1 = s1->next;
    }
  return s_length;
}

int OrdInt::Subset(OrdInt& oi)
{
  OrdIntNode* s1 = head->next;
  OrdIntNode* s2 = oi.head->next;

  while(s1 && s2)
    {
      if (s1->key < s2->key)
	s1 = s1->next;
      else if (s1->key > s2->key)
	return 0;
      else // if (s1->key == s2->key)
        {
	  s1 = s1->next;
	  s2 = s2->next;
	}
    }
  if (s2 == NULL)
    return 1;
  else
    return 0;
}

void OrdInt::negate_elems()
{
  current = head->next;
  if (current == NULL) // empty
      return;
  if (current->next == NULL) // one element
    {
      current->key = - current->key;
      return;
    }
  // reverse pointers inline, thus reverses the list and change sign of key
  OrdIntNode* previous = NULL;
  OrdIntNode* h = current->next;
  while(current->next)
    {
      current->key = - current->key;
      current->next = previous;
      previous = current;
      current = h;
      h = current->next;
    }
  current->key = - current->key;
  current->next = previous;
  head->next = current;
}

void OrdInt::swap(OrdInt& oi1 )
{
  if (oi1.head == head)
    return;
  OrdIntNode* h = oi1.head->next;
  oi1.head->next = head->next;
  head->next = h;
  h = oi1.current;
  oi1.current = h;
  current = h;
}

void OrdInt::Union(OrdInt& oi)
{
  OrdIntNode* s1 = head;
  OrdIntNode* s2 = oi.head;

  while (s1->next && s2->next)
    {
      if (s1->next->key < s2->next->key)
	  s1 = s1->next;
      else if (s1->next->key > s2->next->key)
	{
	  OrdIntNode* h = s1->next;
	  s1->next = new OrdIntNode;
	  s1->next->key = s2->next->key;
	  s1->next->next = h;
	  s1 = s1->next;
	  s2 = s2->next;
	}
      else /* if (s1->next->key == s2->next->key) */
	{	  
	  s1 = s1->next;
	  s2 = s2->next;
	}
    }
  if (s1->next == NULL) // add the rest
    {
      while(s2->next)
	{
	  s1->next = new OrdIntNode;
	  s1->next->key = s2->next->key;
	  s1 = s1->next;
	  s2 = s2->next;
	}
      s1->next = NULL;
    }
}

void OrdInt::Intersection(OrdInt& oi)
{
  OrdIntNode* s1 = head;
  OrdIntNode* s2 = oi.head;

  while (s1->next && s2->next)
    {
      if (s1->next->key < s2->next->key)
        {
          OrdIntNode* h = s1->next;
          
          s1->next = s1->next->next;
          delete h;
        }
      else if (s1->next->key > s2->next->key)
          s2 = s2->next;
      else /* if (s1->next->key == s2->next->key) */
        {
          s1 = s1->next;
          s2 = s2->next;
        }
    }
  while (s1->next)
   {
     OrdIntNode* h = s1->next;
     
     s1->next = s1->next->next;
     delete h;
   }
}

void OrdInt::set_Intersection(OrdInt& oi1, OrdInt& oi2)
{
  OrdIntNode* s1 = oi1.head->next;
  OrdIntNode* s2 = oi2.head->next;
  current = head;

  if (s1 && s2)
    {
      for(;;) // while (s1 && s2) moved in the loop
	{
	  if (s1->key < s2->key)
	    {
	      if ((s1 = s1->next) == NULL)
		break;
	    }
	  else if (s1->key > s2->key)
	    {
	      if ((s2 = s2->next) == NULL)
		break;
	    }
	  else /* if (s1->key == s2->key) */
	    {
	      current = destructive_copy(current,s1->key);
	      if ((s1 = s1->next) == NULL)
		break;
	      if ((s2 = s2->next) == NULL)
		break;
	    }
	}
    }
  free_list(current->next);
  current->next = NULL;
}

// this := oi1 \cap oi2
// oi3 := oi1 \setminus (oi1 \cap oi2)
void OrdInt::set_Intersection_Subtract(OrdInt& oi1, OrdInt& oi2, OrdInt& oi3)
{
  OrdIntNode* s1 = oi1.head->next;
  OrdIntNode* s2 = oi2.head->next;
  current = head;
  OrdIntNode* current3 = oi3.head;

  if (s1 && s2)
    {
      for(;;) // while (s1 && s2) moved in the loop
	{
	  if (s1->key < s2->key)
	    {
	      current3 = destructive_copy(current3,s1->key);
	      if ((s1 = s1->next) == NULL)
		break;
	    }
	  else if (s1->key > s2->key)
	    {
	      if ((s2 = s2->next) == NULL)
		break;
	    }
	  else /* if (s1->key == s2->key) */
	    {
	      current = destructive_copy(current,s1->key);
	      if ((s1 = s1->next) == NULL)
		break;
	      if ((s2 = s2->next) == NULL)
		break;
	    }
	}
    }
  free_list(current->next);
  current->next = NULL;
  while (s1)  // copy remaining from s1
    {
      current3 = destructive_copy(current3,s1->key);
      s1 = s1->next;
    }
  free_list(current3->next);
  current3->next = NULL;
}


void OrdInt::set_Intersection_Subtract_Subtract(
		OrdInt& oi1, OrdInt& oi2, OrdInt& oi3, OrdInt& oi4)
{
  OrdIntNode* s1 = oi1.head->next;
  OrdIntNode* s2 = oi2.head->next;
  current = head;
  OrdIntNode* current3 = oi3.head;
  OrdIntNode* current4 = oi4.head;

  if (s1 && s2)
    {
      for(;;) // while (s1 && s2) moved in the loop
	{
	  if (s1->key < s2->key)
	    {
	      current3 = destructive_copy(current3,s1->key);
	      if ((s1 = s1->next) == NULL)
		break;
	    }
	  else if (s1->key > s2->key)
	    {
	      current4 = destructive_copy(current4,s2->key);
	      if ((s2 = s2->next) == NULL)
		break;
	    }
	  else /* if (s1->key == s2->key) */
	    {
	      current = destructive_copy(current,s1->key);
	      s1 = s1->next;
	      s2 = s2->next;
	      if ((s1 == NULL) || (s2 == NULL))
		break;
	    }
	}
    }
  free_list(current->next);
  current->next = NULL;
  while (s1)  // copy remaining from s1
    {
      current3 = destructive_copy(current3,s1->key);
      s1 = s1->next;
    }
  free_list(current3->next);
  current3->next = NULL;
  while (s2)  // copy remaining from s2
    {
      current4 = destructive_copy(current4,s2->key);
      s2 = s2->next;
    }
  free_list(current4->next);
  current4->next = NULL;
}

int OrdInt::Intersection_length(OrdInt &oi)
{
  OrdIntNode* s1 = head->next;
  OrdIntNode* s2 = oi.head->next;
  int i_length = 0;

  if (s1 && s2)
    for(;;) // while (s1 && s2) moved inline
      {
	if (s1->key < s2->key)
	  {
	    if ((s1 = s1->next) == NULL)
	      break;
	  }
       else if (s1->key > s2->key)
	 {
	   if ((s2 = s2->next) == NULL)
	     break;
	 }
       else // if (s1->next->key == s2->next->key)
	 {
	   i_length++;
	   if ((s1 = s1->next) == NULL)
	     break;
	   if ((s2 = s2->next) == NULL)
	     break;
	  }
      }
  return i_length;
}

int OrdInt::Intersect(OrdInt& oi)
{
  OrdIntNode* s1 = head->next;
  OrdIntNode* s2 = oi.head->next;

  while (s1 && s2)
    {
      if (s1->key < s2->key)
	s1 = s1->next;
      else if (s1->key > s2->key)
	s2 = s2->next;
      else // if (s1->next->key == s2->next->key)
        return 1;
    }
  return 0;
}

void OrdInt::Symdiff(OrdInt& oi)
{
  OrdIntNode* s1 = head;
  OrdIntNode* s2 = oi.head;

  while (s1->next && s2->next)
    {
      if (s1->next->key < s2->next->key)
	s1 = s1->next;
      else if (s1->next->key > s2->next->key)
	{
	  OrdIntNode* h = s1->next;
	  s1->next = new OrdIntNode;
	  s1->next->key = s2->next->key;
	  s1->next->next = h;
	  s2 = s2->next;
	}
      else /* if (s1->next->key == s2->next->key) */
	{	
	  OrdIntNode* h = s1->next;
	  s1->next = s1->next->next;
	  delete h;
	  s2 = s2->next;
	}
    }
  if (s1->next == NULL) // add the rest
    {
      while(s2->next)
	{
	  s1->next = new OrdIntNode;
	  s1->next->key = s2->next->key;
	  s1 = s1->next;
	  s2 = s2->next;
	}
      s1->next = NULL;
    }
}

/* io - routines */

void OrdInt::Print(ostream& out)
{
  OrdIntNode* oi = head->next;

  while(oi)
    {
      out << oi->key << " ";
      oi = oi->next;
    }
}

int OrdInt::initialize(SimpleString& s)
{
  int len = s.length();
  int oi = 0;
  int i  = 0;
  int sign = 1;
  int started = 0;
  
  while (i < len)
    {
      switch (s[i]) {
      case ' '  :
      case '\t' :
	if (started && (oi == 0))
	  { cerr << "OrdInt::initialize : zero read\n"; return 1; }
	if (oi != 0)
	  {
	    add(sign*oi);
	    oi = 0; sign = 1; started = 0;
	  }
	break;
      case '-' :
	if (started)
	  { cerr << "OrdInt::initialize : - sign wrong place\n"; return 1; }
	else
	  { started = 1; sign = -1;}
	break;
      case '+' :
	if (started)
	  { cerr << "OrdInt::initialize : + sign wrong place\n"; return 1; }
	else
	  { started = 1; sign = 1;}
	break;
      default :
	if ((s[i] >= '0') && (s[i] <= '9'))
	  {
	    started = 1;
	    oi = oi*10 + (s[i] - '0');
	  }
	else
	  { cerr << "OrdInt::initialize : unexpected char\n"; return 1; }
	break;
      }
      i++;
    }  
  if (oi != 0) // add last one
    add(sign*oi);
  return 0;
}

int OrdInt::Read(istream& in)
{
  SimpleString s;

  s.Read(in,"\n"); // change to not require a "\n" if neccesary
  if (this->initialize(s))
    return 1;
  else
    return 0;
}

// unordered special modifications, not for public use
void OrdInt::unordered_push(int i)
{
  OrdIntNode* newnode = new OrdIntNode;
  OrdIntNode* h = head;

  if (newnode == NULL)
    cerr << "SHIT \n";
  newnode->key = i;
  newnode->next = h->next;
  h->next = newnode;
}

/////////////////////////////
// OrdIntNode
/////////////////////////////

// should not be used
OrdIntNode::OrdIntNode(const OrdIntNode& oi)
{
  cerr << "Hey, copy constructor of OrdIntNode" << " should not be called\n";
  cerr << oi.key << "\n";
}

// should not be used
void OrdIntNode::operator=(const OrdIntNode& oi)
{
  cerr << "Hey, assignment of OrdIntNode" << " should not be called\n";
  cerr << oi.key << "\n";
}

#ifdef OWN_MEMORY_OrdIntNode
OrdIntNode* OrdIntNode::oifree = (OrdIntNode *) NULL;
#endif
