#ifndef _OrdInt_h
#define _OrdInt_h

#include "streaminclude.h"
#include <math.h>
#include "SimpleString.h"

class OrdIntNode;

class OrdInt {
public:
  OrdInt();
  ~OrdInt();
  OrdInt(const OrdInt& );
  OrdInt(SimpleString& );
  void operator=(const OrdInt& );
  void add(int );
  void operator+=(int );
  void del(int );
  void operator-=(int );
  int pop();
  int pop_random(); // selects a random element and deletes it
  void clear();
  int member(int );
  int length();
  int empty();
  int min();
  int max();
  int compare(OrdInt& );

  // Iterator
  void first();
  int next();
  int last();

  void negate_elems();
  void swap(OrdInt& );
  
  void Union(OrdInt& );
  void operator+=(OrdInt& );
  void Intersection(OrdInt& );  
  void operator*=(OrdInt& );
  void set_Intersection(OrdInt&, OrdInt&);
  // sets this to the intersection of 1st and 2nd; and 3rd arg to 1st setminus intersection
  void set_Intersection_Subtract(OrdInt&, OrdInt&, OrdInt& );
  // sets this to the intersection of 1st and 2nd; 3rd arg to 1st setminus intersection
  // 4th arg to 2nd setminus intersection
  void set_Intersection_Subtract_Subtract(OrdInt&, OrdInt&, OrdInt& ,OrdInt& );
  int Intersection_length(OrdInt &);
  int Intersect(OrdInt& ); // just checks whether intersection is not empty
  void Subtract(OrdInt& );
  void set_Subtract(OrdInt& ,OrdInt& );
  void set_Subtract_Intersectionlen_Subtractlen(OrdInt& ,OrdInt&, int&, int& );
  int Subtract_length(OrdInt& );
  void operator-=(OrdInt& );
  int Subset(OrdInt& );
  void Symdiff(OrdInt& );
  void operator/=(OrdInt& );

  int initialize(SimpleString& s);
  int Read(istream& in=cin);
  void Print(ostream& out=cout);
  friend ostream& operator<<(ostream& ,OrdInt& );
  friend istream& operator>>(istream& ,OrdInt& );

  // special; these routines disregard the ordering restriction
  // whenever an ordint is modified with these routines don't use the
  // previous operations
  void unordered_push(int );

private:
  OrdIntNode* head;
  OrdIntNode* current;
  OrdIntNode* destructive_copy(OrdIntNode*,int);
  void free_list(OrdIntNode*);
friend class Literals;
};

inline void Read(OrdInt& oi, istream& in) {oi.Read(in);}
inline void Print(OrdInt& oi, ostream& out) { oi.Print(out); }
inline istream& operator>>(istream& in ,OrdInt& oi) { oi.Read(in); return in; }
inline ostream& operator<<(ostream& out,OrdInt& oi) { oi.Print(out); return out; }
inline int compare(OrdInt& oi1, OrdInt& oi2) { return oi1.compare(oi2); }

// inlines
inline void OrdInt::operator+=(int i) {  add(i); }
inline void OrdInt::operator-=(int i) {  del(i); }
inline void OrdInt::operator+=(OrdInt& oi) {  Union(oi); }
inline void OrdInt::operator*=(OrdInt& oi) {  Intersection(oi); }
inline void OrdInt::operator-=(OrdInt& oi) {  Subtract(oi); }
inline void OrdInt::operator/=(OrdInt& oi) {  Symdiff(oi); }

class OrdIntNode {
public:
  OrdIntNode() {};
  ~OrdIntNode() {};
  OrdIntNode(const OrdIntNode& oi);
  void operator=(const OrdIntNode& oi);

#define OWN_MEMORY_OrdIntNode
#ifdef OWN_MEMORY_OrdIntNode
  // We use our own memory allocation as described in Stroustrupp page 176
  // Because of a compiler error in cfront we can't use decl. + inline,
  // but have to include the code in the class.
  void* operator new(size_t)
  {
    register OrdIntNode* p = oifree; // first allocate
      
      if (p)
	oifree = p->next;
      else
	{
	  OrdIntNode* q = (OrdIntNode *) new char[NALL * sizeof(OrdIntNode)];
	  for (p=oifree=&q[NALL-1]; q<p; p--)
	    p->next = p-1;
	  (p+1)->next = NULL;
	}
    return p;
  }
  void operator delete(void* p)  // ,size_t ) deleted, does not work with g++2.5.2
  {
    ((OrdIntNode*)p)->next = oifree;
    oifree = (OrdIntNode *) p;    
  }
private:
  int key;
  OrdIntNode* next;
  enum {NALL = 2048 };
  static OrdIntNode* oifree;
#else
private:
  int key;
  OrdIntNode* next;
#endif
friend class OrdInt;
friend class Literals;
};

// inlines which use OrdIntNode; so because of the silly bloddy
// C++ compilers we have to append them at the end
inline void OrdInt::first() {  current = head; }
inline int OrdInt::next() 
{ 
  if (current->next) 
    { current = current->next; 
      return current->key; }
  else return 0;
}
inline int OrdInt::last()  { return (current->next) ? 0 : 1; }
 

// further OrdInt inlines that have to be appended on the end, since
// OrdIntNode* is not really known before !%^$@#
inline int OrdInt::empty() { return (head->next == NULL) ? 1 : 0; }



#endif



