#include "PBCS.h"

int PBCS::coefficient_reduction_bigM(int iid)
{
  Coefficient rhs = S.Pool[iid].get_rhs();
  if (rhs == 1)
    rhs = 2;
  return coefficient_reduction(iid,rhs);
}

int PBCS::coefficient_reduction_bigM()
{
  if (init_enumeration())
    return 0;
  int changed = 0;
  OrdInt alliids = S.alliids; // copy since coeffred may change things
  alliids.first();
  while(!alliids.last())
    {
      changed += coefficient_reduction_bigM(alliids.next());
    }
  return changed;
}

int PBCS::coefficient_reduction(int iid, Coefficient coeffbound)
{
  int changed = 0; // no of changed coefficients
  LPBIneq& lpbi = S.Pool[iid]; // a copy since coeffred may change things
  
  if (verbosity >= 5)
    cout << "[" << iid << "]";
  // Enumeration must be already initialized
  Products term = lpbi.products;
  
  lpbi.first();
  while(!lpbi.last())
    {
      // reduce c_iL_i + cL >= d
      Product& p = lpbi.next();
      if (p.coefficient < coeffbound) 
	break;
      if (verbosity >= 7)
	cout << p.coefficient << "*" << p.literal;
      cout.flush();
      term.del(p.literal);
      Enum.EnumTablesave();
      if (Enum.fix_lit(p.literal) == Enum.UNSAT)
	{
	  // do not do coefficient reduction in that case
	  Enum.EnumTablerestore(); 
	  term.add(p.coefficient,p.literal);
	  break;
	}
      Coefficient lb = Enum.LowerBound(term);
      if (verbosity >= 7)
	{ cout << ">=" << lb << "|";cout.flush(); }
      if (lb >= lpbi.rhs) 
	{
	  // cL >= d must hold. Just let c_iL_i be eliminated
	  changed++;
	  if (verbosity >= 5)
	    cout << "D";
	}
      else if (lpbi.rhs-lb < p.coefficient)
	{
	  // d-lbL_i + cL >= d holds and d-lb < p.coefficient, i.e.
	  // we really can reduce a coefficient
	  Coefficient coefficient = lpbi.rhs-lb;
	  changed++;
	  term.add(coefficient,p.literal);	  
	  if (verbosity >= 5)
	    cout << "C";
	}
      else // nothing changed; just re-add again
	{
	  term.add(p.coefficient,p.literal);
	  // if lb is just 10% of what is needed
	  // assume that we can not do anything more for this ineq
	  if (lpbi.rhs-lb >= 10*p.coefficient)
	    {
	      Enum.EnumTablerestore();
	      break;
	    }
	}
      Enum.EnumTablerestore();
    }
  if (changed)
    {
      LPBIneq addlpbi;
      addlpbi.products = term;
      addlpbi.rhs = lpbi.rhs;
      addlpbi.normalize(); // gcd?
      del_lpbi(iid);
      if (!addlpbi.is_tautology())
	{
	  add(addlpbi);
	}
      // reinitialise enumeration
      switch (init_enumeration()) {
      case -1 : 
	cerr << "PBCS : coefficient reduction : Warning : unsatisfiability detected!\n";
	break;
      case 1 : 
	cerr << "PBCS : coefficient reduction : Warning : satisfiability detected!\n";
      }      
    }
  return changed;
}

int PBCS::coefficient_reduction()
{
  if (init_enumeration())
    return 0;
  int changed = 0;
  OrdInt alliids = S.alliids; // copy since coeffred may change things
  alliids.first();
  while(!alliids.last())
    {
      changed += coefficient_reduction(alliids.next(),2);
    }
  return changed;
}

