#include "PBCS.h"

int PBCS::addcover(int& clen)
{
  LPBIneq lpbi;
  if (GenerateMinimizedCoverIneq(lpbi) == 1) 
    return 1; // satisfiable; solution found
  no_covers++;
  if (is_unsatisfiable())
    return -1;
  clen = lpbi.products.length();
  if (lpbi.products.length() == 1)
    {
      if (verbosity >= 4)
	{ cout << "F"; cout.flush(); }
      fix(lpbi.products.min().literal);      
      fixing();
      enum_dirty = 1;
      if (is_unsatisfiable())
	return -1;
      else
	return 0;
    }
  else
    {      
      if (verbosity >= 4)
	{ cout << "[" << clen << "]"; cout.flush(); }
      delete_implied_by(lpbi); // delete implied ineqs by this cut
      // no add_prepro necessary; a cut!
      int iid = S.add(lpbi);
      Enum.add(iid,lpbi);
      enum_dirty = 1;
      return 0;
    }  
}

int PBCS::GenerateCover(OrdInt& cover)
{
  cover.clear();
  if (is_unsatisfiable())
    return -1;
  int res;
  if ((res = init_enumeration()))
    return res;
  res = Enum.GenerateCover(cover);  
  if (cover.empty() && (res != Enum.SAT))
    {
      set_unsat();
      return -1;
    }
  if (res == Enum.UNSAT) return -1;
  else if (res == Enum.SAT) return 1;
  else if (res == Enum.OPEN) 
    {
      cerr << "Enumerate::GenerateCover return OPEN\n"; 
      return 0;
    }
  return 0;
}

int PBCS::GenerateMinimizedCover(OrdInt& cover)
{
  int res;
  if ((res = GenerateCover(cover)) != -1)
    return res;
  res = Enum.MinimizeCover(cover,max_cut_nodes);
  if (cover.empty() && (res != Enum.SAT))
    {
      set_unsat();
      return -1;
    }
  if (res == Enum.UNSAT) return -1;
  else if (res == Enum.SAT) return 1;
  else if (res == Enum.OPEN) 
    {
      cerr << "Enumerate::MinimizeCover returned OPEN\n";
      return 0;
    }
  cerr << "Enumerate::MinimizeCover unknown return value \n";
  return 0;
}
 
int PBCS::GenerateMinimizedCoverIneq(LPBIneq& lpbi)
{
  OrdInt cover;
  if (GenerateMinimizedCover(cover) == 1)
    { // solution found
      lpbi.products.init(cover,1);
      return 1;
    }
  // 0 should never be returned by GenerateMinimizedCover!
  lpbi.products.init(cover,-1);
  lpbi.products.constant = 0;
  lpbi.rhs = 1;
//  lpbi.Print(cout,atoms);
  return -1;
}

OrdInt PBCS::get_last_solution()
{
  OrdInt oi = Enum.last_solution.to_OrdInt();
  return oi;
}
