#include "PBCS.h"
#include "Implications.h"

int PBCS::equate(int l1, int l2)
{
  enum_dirty = 1;
  Eq.add(l1,l2);
  int left = Eq(l1);
  int right = Eq(l2);
  if (left != l1)
    replace_lit(l1,left);
  if (right != l2)
    replace_lit(l2,right);
  return 0;
}

int PBCS::replace_lit(int left, int right)
{
  RArray<LitRef>& pos_ref = Enum.litref[left];   // positive replacing
  replace_lit_ref(pos_ref,left,right);
  RArray<LitRef>& neg_ref = Enum.litref[-left];  // negative replacing
  replace_lit_ref(neg_ref,-left,-right);
  return 0;
}

int PBCS::replace_lit_ref(RArray<LitRef> ref, int left, int right)
{
  for (int i=1; i <= ref[0].iid; i++)
    {
      LPBIneq& lpbi = S.Pool[ref[i].iid];
      lpbi.equate(right,left); // yes this is ok even if it seems unnatural
      if (lpbi.is_tautology())
	del_lpbi(ref[i].iid); // now it pays that Enum keeps a copy :-)
      else
	Enum.replace(ref[i].iid,lpbi);
    }
  return 0;
}

int PBCS::eq_check()
{
  Equations eq;
  Implications implications(128,&eq);
  OrdInt fixed_lits;
  int no_neweqs = 0;
  
  int res;
  if ((res = init_enumeration())) 
    return res;
    
  if (Enum.verbosity >= 6)
    { cout << "PBCS : eq_check \n";cout.flush(); }
  for(int i=1; i<=Enum.LitValue.upper(); i++)
    {
      if (Enum.litref.var_occurs(i)) // literal occurs in problem
	{	  
	  if (Enum.verbosity >= 6)
	    { cout << ".";cout.flush(); }
	  Enum.EnumTablesave();
	  if (Enum.fix_lit(i) != Enum.UNSAT)
	    {
	      fixed_lits = Enum.current_fixed();
	      fixed_lits.first();
	      while(!fixed_lits.last())
		implications.add(i,fixed_lits.next());
	      Enum.EnumTablerestore();
	      Enum.EnumTablesave();
	      if (Enum.fix_lit(-i) != Enum.UNSAT)
		{
		  fixed_lits = Enum.current_fixed();
		  fixed_lits.first();
		  while(!fixed_lits.last())
		    implications.add(-i,fixed_lits.next());
		}
	    }
	  Enum.EnumTablerestore();
	}
    }  
  implications.calc_equations();
  int l1,l2;
  while(!implications.equations_empty())
    {
      implications.pop_equation(l1,l2);
      equate(l1,l2);
      no_neweqs++;
    }
  return no_neweqs;
}
