#include "PBCS.h"
#include "Implications.h"
extern "C" {
#include "stdlib.h"
}

int PBCS::fix(int lit)
{
  if (is_unsatisfiable())
    return -1;
  if (Fixed.member(lit))
    return 0;
  if (Fixed.member(-lit))
    {
      set_unsat();
      return -1;
    }  
  Fixed += lit;
  enum_dirty = 1;
  RArray<LitRef>& pos_ref = Enum.litref[lit];   // positive fixing
  fix_ref(pos_ref,lit);
  RArray<LitRef>& neg_ref = Enum.litref[-lit];  // negative fixing
  fix_ref(neg_ref,lit);
  return 0;
}

int PBCS::fix_ref(RArray<LitRef> ref, int lit)
{
  for (int i=1; i <= ref[0].iid; i++)
    {
      LPBIneq& lpbi = S.Pool[ref[i].iid];
      lpbi.fix(lit);
      if (lpbi.is_tautology())
	del_lpbi(ref[i].iid); // now it pays that Enum keeps a copy :-)
      else if (lpbi.is_contradiction())
	{ // that is it; give up
	  set_unsat();
	}
      else
	Enum.replace(ref[i].iid,lpbi);
    }
  return 0;
}

int PBCS::fixed_literal()
{
  return S.fixed_literal();
}

OrdInt PBCS::fixing()
{
  OrdInt fixed_lits;
  int lit;
  while ((lit = fixed_literal()) != 0)
    {
      fix(lit);
      fixed_lits += lit;
    }
  return fixed_lits;
}

OrdInt PBCS::fixed_by(int literal, int& isunsat)
{
  OrdInt fixed_lits;
  // initialize enumeration
  if (init_enumeration()) 
    return fixed_lits;
  Enum.EnumTablesave();
  if (Enum.fix_lit(literal) == Enum.UNSAT)
    {
      isunsat = 1;
      set_unsat();
      return fixed_lits;
    }
  fixed_lits = Enum.current_fixed();
  Enum.EnumTablerestore();
  fixed_lits -= literal;
  return fixed_lits;
}

OrdInt PBCS::fixingcheck()
{
  OrdInt fixed_lits;

  if (init_enumeration()) 
    return fixed_lits;
  Enum.show_enum_table();
  if (Enum.verbosity >= 6)
    { cout << "Fixingcheck : ";cout.flush(); }
  OrdInt posfixed,negfixed;
  for(int i=1;i<=Enum.LitValue.upper();i++)
    {
      if (Enum.litref.var_occurs(i)) // literal occurs in problem
	{
	  if (Enum.verbosity >= 6)
	    { cout << ".";cout.flush(); }
	  Enum.EnumTablesave();
	  if (Enum.fix_lit(i) == Enum.UNSAT)
	    {
	      fixed_lits += -i;
	      if (Enum.verbosity >= 5)
		{ cout << "n"; cout.flush(); }
	    }
	  else
	    {
	      posfixed = Enum.current_fixed();
	      Enum.EnumTablerestore();
	      Enum.EnumTablesave();
	      if (Enum.fix_lit(-i) == Enum.UNSAT)
		{
		  fixed_lits += i;
		  if (Enum.verbosity >= 5)
		    { cout << "p"; cout.flush();}
		}
	      else
		{
		  negfixed = Enum.current_fixed();
		  negfixed.Intersection(posfixed);
		  if (!negfixed.empty()) // A->B and A->~B
		    fixed_lits += negfixed;
		}
	    }
	  Enum.EnumTablerestore();
	}
    }
  if (Enum.verbosity >= 6)
    cout << "\n";
  return fixed_lits;
}


