#include "PBTerm.h"
#include "RArray.h"
#include "PBToken.h"

PBTerm::PBTerm()
{
  what = CONSTANT;
  value = 0;
}

PBTerm::PBTerm(const PBTerm& t)
{
  what = CONSTANT;
  this->operator=(t);
}

PBTerm::PBTerm(int i)
{
  what = CONSTANT;
  value = i;
}

PBTerm::PBTerm(SimpleString& s,Atoms* atoms)
{
  what = LITERAL;
  value = (* atoms)(s);
}

PBTerm::PBTerm(char* s,Atoms *atoms)
{
  what = LITERAL;
  value = (* atoms)(s);
}

PBTerm::PBTerm(int OP,PBTerm& l, PBTerm& r)
{
  switch(OP) {
  case PLUS:
  case MINUS:
  case TIMES:
    what = OP;
    left = new PBTerm; (* left) = l;
    right = new PBTerm; (* right) = r;
    break;
  default:
    cerr << "PBTerm::PBTerm(OP,l,r) : OP not PLUS, MINUS, TIMES\n";
    exit(-1);
    break;
  }  
}

PBTerm& PBTerm::operator=(const PBTerm& t)
{
  if (this == &t)
    return *this; 
  int savewhat = what;
  PBTerm* saveleft = left;
  PBTerm* saveright = right;
  Copy(t);
  if ((savewhat != LITERAL) && (savewhat != CONSTANT))
    {
      saveleft->RecDelete();
      saveright->RecDelete();
      delete saveleft;
      delete saveright;
    }
  return *this;
}

 
PBTerm::~PBTerm()
{
//  cout << "DESTRUCTOR ";
  RecDelete();
}

void PBTerm::Print(ostream& out, Atoms* atoms)
{
  if (what == LITERAL) 
    {
      if (atoms)
	{
	  if (value > 0)
	    out << (*atoms)(value);
	  else
	    out << "~" << (*atoms)(-value);
	}
      else
	out << "X" << value;
    }
  else if (what == CONSTANT)
    {
      out << value;
    }
  else
    {
      out << "(";
      left->Print(out,atoms);
      if (what == PLUS)
	cout << "+";
      else if (what == MINUS)
	cout << "-";
      else // if (what == TIMES)
	cout << "*";
      right->Print(out,atoms);
      out << ")";
    }
}

#include "PBTermParse.C"

int PBTerm::initialize(char* s, Atoms* atoms)
{
  SimpleString ss(s);
  return initialize(ss,atoms);
}

// basic simplifications
//  eliminates MINUS
//  makes associate operator chains left associate
//  and simplifies values 
int PBTerm::simplify()
{
  elim_minus();
  group_ass();
  simpl_consts();
  return 0;
}

int PBTerm::elim_minus()
{
  switch (what) {
  case PLUS:
  case TIMES:
    left->elim_minus();
    right->elim_minus();
    break;
  case MINUS:
    {
      what = PLUS;
      PBTerm* minus1 = new PBTerm;
      minus1->what = CONSTANT;
      minus1->value = -1;
      PBTerm* newtimes = new PBTerm;
      newtimes->what = TIMES;
      newtimes->left = minus1;
      newtimes->right = right;
      right = newtimes;
    }
      break;
  default:
    break;
  }
  return 0;
}

int PBTerm::group_ass()
{
  switch (what) {
  case PLUS:
  case TIMES:
    group_ass(what);
    break;
  default:
    break;
  }
  return 0;
}
 
int PBTerm::group_ass(int type)
{
  // both are the same
  if ((left->what == type) && (right->what == type))
    {
      PBTerm* h = right;
      right = left;
      left = right->right;
      right->right = h;
      group_ass(type); // funny recursion :-)
    }
  else if (left->what == type) // only left is the same -> swap
    {
      PBTerm* h = right;
      right = left;
      left = h;
    }
  // else group in subterms
  left->group_ass();
  right->group_ass();
  return 0;
}

int PBTerm::simpl_consts()
{
  switch (what) {
  case PLUS:
  case TIMES:
    {
      PBTerm* p = this;
      PBTerm* last = this;
      while(p->what == what)
	{
//	  cout << "    last: ";last->Print(cout);cout << "\n";
//	  cout << "    p:    ";p->Print(cout);cout << "\n";
	  p->left->simpl_consts();
	  if (p->left->what == CONSTANT)
	    {
	      if (p != this)
		{
		  if (left->what != CONSTANT)
		    {
		      // swap
		      PBTerm* h = this->left;
		      left = p->left;
		      p->left = h;
		      last = p;
		      p = p->right;
		    }
		  else // merge constant with existing constant and delete entry
		    {
		      if (what == PLUS)
			left->value += p->left->value;
		      else // if (what == TIMES)
			left->value *= p->left->value;
		      delete p->left;
		      last->right = p->right;
		      p->what = CONSTANT;
		      delete p;
		      p = last->right;
		    }
		}
	      else
		{
		  last = p;
		  p = p->right;
		}
	    }
	  else 
	    {
	      last = p;
	      p = p->right;
	    }
	}
      // check for right entry
      p->simpl_consts();
//      cout << "    last: ";last->Print(cout);cout << "\n";
//      cout << "    p:    ";p->Print(cout);cout << "\n";
      if (p->what == CONSTANT)
	{
	  if (left->what != CONSTANT)
	    {
	      // swap
	      PBTerm* h = left;
	      left = p;
	      last->right = h;
	    }
	  else 
	    {	      
	      if (last == this)
		{
		  value = left->value;
		  if (what == PLUS)
		    value += p->value;
		  else
		    value *= p->value;
		  what = CONSTANT;
		  delete left;
		  delete p;
		}
	      else
		{
		  if (what == PLUS)
		    left->value += p->value;
		  else
		    left->value *= p->value;
		  delete p;
		  PBTerm* h = last->left;
		  last->what = last->left->what;
		  last->value = last->left->value;
		  last->right = last->left->right;
		  last->left = last->left->left;		  
		  delete h;
		}
	    }
	}
    }
  }
  return 0;
}


int PBTerm::polform()
{
//  cout << "PF:";Print(cout);cout << "\n";cout.flush();
  switch (what) {
  case MINUS:
    left->polform();
    right->polform();
    right->mult_polform_const(-1);
    what = PLUS;
    merge_plus();
    break;
  case PLUS:
    left->polform();
    right->polform();
    merge_plus();
    break;
  case TIMES:
    left->polform();
    right->polform();
    distrib_times();
  case LITERAL:
  case CONSTANT:
    break;
  default:
    break;
  }
  return 0;
}

int PBTerm::identical_literals(PBTerm* l,PBTerm* r)
{
  return (subset_literals(l,r) && subset_literals(r,l));
}

int PBTerm::subset_literals(PBTerm* l,PBTerm* r)
{
  RAssocItPBTerm itl(l,TIMES);
  PBTerm* pl;

  while ((pl = itl.get_current()))
    {
      if (pl->what != LITERAL)
	return 0; // no literal
      RAssocItPBTerm itr(r,TIMES);
      PBTerm* pr;
      do {
	pr = itr.get_current();
	itr.next();
	if (pr == 0) 
	  return 0; // literal not found
	if (pr->what != LITERAL)
	  return 0; // no literal
      } while (pr->value != pl->value);
      itl.next();
    }
  return 1; // ok; subset!
}

int PBTerm::ensure_coefficient(PBTerm* t)
{
  int add1 = 0;
  if (t->what == LITERAL)
    add1 = 1;
  else if (t->what == CONSTANT)
    add1 = 0;
  else if (t->what == TIMES)
    {
      if (t->left->what == CONSTANT)
	add1 = 0;
      else if (t->left->what == LITERAL)
	add1 = 1;
    }
  if (!add1)
    return 0;
  PBTerm* copyt = new PBTerm;
  if ((t->what == CONSTANT) || (t->what == LITERAL))
    {
      copyt->what = t->what;
      copyt->value = t->value;
    }
  else
    {
      copyt->what = t->what;
      copyt->left = t->left;
      copyt->right = t->right;
    }
  t->what = TIMES;
  PBTerm* one = new PBTerm;
  one->what = CONSTANT;
  one->value = 1;
  t->left = one;
  t->right = copyt;
  return 1;
}


int PBTerm::merge_plus()
{
//  cout << "MERGEPLUS : ";Print(cout);cout << "\n";cout.flush();
  merge(PLUS);
  // ensures coeffcients on all products
  RAssocItPBTerm ite(this,PLUS);
  PBTerm* p1;  
  while ((p1 = ite.get_current()))
    {
      ensure_coefficient(p1);
      ite.next();
    }
  // group identicals
  RAssocItPBTerm it1(this,PLUS);
  while ((p1 = it1.get_current()))
    {
      PBTerm* f = it1.get_current_father();
      if ((f->what != PLUS) || (f->left != p1))
	break; // last element in plus chain
      RAssocItPBTerm it2(it1.get_current_father()->right,PLUS);
      PBTerm* p2;
      while ((p2 = it2.get_current()))
	{
	  if ((p1->what == CONSTANT) || (p2->what == CONSTANT))
	    {
	      if ((p1->what == CONSTANT) && (p2->what == CONSTANT))
		{
		  p1->value += p2->value;
		  it2.del_current_next();
		}
	      else
		it2.next();
	    }
	  else
	    {
	      if (identical_literals(p1->right,p2->right))
		{
		  p1->left->value += p2->left->value;
		  it2.del_current_next();
		}
	      else
		it2.next();
	    }
	}
      it1.next();
    }
  // delete zeros
  RAssocItPBTerm it(this,PLUS);
  PBTerm* p;
  while ((p = it.get_current()))
    {
      if ((p->what == CONSTANT) && (p->value == 0))
	it.del_current_next();
      else
	it.next();
    }
  return 0;
}

int PBTerm::merge(int op)
{
//  cout << "Merge:";Print(cout);cout << "\n";cout.flush();
  if (left->what != op)
    return 0;
  PBTerm* l = left;
  left = l->left;
  PBTerm* r = this;
  while (r->right->what == op)
    r = r->right;
  l->left = r->right;
  r->right = l;
  return 0;
}

int PBTerm::mult_polform_const(int c)
{
//  cout << "MultPol:";Print(cout);cout << "\n";cout.flush();
  RAssocItPBTerm ite(this,PLUS);
  PBTerm* p1;  
  while ((p1 = ite.get_current()))
    {
      ensure_coefficient(p1);
      if (p1->what == CONSTANT)
	p1->value *= c;
      else
	p1->left->value *= c;
      ite.next();
    }
//  cout << "MultPolDone:";Print(cout);cout << "\n";cout.flush();
  return 0;
}

int PBTerm::del_consts_times()
{
  if (what == CONSTANT)
    return 0;
  ensure_coefficient(this);
  RAssocItPBTerm ite(this,TIMES);
  PBTerm* p1;  
  ite.next(); // the first is the coefficient
  while ((p1 = ite.get_current()))
    {
      if (p1->what == CONSTANT)
	{
	  left->value *= p1->value;
	  ite.del_current_next();
	}
      else
	ite.next();
    }  
  if (what == CONSTANT)
    return 0;
  if (left->value == 0)
    {
      RecDelete();
      what = CONSTANT;
      value = 0;
    }
  return 0;
}

int PBTerm::del_doublelits_times()
{
  PBTerm* p1 = this;
  while (p1->what == TIMES)  // last literal cannot have double occurence :-)
    {
      if (p1->left->what == LITERAL)
	{
	  // delete all further occurences of that literal
	  PBTerm* p2 = p1->right;
	  PBTerm* father = p1;
	  PBTerm* h;
	  int last = 0;
	  while ((p2->what == TIMES) || (last == 0))
	    {
	      h = p2->left;
	      if (p2->what != TIMES)
		{
		  h = p2;
		  last = 1;
		}
	      if ((h->what == LITERAL) && (h->value == p1->left->value)) // delete identical
		{		  
		  if (last == 0) // left son
		    {
		      delete h;
		      father->right = p2->right;
		      p2->what = CONSTANT;
		      delete p2;
		      p2 = father->right;
		    }
		  else // last right son
		    {
		      delete father->right;
		      PBTerm* todel = father->left;
		      father->what = todel->what;
		      father->value = todel->value;
		      todel->what = CONSTANT;
		      delete todel;	      
		    }
		}
	      else if ((h->what == LITERAL) && (h->value == -(p1->left->value))) // A*~A -> del whole product
		{
		  RecDelete();
		  what = CONSTANT;
		  value = 0;
		  return 0;
		}
	      else if (last == 0)
		{
		  father = p2;
		  p2 = p2->right;
		}
	      if (last == 1)
		break;
	    }
	}
      if (p1->what == TIMES)
	p1 = p1->right;
    }
  return 0;
}

PBTerm* PBTerm::merge_mults(PBTerm* l, PBTerm* r)
{ 
//  cout << "MERGEMULT:";Print(cout);cout << "\n";cout.flush();
  PBTerm* ret = new PBTerm;
  ret->what = TIMES;
  ret->left = new PBTerm;
  ret->left->Copy(*l);
  ret->right = new PBTerm;
  ret->right->Copy(*r);
  // normalize term
  ret->merge(TIMES); // flatten it
  ret->del_consts_times(); // delete constants and merge them to the beginning
  ret->del_doublelits_times(); 
  return ret;
}

int PBTerm::distrib_times()
{
//  cout << "DISTRIBTIMES:";Print(cout);cout << "\n";cout.flush();
  if ((left->what != PLUS) && (right->what != PLUS))
    {
      merge(TIMES);
      del_consts_times();
      del_doublelits_times();
      return 0;
    }

  PBTerm* top = new PBTerm;
  PBTerm* free = top;
  PBTerm* beforefree = top;

  PBTerm* l = left;
  int llast = 0;
  while ((l->what == PLUS) || (llast == 0)) // for all elements
    {
      PBTerm* lt = l->left;
      if (l->what != PLUS)
	{
	  llast = 1;
	  lt = l;
	}	
      PBTerm* r = right;
      int rlast = 0;
      while ((r->what == PLUS) || (rlast == 0)) // for all elements
	{
	  PBTerm* rt = r->left;
	  if (r->what != PLUS)
	    {
	      rlast = 1;
	      rt = r;
	    }
	  free->what = PLUS;
	  free->left = merge_mults(lt,rt);
	  free->right = new PBTerm;
	  beforefree = free; // is called at least once!
	  free = free->right;
	  if (rlast == 0)
	    r = r->right;
	}
      if (llast == 0)
	l = l->right;
    }
  delete beforefree->right;
  free = beforefree->left;
  beforefree->what = free->what;
  beforefree->left = free->left;
  beforefree->right = free->right;
  beforefree->value = free->value;
  free->what = CONSTANT;
  delete free;
  RecDelete();
  left = top->left;
  right = top->right;
  what = PLUS;
  top->what = CONSTANT;
  delete top;
  merge_plus();
  return 0;
}

int PBTerm::Read(istream& in, Atoms* atoms)
{
  SimpleString s;
  s.Read(in,"\n");
  if (atoms == 0)
    {
      cerr << "PBTerm::Read atoms must be != 0\n";
      exit(-1);
    }
  return 0;
}

void PBTerm::Copy(const PBTerm& t)
{
  if ((t.what == LITERAL) || (t.what == CONSTANT))
    {
      what = t.what;
      value = t.value;
    }
  else
    {
      what = t.what;
      left = new PBTerm;
      left->Copy(*t.left);
      right = new PBTerm;      
      right->Copy(*t.right);
    }
}

void PBTerm::RecDelete()
{
  if ((what != LITERAL) && (what != CONSTANT))
    {
      delete left;
      delete right;
    }
}

