#ifndef _PBTerm_h
#define _PBTerm_h

// general pseudo-Boolean terms

#include "SimpleString.h"
#include "Atoms.h"

class PBTerm {
public:
  PBTerm();
  PBTerm(const PBTerm&);
  PBTerm(int i);
  PBTerm(SimpleString&, Atoms*);
  PBTerm(char*, Atoms*);
  PBTerm(int OP,PBTerm&, PBTerm&);
  PBTerm& operator=(const PBTerm&);
  ~PBTerm();

  enum { PLUS, MINUS, TIMES, LITERAL, CONSTANT };  

  PBTerm& operator[](int i);
  int kind();

  int initialize(SimpleString&, Atoms* );
  int initialize(char* s, Atoms* );
  int simplify(); // very basic simplifications
  int polform(); // polynomial form

  int Read(istream& in=cin, Atoms* atoms = 0);
  void Print(ostream& out=cout, Atoms* atoms = 0);

private:
  int what;
  // make a union from value or left and right?
  int value;
  PBTerm* left;
  PBTerm* right;

  void Copy(const PBTerm& t);
  void RecDelete();

  int parse(int, int);

  // for simplify
  int elim_minus();
  int group_ass();
  int group_ass(int type);
  int simpl_consts();

  // for pol_form
  int distrib_times();
  int mult_polform_const(int );  
  int merge(int );
  int subset_literals(PBTerm* ,PBTerm* );
  int identical_literals(PBTerm* ,PBTerm* );
  int ensure_coefficient(PBTerm* );
  int merge_plus();
  int del_consts_times();
  int del_doublelits_times();
  PBTerm* merge_mults(PBTerm*, PBTerm* );

  // memory allocation
  void* operator new(size_t)
  {
    PBTerm* p = (PBTerm *) new char[sizeof(PBTerm)];
//    cout << "allocating " << p << "\n";
    return p;
  }
  void operator delete(void *p)
  {
//    cout << "freeing " << p << "\n";
    delete [] p;
  }

  friend class PolPBTerm;
  friend class RAssocItPBTerm;
};

inline PBTerm& PBTerm::operator[](int i)
{
  if ((i < 0) || (i > 1))
    {
      cerr << "PBTerm::operator[] : index not 0 or 1\n";
      exit(-1);
    }
  if ((what == CONSTANT) || (what == LITERAL))
    {
      cerr << "PBTerm::operator[] : no subterm\n";
      exit(-1);      
    }
  return i == 0 ? *left : *right;
}

inline int PBTerm::kind() { return what; }


// an iterator class for rightnormalized PBTerms
class RAssocItPBTerm
{
private:
  int op;
  PBTerm* root;
  PBTerm* father;
  PBTerm* entry;
public:
  RAssocItPBTerm(PBTerm* t, int o) 
  {
    root = father = t;
    op = o; 
    if (t==0)
      {
	entry = 0;
	return;
      }
    else if (t->what == o)
      {
	father = t;
	entry = t->left;
      }
    else 
      {
	father = entry = t;
      }
  }
  
  PBTerm* get_current() 
  {
    return entry;
  }

  PBTerm* get_current_father() 
  {
    return father;
  }

  void next()
  {
    if (father == 0)
      return;
    if ((father->what == op) && (father->left == entry))
      {
	if (father->right->what == op)
	  {
	    father = father->right;
	    entry = father->left;
	  }
	else
	  {
	    entry = father->right;
	  }
      }
    else
      {
	father = 0;
	entry = 0;
      }
  }

  int last() { return father == 0; }

  void del_current_next()
  {
    if ((entry == 0) || (father ==0))
      return;
    if (father == entry)
      {
	// default stuff
	entry->what = PBTerm::CONSTANT;
	if (op == PBTerm::PLUS)
	  entry->value = 0;
	else // TIMES
	  entry->value = 1;
	father = entry = 0;
      }
    else // father->left == entry || father->right == entry
      {
	PBTerm* todel;
	int ewasleft;
	if (father->left == entry)
	  {
	    todel = father->right;
	    ewasleft = 1;
	  }
	else // father->right == entry
	  {
	    todel = father->left;
	    ewasleft = 0;
	  }
	entry->RecDelete();
	entry->what = PBTerm::CONSTANT;
	delete entry;
	father->what = todel->what;
	if ((father->what == PBTerm::CONSTANT) || (father->what == PBTerm::LITERAL))
	  {
	    father->value = todel->value;
	  }
	else 
	  {
	    father->left = todel->left;
	    father->right = todel->right;
	  }
	todel->what = PBTerm::CONSTANT;
	delete todel;
	if (ewasleft)
	  {
	    if (father->what == op)
	      entry = father->left;
	    else 
	      {
		if (father == root)
		  entry = father;
		else
		  {
		    PBTerm* h = root;
		    while (h->right != father)
		      h = h->right;
		    father = h;
		    entry = father->right;
		  }
	      }
	  }
	else
	  entry = father = 0;
      }
  }

};




#endif
