// parse a string and generate a PBTerm
// Attention! - is rightassoc (probably I should change this)

static RArray<PBToken> pbtokens;
static int pbtokensi;

Atoms* pbt_ga;

int show_tokens(void );

int add_token(int what, int pos)
{
  pbtokens.ensure(pbtokensi);
  pbtokens[pbtokensi].what = what;
  pbtokens[pbtokensi].pos = pos;
  pbtokensi++;
  return 0;
}

int add_constant(SimpleString& s, int& i, int len)
{
  pbtokens.ensure(pbtokensi);
  pbtokens[pbtokensi].what = PBToken::CONSTANT;
  pbtokens[pbtokensi].pos = i;
  int value = 0;
  while ((i < len) && (s[i] >= '0') && (s[i] <= '9'))
    {
      value = 10*value + (s[i] - '0');
      i++;
    }
  pbtokens[pbtokensi].value = value;
  pbtokensi++;  
  return 0;
}

int add_literal(SimpleString& s, int& i, int len)
{
  pbtokens.ensure(pbtokensi);
  pbtokens[pbtokensi].what = PBToken::LITERAL;
  pbtokens[pbtokensi].pos = i;
  int neg = 0;
  if (s[i] == '~')
    {
      neg = 1;
      i++;
      while( (i < len) && ((s[i] == ' ') || (s[i] == '\t') || (s[i] == '\n')))
	i++;
      if (i >= len)
	{
	  cerr << "tokenise : ~ without literal at position " << i << "\n";
	  return 1;
	}
      if (! (((s[i] <= 'z') && (s[i] >= 'a')) ||
	     ((s[i] <= 'Z') && (s[i] >= 'A')) ||
	     (s[i] == '_')))
	{
	  cerr << "tokenise : ~ without literal at position " << i << "\n";
	  return 1;
	}
    }
  int j = i;
  int stop = 0;
  do {
    if (((s[i] <= 'z') && (s[i] >= 'a')) ||
        ((s[i] <= 'Z') && (s[i] >= 'A')) ||
        (s[i] == '_') ||
        ((s[i] <= '9') && (s[i] >= '0')) )
      {
	i++;
      }
    else if ((s[i] == ' ') || (s[i] == '\t') || (s[i] == '\n') ||
	     (s[i] == '(') || (s[i] == ')') || 
	     (s[i] == '+') || (s[i] == '-') || (s[i] == '*'))
      {
	stop = 1;
      }
    else
      {
	cerr << "tokenise : unexpected char " << s[i] << " on pos " << i << "\n";
	return 1;
      }
  } while (!stop && (i < len));
  if (j >= i)
    {
      cerr << "tokenise : no literal at position " << i << "\n";
      return 1;
    }
  SimpleString var = s(j,i-1);
  int value = pbt_ga->operator()(var);
  pbtokens[pbtokensi].value = neg ? -value : value;
  pbtokensi++;  
  return 0;
}

int tokenise(SimpleString& s)
{
  // gets a token, checks what it is and pushes it on the stack
  // Literals are inserted in Atoms first
  pbtokensi = 0;
  int i = 0;
  int len = s.length();
  int skip;
  while(i < len)
    {
      // skip seperating chars
      skip = 1;
      do {
	switch (s[i]) {
	case ' ' :
	case '\t':
	case '\n':
	  i++;	  
	  skip = 1;
	  break;
	default:
	  skip = 0;
	} 
      } while (skip && (i < len));
      if (i >= len)
	break; // parse done
      // get token
      switch (s[i]) {
      case '(':
	add_token(PBToken::LPAR,i);i++;break;
      case ')':
	add_token(PBToken::RPAR,i);i++;break;
      case '+':
	add_token(PBToken::PLUS,i);i++;break;
      case '-':
	add_token(PBToken::MINUS,i);i++;break;
      case '*':
	add_token(PBToken::TIMES,i);i++;break;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
	if (add_constant(s,i,len))
	  return -1;
	break;
      default:
	if (add_literal(s,i,len))
	  return -1;
	break;
      }
    }
  return 0;
}

// for parsing
int matching_rpar(int i, int level)
{
  while (i < pbtokensi)
    {
      if (pbtokens[i].what == PBToken::RPAR)
	{
	  if (--level == 0)
	    return i;      
	}
      else if (pbtokens[i].what == PBToken::LPAR)
	level++;
      i++;
    }
  return -1;
}

int get_principal_binop(int from, int to)
{
  int level = 0;
  int save_mult = -1;

  while (from <= to)
    {
      if ((pbtokens[to].what == PBToken::PLUS) || (pbtokens[to].what == PBToken::MINUS))
	{
	  if ((level == 0) && (to > from))
	    {
	      if ((pbtokens[to-1].what == PBToken::PLUS) || 
		  (pbtokens[to-1].what == PBToken::MINUS))
		return to-1;
	      else
		return to;
	    }
	}
      else if (pbtokens[to].what == PBToken::TIMES)
	{
	  if ((save_mult == -1) && (level == 0))
	    save_mult = to;
	}
      else if (pbtokens[to].what == PBToken::RPAR)
	{
	  level++;
	}
      else if (pbtokens[to].what == PBToken::LPAR)
	{
	  if (--level == -1)
	    return -1;
	}
      else
	{
	  // nothing
	}
      to--;
    }
  return save_mult;
}

// works by searching top operator and splitting
int PBTerm::parse(int from, int to)
{
  if (from > to)
    {
      cerr << "parse : no subexpression " << "at pos " << pbtokens[from].pos << "\n";
      return -1;
    }
  if (pbtokens[from].what == PBToken::RPAR)
    {
      cerr << "parse : unexpected right parantheses";
      cerr << " at pos " << pbtokens[from].pos << "\n";
      return -1;
    }
  if (pbtokens[from].what == PBToken::LPAR)
    {
      // delete matching RPAR if these are outer ones
      if (matching_rpar(from+1,1) == to)
	return parse(from+1,to-1);
    }
  if (from == to)
    {
      if (pbtokens[from].what == PBToken::CONSTANT)
	{
	  what = CONSTANT;
	  value = pbtokens[from].value;
	  return 0;
	}
      else if (pbtokens[from].what == PBToken::LITERAL)
	{
	  what = LITERAL;
	  value = pbtokens[from].value;
	  return 0;
	}
      else
	{
	  cerr << "parse error : literal or constant expected";
	  cerr << " at pos " << pbtokens[from].pos << "\n";
	  return -1;
	}
    }
  else if (from+1 == to) // two entries to parse
    {
      what = TIMES;
      left = new PBTerm;
      right = new PBTerm;
      if (pbtokens[from].what == PBToken::PLUS)
	{
	  left->what = CONSTANT;
	  left->value = 1;
	  return right->parse(from+1,to);
	}
      else if (pbtokens[from].what == PBToken::MINUS)
	{
	  left->what = CONSTANT;
	  left->value = -1;
	  return right->parse(from+1,to);
	}
      else // an abbreviated *
	{
	  int ret = left->parse(from,from);
	  if (ret)
	    return ret;
	  return right->parse(to,to);
	}
    }
  else // more than two entries to parse
    {
      int i = get_principal_binop(from,to);
      if (i == -1) // no prinicipal binop hence * abbrev
	{
	  what = TIMES;
	  left = new PBTerm;
	  right = new PBTerm;
	  int ret,f0,t0,f1,t1;	  
	  if (pbtokens[from].what == PBToken::LPAR)
	    {	      
	      int rpar = matching_rpar(from+1,1);
	      if (rpar == -1)
		{
		  cerr << "parse : parentheses mismatch\n";
		  return -1;
		}
	      f0 = from+1;
	      t0 = rpar-1;
	      f1 = rpar+1;
	      t1 = to;
	    }
	  else if ((pbtokens[from].what == PBToken::PLUS) ||
		   (pbtokens[from].what == PBToken::MINUS))
	    {
	      f0 = from;
	      t0 = from+1;
	      f1 = from+2;
	      t1 = to;
	    }
	  else
	    {
	      f0 = from;
	      t0 = from;
	      f1 = from+1;
	      t1 = to;
	    }
	  if ((ret = left->parse(f0,t0)))
	    return ret;
	  return right->parse(f1,t1);	    
	}
      else
	{
	  if (pbtokens[i].what == PBToken::PLUS)
	    what = PLUS;
	  else if (pbtokens[i].what == PBToken::MINUS)
	    what = MINUS;
	  else
	    what = TIMES;
	  left = new PBTerm;
	  right = new PBTerm;
	  int ret;
	  if ((ret = left->parse(from,i-1)))
	    return ret;
	  return right->parse(i+1,to);
	}
    }
//  return 0;
}

int PBTerm::initialize(SimpleString& s,Atoms* atoms)
{
  if (atoms == 0)
    {
      cerr << "PBTerm::initialize Atoms must be != 0\n";
      cerr << "setting to empty term\n";
      return 1;
    }
  pbt_ga = atoms;
  RecDelete();
  what = CONSTANT;
  value = 0;
  int ret;
  if ((ret = tokenise(s)))
    return ret;
  if (pbtokensi == 0) // nothing to read
    return 0;
//  show_tokens();  
  return parse(0,pbtokensi-1);
}
 


// verbose functions
int show_tokens(void )
{
  for(int i=0;i < pbtokensi;i++)
    {
      cout << i << " : ";
      switch (pbtokens[i].what) {
      case PBToken::CONSTANT:
	cout << pbtokens[i].value;break;
      case PBToken::LITERAL:
	if (pbtokens[i].value > 0)
	  cout << pbt_ga->operator()(pbtokens[i].value);
	else
	  cout << "~" << pbt_ga->operator()(-pbtokens[i].value);	
	break;
      case PBToken::PLUS: cout << "+";break;
      case PBToken::MINUS: cout << "-";break;
      case PBToken::TIMES: cout << "*";break;
      case PBToken::LPAR: cout << "(";break;
      case PBToken::RPAR: cout << ")";break;
      default:
	cout << "unknown";
      }
      cout << "\n";
    }
  return 0;
}

#ifdef __GNUC__
template class RArray<PBToken>; 
#endif
