#include "PolPBTerm.h"

PolPBTerm::PolPBTerm() : products(0), constant(0)
{
}

PolPBTerm::PolPBTerm(PolPBTerm& t) : products(0), constant(0)
{
  operator=(t);
}

PolPBTerm::PolPBTerm(PBTerm& t) : products(0), constant(0)
{
  operator=(t);
}

PolPBTerm& PolPBTerm::operator=(PBTerm& tt)
{
  tt.polform();
  if (tt.kind() == PBTerm::CONSTANT)
    {
      products = 0;
      constant = tt.value;
      return *this;
    }      
  constant = 0;
  products = new PolPBTermNode;
  PolPBTermNode* h = products;
  PBTerm* t = &tt;
  int last = 0;
  PolPBTermNode* lasth = h;
  while ((t->what == PBTerm::PLUS) || (last == 0))
    {      
      PBTerm* tc = t->left;
      if (t->what != PBTerm::PLUS)
	{
	  tc = t;
	  last = 1;
	}
//      cout << "adding"; tc->Print(cout);cout << "\n";cout.flush();
      if (tc->what == PBTerm::CONSTANT)
	{
	  constant += tc->value;
	  if (last == 1)
	    {
	      delete h;
	      lasth->next = 0;
	    }
	}
      else if (tc->what == PBTerm::LITERAL)
	{
	  h->coefficient = 1;
	  h->literals += tc->value;
	  if (last == 0)
	    {
	      h->next = new PolPBTermNode;
	      lasth = h;
	      h = h->next;
	    }
	}
      else if (tc->what == PBTerm::TIMES)
	{
	  if (tc->left->what == PBTerm::CONSTANT)
	    {
	      h->coefficient = tc->left->value;
	      tc = tc->right;
	    }
	  else
	    {
	      h->coefficient = 1;	      
	    }
	  // copy literals
	  while (tc->what == PBTerm::TIMES) 
	    {
	      h->literals += tc->left->value;
	      tc = tc->right;
	    }
	  h->literals += tc->value;
	  if (last == 0)
	    {
	      lasth = h;
	      h->next = new PolPBTermNode;
	      h = h->next;
	    }
	}
      if (last == 1)
	{
	  h->next = 0;
	  break;
	}
      else
	{
	  t = t->right;
	}
    }   
  return *this;
}

PolPBTerm& PolPBTerm::operator=(PolPBTerm& t)
{
  constant = t.constant;  
  if (t.products == 0)
    {
      products = 0;
      return *this;
    }
  PolPBTermNode* p = t.products;  
  products = new PolPBTermNode;
  PolPBTermNode* h = products;
  while(p)
    {
      h->literals = p->literals;
      h->coefficient = p->coefficient;      
      p = p->next;
      if (p)
	{
	  h->next = new PolPBTermNode;
	  h = h->next;
	}
      else
	h->next = 0;
    }  
  return *this;
}

PolPBTerm::~PolPBTerm()
{
  if (products == 0)
    return;
  PolPBTermNode* h = products;
  while (products->next)
    {
      h = products;
      products = products->next;
      delete h;
    }
  delete products;
}

int PolPBTerm::is_linear()
{
  PolPBTermNode* h = products;
  while(h)
    {
      if (h->literals.length() > 1) // not very efficient!
	return 0;
      h = h->next;
    }
  return 1;
}

int PolPBTerm::no_products()
{
  PolPBTermNode* h = products;
  int no = 0;
  while(h)
    {
      no++;
      h = h->next;
    }
  return no;
}


int PolPBTerm::mul_constant(Coefficient c)
{
  if (c == 1)
    return 0;
  PolPBTermNode* h = products;
  while(h)
    {
      h->coefficient *= c;
      h = h->next;
    }
  constant *= c;
  return 0;
}

int PolPBTerm::add(Coefficient coefficient, OrdInt& literals)
{
  if (coefficient == 0)
    return 0;
  if (literals.empty())
    {
      constant += coefficient;
      return 0;
    }
  // nonempty literals
  PolPBTermNode* h = products;
  while(h)
    {
      if (literals.compare(h->literals) == 0) // same literals set
	{
	  h->coefficient += coefficient;
	  if (h->coefficient == 0)
	    return del(literals);
	  return 0;
	}
      h = h->next;
    }
  // no same literals found
  h = products;
  products = new PolPBTermNode;
  if (products == 0)
    {
      cerr << "PolPBTerm : add : no more memory \n";
      exit(-1);
    }
  products->next = h;
  products->coefficient = coefficient;
  products->literals = literals;
  return 0;
}

int PolPBTerm::del(OrdInt& literals)
{
  if (products == 0)
    return 0;
  if (products->next == 0)
    {
      if (literals.compare(products->literals) == 0)
	{
	  delete products;
	  products = 0;
	  return 0;
	}
    }
  PolPBTermNode* h = products;

  while(h->next)
    {
      if (literals.compare(h->next->literals) == 0)
	{
	  PolPBTermNode* hh = h->next;
	  h->next = hh->next;
	  delete hh;
	  return 0;
	}
      h = h->next;
    }
  return 0;
}

int PolPBTerm::Print(ostream& out, Atoms* atoms)
{
  PolPBTermNode* h = products;

  while(h)
    {
      cout << h->coefficient << "*";
      h->literals.first();
      while(!h->literals.last())
	{
	  int lit = h->literals.next();
	  if (atoms == 0)
	    out << lit;
	  else
	    {
	      if (lit > 0)
		out << (* atoms)(lit);
	      else
		out << "~" << (* atoms)(-lit);
	    }	  
	  if (!h->literals.last())
	    out << "*";
	}
      h = h->next;
      if (h || (constant != 0))
	out << " + ";
    }
  if (constant !=0)
    out << constant; 
  return 0;
}

int PolPBTerm::initialize(SimpleString& s, Atoms* atoms)
{
  PBTerm t;
  int ret;
  if ((ret = t.initialize(s,atoms)))
    return ret;
  this->operator=(t);
  return 0;
}
