#include "Products.h"
#include "stdlib.h"

Products::Products()
{
  head = new ProductNode;
  head->next = NULL;
  head->product.coefficient = 0;
  head->product.literal = 0;
  constant = 0;
  sumcoeffs = 0;
}

Products::~Products()
{
#ifdef OWN_MEMORY_ProductNode
  // insert the whole chain at once at the beginning of the freelist
  ProductNode* p = head;
  
  while (p->next) 
    p = p->next;
  p->next = ProductNode::pfree;
  ProductNode::pfree = head;
#else
  ProductNode* p = head->next;
  ProductNode* pnext;

  while(p)
    {
      pnext = p->next;
      delete p;
      p = pnext;
    }
  delete head;
#endif
}

inline void Products::free_list(ProductNode* p)
{
#ifdef OWN_MEMORY_ProductNode
  // insert the whole chain at once at the beginning of the freelist
  if (p == NULL)
    return;
  ProductNode* first = p;
  
  while (p->next) 
    p = p->next;
  p->next = ProductNode::pfree;
  ProductNode::pfree = first;
#else
  ProductNode* pnext;

  while(p)
    {
      pnext = p->next;
      delete p;
      p = pnext;
    }
#endif  
}

Products::Products(const Products& p) : constant(p.constant), sumcoeffs(p.sumcoeffs)
{
#ifdef OWN_MEMORY_ProductNode
  ProductNode* tocopy = p.head;

  head = new ProductNode;
  head->next = ProductNode::pfree;
  head->product.coefficient = 0;
  head->product.literal = 0;
  current = head;
  while (tocopy->next)
    {
      if (ProductNode::pfree) // fine, we have some left
	{
	  ProductNode::pfree->product = tocopy->next->product;
	  tocopy = tocopy->next;
          current->next = ProductNode::pfree;  // neccesary
	  current = ProductNode::pfree;
	  ProductNode::pfree = ProductNode::pfree->next;
	}
      else // standard allocation
        {
	  ProductNode* newnode = new ProductNode;
	  
	  newnode->product = tocopy->next->product;
	  current->next = newnode;
	  current = newnode;
	  tocopy = tocopy->next;	  
	}
    }
  current->next = NULL;
#else
  ProductNode* tocopy = p.head;
  ProductNode* current;

  head = new ProductNode;
  head->next = NULL;
  head->product.coefficient = 0;
  head->product.literal = 0;
  current = head;
  while (tocopy->next)
    {
      ProductNode* newnode = new ProductNode;
      
      newnode->product = tocopy->next->product;
      current->next = newnode;
      current = newnode;
      tocopy = tocopy->next;
    }      
  current->next = NULL;
#endif
}

Products::Products(SimpleString& s,Atoms* atoms)
{
  head = new ProductNode;
  head->next = NULL;
  head->product.coefficient = 0;
  head->product.literal = 0;
  constant = 0; 
  sumcoeffs = 0;
  this->initialize(s,atoms);
}

Products::Products(char *s,Atoms* atoms)
{
  head = new ProductNode;
  head->next = NULL;
  head->product.coefficient = 0;
  head->product.literal = 0;
  constant = 0; 
  sumcoeffs = 0;
  SimpleString ss(s);
  this->initialize(ss,atoms);
}

void Products::operator=(const Products& p)
{
  ProductNode *tocopy;

  constant = p.constant;
  sumcoeffs = p.sumcoeffs;
  if (head == p.head) // nothing to do, the same Products
    return;
  tocopy = p.head;
  current = head;
  while (tocopy->next && current->next) // Just copy the information as long as possible
    {
      current->next->product = tocopy->next->product;
      tocopy = tocopy->next;
      current = current->next;
    }
  if (tocopy->next == NULL)  // All copied, just free the remaining nodes of current
    {
      ProductNode* pn = current->next;
      ProductNode* pnext;

      while(pn)
	{
	  pnext = pn->next;
	  delete pn;
	  pn = pnext;
	}
      current->next = NULL;
    }
  else // if (current->next == NULL) // Construct New ones for the rest
    {
      while (tocopy->next)
	{
	  ProductNode* newnode = new ProductNode;

	  newnode->product = tocopy->next->product;
	  current->next = newnode;
	  current = newnode;
	  tocopy = tocopy->next;
	}      
      current->next = NULL;
    }	  
}

void Products::init(OrdInt& oi,Coefficient coefficient)
{
  // go not for speed
  clear();
  oi.first();
  while(!oi.last())
    add(coefficient,oi.next());
}

void Products::add(Coefficient coefficient,int literal)
{
  // add only product with positive coefficient
  if (coefficient < 0)
    {
      // ci*Li = ci + (-ci)*(1-Li)
      literal = -literal;
      constant += coefficient;
      coefficient = -coefficient;
    }
  ProductNode* newnode; // the new node (either updated or allocated)

  // is there is another product with the same variable
  ProductNode* actual = head->next;
  ProductNode* previous = head;

  while (actual && (actual->product.literal != literal)
                && (actual->product.literal != -literal))
    {
      previous = actual;
      actual = actual->next;
    }
  if (actual) // then (actual->product.literal == (+/- literal)
    {
      // there is another product with the same variable
      previous->next = actual->next; // link node out
      if (actual->product.literal == literal)
	{
	  actual->product.coefficient += coefficient;
	  sumcoeffs += coefficient;
	  newnode = actual;
        }
      else
	{
	  if (actual->product.coefficient > coefficient)
	    {
	      constant += coefficient;
	      sumcoeffs -= coefficient;
	      actual->product.coefficient -= coefficient;
	      newnode = actual;
	    }
	  else if (actual->product.coefficient < coefficient)
	    {
	      constant += actual->product.coefficient;
	      sumcoeffs = sumcoeffs + coefficient 
		- 2*(actual->product.coefficient);
	      actual->product.coefficient = 
		coefficient - actual->product.coefficient;
	      actual->product.literal *= -1;
	      newnode = actual;
	    }
	  else // node becomes deleted
	    {
	      constant += coefficient;
	      sumcoeffs -= coefficient;
	      delete actual;
	      newnode = 0;
	    }
        }
    }
  else // newnode needs to be constructed
    {
      newnode = new ProductNode;
      if (newnode == NULL)
	cerr << "SHIT \n";
      newnode->product.coefficient = coefficient;
      newnode->product.literal = literal;
      sumcoeffs += coefficient;
    }
  if (newnode == 0)
    return; // nothing to do
  // now newnode contains a node that needs to be inserted
  // but no further interaction with other products is possible
  actual = head;

  while (actual->next && 
         (actual->next->product.coefficient > newnode->product.coefficient))
    actual = actual->next;
  if (actual->next == 0)
    {
      newnode->next = 0;
      actual->next = newnode;
    }
  else
    {
      newnode->next = actual->next;
      actual->next = newnode;
    }
}

void Products::mult(Coefficient coefficient)
{
  if (coefficient == 0)
    {
      Coefficient i = constant;
      clear();
      constant = i;
      sumcoeffs = 0;
      return;
    }
  if (coefficient < 0)
    {
      coefficient = -coefficient;
      ProductNode* h = head->next;

      while (h)  // negate literals
	{
	  h->product.literal *= -1;
	  h = h->next;
	}
      constant = -(constant + sumcoeffs);
    }
  // now coefficient > 0
  ProductNode* h = head->next;

  while (h)  // negate literals
    {
      h->product.coefficient *= coefficient;
      h = h->next;
    }  
  sumcoeffs *= coefficient;
}

// not very useful for products
void Products::del(int literal)
{
  ProductNode* h = head;

  while (h->next && (h->next->product.literal != literal))
    h = h->next;
  if (h->next != NULL) // then (h->next->product.literal == literal)
    {
      ProductNode* hh = h->next;
      h->next = h->next->next;
      sumcoeffs -= hh->product.coefficient;
      delete hh;
    }
}

int Products::fix(int literal)
{
  ProductNode* actual = head->next;
  ProductNode* previous = head;
  
  while (actual && (actual->product.literal != literal)
                && (actual->product.literal != -literal))
    {
      previous = actual;
      actual = actual->next;
    }
  if (actual) // then (actual->product.literal == (+/- literal)
    {
      sumcoeffs -= actual->product.coefficient;
      if (actual->product.literal == literal)
	{
	  // fix to 1
	  previous->next = actual->next;
	  constant += actual->product.coefficient;
	  delete actual;
	}
      else
	{
	  // fix negation to 1; thus to 0
	  previous->next = actual->next;
	  delete actual;
	}
      return 1;
    }
  return 0;
}

int Products::equate(int lit1, int lit2)
{
  ProductNode* actual2 = head->next;
  ProductNode* previous2 = head;

//  cout << "start";cout.flush();  
  while (actual2 && (actual2->product.literal != lit2)
	         && (actual2->product.literal != -lit2))
    {
      previous2 = actual2;
      actual2 = actual2->next;
    }
  if (!actual2)
    return 0; // nothing to do
// cout << "found";cout.flush();
  ProductNode* actual1 = head->next;
  ProductNode* previous1 = head;
  
  while (actual1 && (actual1->product.literal != lit1)
	         && (actual1->product.literal != -lit1))
    {
      previous1 = actual1;
      actual1 = actual1->next;
    }
  if (!actual1) // not lit1 in ineq => just rename
    {
//      cout << "rename";cout.flush();
      if (actual2->product.literal == lit2)
	actual2->product.literal = lit1; 
      else // -lit1 = -lit2
	actual2->product.literal = -lit1; 
      return 1;
    }
  else // both in the left-hand side
    {
      // unlink lit2 but hold for data
      if (previous2->next == previous1)
	previous1 = previous2; // assure that previous1 is correct
      previous2->next = actual2->next; 
      // push the well coefficients and sign to actual1
      if (((actual2->product.literal ==  lit2) && (actual1->product.literal == lit1)) ||
	  ((actual2->product.literal == -lit2) && (actual1->product.literal == -lit1)))
	{
//	  cout << "same sign";cout.flush();
	  // the same sign
	  actual1->product.coefficient += actual2->product.coefficient;
	  // sumcoeffs and constant does not change
	  previous1->next = actual1->next; // unlink but hold
	}
      else // different sign
	{
//	  cout << "complicated";cout.flush();
	  if (actual1->product.coefficient == actual2->product.coefficient)
	    {
//	      cout << "equals";cout.flush();
	      // delete all nodes
	      sumcoeffs -= 2*actual1->product.coefficient;
	      constant += actual1->product.coefficient;
	      previous1->next = actual1->next; // unlink
	      delete actual1; // and delete
	      actual1 = 0;
	    }
	  else if (actual1->product.coefficient > actual2->product.coefficient)
	    {
//	      cout << "greater";cout.flush();
	      // retain sign
	      sumcoeffs -= 2*actual2->product.coefficient;
	      constant += actual2->product.coefficient;	      
	      actual1->product.coefficient -= actual2->product.coefficient;
	      previous1->next = actual1->next; // unlink
	    }
	  else // if (actual1->product.coefficient < actual2->product.coefficient)
	    {
//	      cout << "smaller";cout.flush();
	      sumcoeffs -= 2*actual1->product.coefficient;
	      constant += actual1->product.coefficient;	      	      
	      actual1->product.literal *= -1; // flip sign
	      actual1->product.coefficient = actual2->product.coefficient - actual1->product.coefficient;
	      previous1->next = actual1->next; // unlink
	    }
	}
      // delete lit2
      delete actual2;
      if (actual1 != 0) // remains to link in
	{
	  actual2 = head->next;
	  previous2 = head;
	  while (actual2 && (actual2->product.coefficient > actual1->product.coefficient))
	    {
	      previous2 = actual2;
	      actual2 = actual2->next;
	    }
	  actual1->next = actual2;
	  previous2->next = actual1;
	}
      return 2;
    }
}

void Products::clear()
{
  ProductNode* h = head->next;
  ProductNode* hh;

  constant = 0;
  sumcoeffs = 0;
  while (h)
    {
      hh = h->next;
      delete h;
      h = hh;
    }
  head->next = 0;
}

Coefficient Products::member(int literal)
{
  ProductNode* actual = head->next;
  
  while (actual && (actual->product.literal != literal)
                && (actual->product.literal != -literal))
    actual = actual->next;
  if (actual) // then (actual->product.literal == (+/- literal)
    {
      if (actual->product.literal == literal)
	return actual->product.coefficient;
      else
	return -actual->product.coefficient;
    }
  return 0; 
}

int Products::length()
{
  int len = 0;
  ProductNode* p = head->next;

  while(p)
    {
      len++;
      p = p->next;
    }
  return len;
}

Product& Products::max()
{
  ProductNode* actual = head->next;
  
  if(!actual)
    {
      cerr << "Products : max of emptyproducts; ";
      exit(-1);
    }
  return actual->product;
}

Product& Products::min()
{
  ProductNode* actual = head;
  
  if (!actual->next)
    {
      cerr << "Products : min of emptyset;";
      exit(-1);
    }
  while(actual->next)
    actual = actual->next;
  return actual->product;
}

Coefficient Products::get_upperbound()
{
  ProductNode* actual = head->next;
  
  Coefficient val = constant;
  while(actual)
    {
      if (actual->product.coefficient > 0)
	val += actual->product.coefficient;
      actual = actual->next;
    }
  return val;
}

Coefficient Products::get_lowerbound()
{
  ProductNode* actual = head->next;
  
  Coefficient val = constant;
  while(actual)
    {
      if (actual->product.coefficient < 0)
	val += actual->product.coefficient;
      actual = actual->next;
    }
  return val;
}


Coefficient Products::get_value(OrdInt& fixed, Coefficient& val ,Coefficient& remaining)
{
  ProductNode* actual = head->next;
  
  val = constant;
  remaining = 0;
  while(actual)
    {
      if (fixed.member(actual->product.literal))
	val += actual->product.coefficient;
      else if (fixed.member(-(actual->product.literal)))
	;
      else
	remaining += actual->product.coefficient;
      actual = actual->next;
    }
  return val;
}

/* io - routines */

void Products::Print(ostream& out, Atoms* atoms)
{
  ProductNode* p = head->next;

  while(p)
    {
      out << p->product.coefficient << "*";
      if (atoms)
	{
	  if (p->product.literal > 0)
	    out << (*atoms)(p->product.literal) << " ";
	  else
	    out << "~" << (*atoms)(-p->product.literal) << " ";
	}
      else
	out << p->product.literal << " ";
      p = p->next;
      if (p) 
	out << "+ ";
    }
  if (constant != 0)
    out << "+ " << constant << " ";
}

void pr_read_error(SimpleString& ,int i,const char* text)
{
  cerr << "Products : initialize : at pos " << i << " :" << text << "\n";
}

int skip_tab(SimpleString& s, int len, int& i)
{
  while(i < len)
    {
      switch (s[i]) {
      case ' ' :
      case '\t' :
	break;
      default :
	return 0;
      }
      i++;
    }
  return 1;
}

int get_coefficient(SimpleString& s, int len, int& i, Coefficient& coefficient)
{
  Coefficient oi = 0;
  int sign = 1;
  int started = 0;
  
  while(i < len)
    {
      switch (s[i]) {
      case '>'  :
      case '<'  :
	return 1;
      case ' '  :
      case '\t' :
      case '*'  :
	if (started && (oi == 0))
	  { 
	    pr_read_error(s,i,"zero read"); 
	    return 1; 
	  }
	if (oi != 0)
	  {
	    coefficient = sign*oi;
	    return 0;
	  }
	break;
      case '-' :
	if (started)
	  { 
	    pr_read_error(s,i,"- sign wrong place"); 
	    return 1; 
	  }
	else
	  { 
	    started = 1; 
	    sign = -1;
	  }
	break;
      case '+' :
	if (started)
	  { 
	    pr_read_error(s,i,"+ sign wrong place"); 
	    return 1; 
	  }
	else
	  { 
	    started = 1; 
	    sign = 1;
	  }
	break;
      default :
	if ((s[i] >= '0') && (s[i] <= '9'))
	  {
	    started = 1;
	    oi = oi*10 + (s[i] - '0');
	  }
	else
	  { 
	    pr_read_error(s,i,"unexpected char"); 
	    return 1; 
	  }
	break;
      }
      i++;
    }  
  if (oi != 0) // add last one
    {
      coefficient = sign*oi;
      return 0;
    }
  else 
    return 1;
}

int get_string_literal(SimpleString& s,int len,int& i,SimpleString& literal,int& negation)
{
  literal.clear();
  if (s[i] == '~')
    {
      i++;
      negation = 1;
    }
  int start = i;
  while(i<len)
    {
      if (((s[i] >= 'a') && (s[i] <= 'z')) ||
	  ((s[i] >= 'A') && (s[i] <= 'Z')) ||
	  ((s[i] >= '0') && (s[i] <= '9')) ||
	  (s[i] == '_'))
	{
	  i++;
	}
      else
	break;
    }
  if (start == i)
    { 
      return 1; 
    }
  literal.set_substring(s,start,i-1);
  return 0;
}

int get_opsign(SimpleString& s,int len,int& i)
{
  if (i >= len)
    return 1;
  if (s[i] == '+')
    {
      i++;
      return 1;
    }
  else if (s[i] == '-')
    {
      i++;
      return -1;
    }
  else 
    return 1;
}

int Products::initialize(char* s, Atoms* atoms)
{
  SimpleString ss(s);
  return this->initialize(ss,atoms);
}

int Products::initialize(SimpleString& s, Atoms* atoms)
{
  int len = s.length();
  int i = 0;
  int last_i=-1;
  Coefficient coefficient;
  int opsign; // 1 if coeff/lit is added -1 if subtracted
  int literal=0;
  int justconstant;

  clear();
  while (i < len)
    {
      if (i == last_i)
	{
	  pr_read_error(s,i,"unexpected character");
	  return 1;
	}
      last_i = i;
      justconstant = 0;
      skip_tab(s,len,i);
      // is there a leading + or -; if yes eat it
      opsign = get_opsign(s,len,i);
      skip_tab(s,len,i);
      if (i+1 < len) 
	{ 
	  if (((s[i] == '>') || (s[i] == '<')) && (s[i+1] == '='))
	    {
	      i += 2;
	      break;
	    }
	}
      if (s[i] == '*')
	{
	  pr_read_error(s,i,"unexpected '*'");
	  return 1;
	}
      if (((s[i] > '9') || (s[i] < '0')) && (s[i] != '+') && (s[i] != '-'))
	{
	  // assume a coefficient of 1
	  coefficient = 1;
	}
      else if (get_coefficient(s,len,i,coefficient))
	{
	  pr_read_error(s,i,"expected Product or >="); 
	  return 1; 
	}
      skip_tab(s,len,i);
      // just space is ok; '*' allowed
      if (i >= len) 
	{
	  justconstant = 1;
	}
      else
	{
	  if (s[i] == '*')
	    {
	      i++;
	      skip_tab(s,len,i);
	    }
	  if (atoms)
	    {
	      SimpleString stringliteral;
	      int negation=0;
	      skip_tab(s,len,i);
	      if (get_string_literal(s,len,i,stringliteral,negation))
		{
		  // if we have atoms, then we may also read just 
		  // constants
		  justconstant = 1;
		  // pr_read_error(s,i,"literal expected"); 	     
		  // return 1; 
		}
	      else 
		{ 
		  if (negation)
		    literal = -(*atoms)(stringliteral);
		  else 
		    literal = (*atoms)(stringliteral);
		}
	    }
	  else
	    {
	      Coefficient __coefficient;
	      if (get_coefficient(s,len,i,__coefficient))
		{ 
		  pr_read_error(s,i,"literal expected"); 
		  return 1; 
		}
	      literal = coefficient_to_int(__coefficient);
	    }
	}
      skip_tab(s,len,i);
      if (justconstant)
	constant += opsign*coefficient;
      else
	add(opsign*coefficient,literal);
    }
  if (i >= len)
    return 0;
  // there is a {<,>}= read
  int swap = 0;
  if (s[i-2] == '<') // it was a <=
    swap = 1;
  Coefficient rhs;
  skip_tab(s,len,i);
  int savei = i;
  while ((i < len) && (s[i] == '0'))
    i++;
  if (i >= len)
    {
      if (savei < i)
	rhs = 0;
      else
	{
	  pr_read_error(s,i,"expected rhs after '>='"); 
	  return 1;
	}
    }
  else if ((s[i] == '+') || (s[i] == '-') ||
      ((s[i] >= '0') && (s[i] <= '9')))
    {
      if (get_coefficient(s,len,i,rhs))
	{
	  pr_read_error(s,i,"expected rhs after '>='"); 
	  return 1;
	}
    }
  else
    {
      pr_read_error(s,i,"expected rhs after '>='"); 
      return 1;
    }
  constant -= rhs; // we then represent term >= 0!
  if (swap) // it was a <=
    {
      Coefficient rhs2 = -constant;
      constant = 0;
      mult(-1);
      constant += rhs2; // I too don't like that hack
    }
  return 0;
}

int my_strlen(const char *s)
{
  int len = 0;
  while(*s++)
    len++;
  return len;
}

char* product_append_chars(char* s,char* toapp)
{
  if (s)
    {
      char* sptr = s;
      int len = my_strlen(s);
      len += my_strlen(toapp);
      char* newbuf = new char[len+2];
      char* newbufptr = newbuf;
      if (newbuf == 0)
	{
	  cerr << "no more memory\n";
	  exit(-1);
	}
      while(*s != 0x00)
	*newbuf++ = *s++;
      free(sptr);
      while(*toapp != 0x00)
	*newbuf++ = *toapp++;
      *newbuf = 0;
      return newbufptr;
    }
  else
    {
      int len = my_strlen(toapp);
      char* newbuf = new char[len+2];
      char* newbufptr = newbuf;
      if (newbuf == 0)
	{
	  cerr << "no more memory\n";
	  exit(-1);
	}
      while(*toapp != 0x00)
	*newbuf++ = *toapp++;
      *newbuf = 0;
      return newbufptr;
    }
}

char* products_read_a_line(istream& in)
{
  char buf[512];
  char* p = buf;
  char* q = p+511;
  char* s=0;

  do {
    while (p < q && in.get(*p))
      { 
	if (*p == '\n') 
	  break;
	p++;
      } 
    if ((*p == '\n') || !in)
      {
	*p = '\0';
	s = product_append_chars(s,buf);
	break;
      }
    else
      {
	*p = '\0';
	s = product_append_chars(s,buf);
	p = buf;
      }
  } while(1);
  return s;
}

int Products::Read(istream& in, Atoms* atoms)
{
  char* s;

  s = products_read_a_line(in); 
  if (this->initialize(s,atoms))
    {
      free(s);
      return 1;
    }
  else
    {
      free(s);
      return 0;
    }
}


/////////////////////////////
// ProductNode
/////////////////////////////

// should not be used
ProductNode::ProductNode(const ProductNode& p)
{
  cerr << "Hey, copy constructor of ProductNode" << " should not be called\n";
  cerr << p.product.coefficient << "*" << p.product.literal << "\n";
}

// should not be used
void ProductNode::operator=(const ProductNode& p)
{
  cerr << "Hey, assignment of ProductNode" << " should not be called\n";
  cerr << p.product.coefficient << "*" << p.product.literal << "\n";
}

#ifdef OWN_MEMORY_ProductNode
ProductNode* ProductNode::pfree = (ProductNode *) NULL;
#endif
