#ifndef _Products_h
#define _Products_h

#include "streaminclude.h"
#include "OrdInt.h"
#include "SimpleString.h"
#include "Atoms.h"
#include "Coefficient.h"

class ProductNode;

struct Product {
  Coefficient coefficient;
  int literal;
  Product() {};
  Product(const Product& p) : coefficient(p.coefficient), literal(p.literal) {}
  Product& operator=(const Product& p) { coefficient = p.coefficient; literal = p.literal; return *this; }
  ~Product() {};
};

class Products {
public:
  Products();
  ~Products();
  Products(const Products& );
  Products(SimpleString& , Atoms* atoms = 0);
  Products(char *, Atoms* atoms = 0);
  void operator=(const Products& );
 
  void init(OrdInt& ,Coefficient coefficient=1);
  void add(Coefficient coefficient, int literal);
  void mult(Coefficient coefficient);
  void del(int literal); // not very usable
  int fix(int literal);
  int equate(int lit1, int lit2); // lit1 = lit2; lit2 does not occur any longer
  void clear();
  Coefficient member(int literal); // returns positive coefficient; 0 otherwise
  int length();
  int empty();
  Product& max();
  Product& min();
  Coefficient get_constant() { return constant; }
  Coefficient get_sumcoeffs() { return sumcoeffs; }
  void set_constant(Coefficient c) { constant = c; }
  Coefficient get_upperbound();
  Coefficient get_lowerbound();
  Coefficient get_value(OrdInt& fixed, Coefficient& val ,Coefficient& remaining);

  // Iterator
  void first();
  Product& next();
  int last();

  int initialize(SimpleString& s, Atoms* atoms = 0);
  int initialize(char* s, Atoms* atoms = 0);
  int Read(istream& in=cin, Atoms* atoms = 0);
  void Print(ostream& out=cout, Atoms* atoms = 0);
  friend ostream& operator<<(ostream& ,Products& );
  friend istream& operator>>(istream& ,Products& );

private:
  ProductNode* head;
  ProductNode* current;
  Coefficient constant;
  Coefficient sumcoeffs;
  void free_list(ProductNode*);
friend class LPBIneq;
friend class LPBIneqSet;
friend class LPBIRefs;
friend class Enumerate;
friend class PBCS;
};

inline void Read(Products& p, istream& in) {p.Read(in);}
inline void Print(Products& p, ostream& out) { p.Print(out); }
inline istream& operator>>(istream& in ,Products& p) { p.Read(in); return in; }
inline ostream& operator<<(ostream& out,Products& p) { p.Print(out); return out; }


class ProductNode {
public:
  ProductNode() {};
  ~ProductNode() {};
  ProductNode(const ProductNode& p);
  void operator=(const ProductNode& p);

// own memory allocation and management works only with native types
#if (COEFFICIENT_int || COEFFICIENT__int64)
#define OWN_MEMORY_ProductNode
#endif

#ifdef OWN_MEMORY_ProductNode
private:
  Product product;
  ProductNode* next;
  enum {NALL = 2048 };
  static ProductNode* pfree;
  // We use our own memory allocation as described in Stroustrupp page 176
  // Because of a compiler error in cfront we can't use decl. + inline,
  // but have to include the code in the class.
  void* operator new(size_t)
  {
    register ProductNode* p = pfree; // first allocate
      
      if (p)
	pfree = p->next;
      else
	{
	  ProductNode* q = (ProductNode *) new char[NALL * sizeof(ProductNode)];
	  for (p=pfree=&q[NALL-1]; q<p; p--)
	    {
	      p->next = p-1;
	    }
	  (p+1)->next = NULL;
	}
    return p;
  }
  void operator delete(void* p)  // ,size_t ) deleted, does not work with g++2.5.2
  {
    ((ProductNode*)p)->next = pfree;
    pfree = (ProductNode *) p;    
  }
#else
private:
  Product product;
  ProductNode* next;
#endif
friend class Products;
friend class LPBIneq;
friend class LPBIneqSet;
friend class LPBIRefs;
friend class Enumerate;
};

// inlines which use ProductNode; so because of the silly bloddy
// C++ compilers we have to append them at the end
inline void Products::first() {  current = head; }
inline Product& Products::next() 
{ 
  if (current->next) 
    { current = current->next; 
      return current->product; }
  else 
    {
      return head->product; 
    }
}
inline int Products::last()  { return (current->next) ? 0 : 1; }
 

// further Products inlines that have to be appended on the end, since
// ProductNode* is not really known before !%^$@#
inline int Products::empty() { return (head->next == NULL) ? 1 : 0; }



#endif



