#ifndef RArrayList_h
#define RArrayList_h

template <class T> class RArrayList {
private:
  struct Node {
    T t;
    int prev;
    int next;
  };
  int size;
  int first_index;
  Node* a;

  int valid_node(Node* n) { return n->next > 0; }  
  void move(int,int);

public:
  RArrayList(int s=32);
  ~RArrayList();
  RArrayList(const RArrayList&);
  void operator=(const RArrayList&);

  T& operator[](int i) { return a[i].t;} // no check but fast!
  void enlarge(int );
    
  // list functions
  // a[0] is last element and not used
  // a[0].next = a[0].prev = 0
  int next(int i) { return a[i].next; }
  int prev(int i) { return a[i].prev; }
  void unlink(int i) 
  {
    // Note: unsafe if empty since next then is -1!!! (but faster :-)
    a[a[i].next].prev = a[i].prev;
    a[a[i].prev].next = a[i].next;
    if (i == first_index)
      first_index = a[i].next;
    a[i].next = -1;
  }  
  int active_index(int i) { return i != 0; }
  int active_elem(int i)  { return a[i].next >= 0; }

  // destructive unsafe, but fast, routines
  void mark_inactive(int i) { a[i].next = -1; } // unsafe if linked in list!
  // again unsafe but fast
  void copy_node(int i,RArrayList& ra) { a[i] = ra.a[i]; }
  void copy_first_index(RArrayList& ra) { first_index = ra.first_index; } 

  int first() { return first_index; }
  int empty() { return first_index == 0; }

  void make_list(int i) { make_list(1,i); } // initializes the list from 1 to i
  void make_list(int i, int j); // initializes the list from i to j
  void swap(int i, int j); // swaps in list not in place!
  void insert_after(int i, int j); // inserts a[i] after a[j] in list not in place
  void insert_front(int i); // inserts a[i] as first elem in list not in place
};

#ifdef __GNUC__
#ifndef RArrayList_C
#include "RArrayList.C"
#endif
#endif

#endif
