#ifndef SimpleList_h
#define SimpleList_h

template <class T> class SimpleListNode {
public:
  T info; 
  SimpleListNode<T>* next;
  SimpleListNode<T>* prev;

  SimpleListNode() { next = prev = 0; }
  SimpleListNode(T& t) : info(t) { next = prev = 0; }
  ~SimpleListNode() { }
};

template <class T> class SimpleList {
public:
  SimpleList();
  ~SimpleList();
  SimpleList(const SimpleList&);
  SimpleList& operator=(const SimpleList& );

  int length();
  int empty();
  void clear();

  void push(T& );
  T& top();
  void pop();

  // iteration
  void first();
  int is_first();
  void last();
  int is_last();
  T& get_current();
  int is_valid_current();
  void forward(); 
  void backward(); 

  void insert_after_current(T& );  // doesn't change current !
  void insert_before_current(T& ); // doesn't change current !
  void delete_current(int dir = -1); // dir<0 -> current=prev; >0 =next; =0 =0
  
private:
  // head is a node containing no information, but its next pointer
  // points to the first element and its previous pointer to the 
  // last element of the list
  // if the list is empty, then head->next = head->previous = head
  SimpleListNode<T>* head; 
  // current is a pointer to some node in the list
  // it is invalid if it points to head
  SimpleListNode<T>* current;
};

#ifdef __GNUC__
#ifndef SimpleList_C
#include "SimpleList.C"
#endif
#endif
  

#endif
