#include "SimpleString.h"

SimpleString::SimpleString()
{
  alloclen = ALLOCMIN;
  s = new char[alloclen];
  if (s == 0)
    {
      cerr << "SimpleString : SimpleString : no more memory \n";
      exit(-1);
    }
  s[0] = 0;
  len = 0;
}

SimpleString::~SimpleString()
{
  delete [] s;
}

SimpleString::SimpleString(const SimpleString& ss)
{
  alloclen = ss.alloclen;
  s = new char[alloclen];
  if (s == 0)
    {
      cerr << "SimpleString : SimpleString : no more memory \n";
      exit(-1);
    }
  char* ssptr = ss.s;
  char* sptr = s;
  while((*sptr++ = *ssptr++))
    ;
  len = ss.len;
}

SimpleString::SimpleString(const char* ssptr)
{
  int sslen = strlen(ssptr);
  alloclen = alloclen_for(sslen+1);
  s = new char[alloclen];
  if (s == 0)
    {
      cerr << "SimpleString : SimpleString : no more memory \n";
      exit(-1);
    }
  char* sptr = s;
  while((*sptr++ = *ssptr++))
    ;
  len = sslen;
}
 
SimpleString& SimpleString::operator=(const SimpleString& ss)
{
  if (alloclen < ss.alloclen)
    {
      delete [] s;
      alloclen = ss.alloclen;
      s = new char[alloclen];
      if (s == 0)
	{
	  cerr << "SimpleString : SimpleString : no more memory \n";
	  exit(-1);
	}
    }
  char* ssptr = ss.s;
  char* sptr = s;
  while((*sptr++ = *ssptr++))
    ;
  len = ss.len;
  return *this;
}

SimpleString& SimpleString::operator=(const char *ssptr)
{
  int sslen = strlen(ssptr);
  if (alloclen < sslen+1)
    {
      delete [] s;
      alloclen = alloclen_for(sslen);
      s = new char[alloclen];
      if (s == 0)
	{
	  cerr << "SimpleString : SimpleString : no more memory \n";
	  exit(-1);
	}
    }
  char* sptr = s;
  while((*sptr++ = *ssptr++))
    ;
  len = sslen;
  return *this;
}

void SimpleString::clear()
{
  len = 0;
  s[0] = 0;
}

int SimpleString::length()
{
  return len;
}

SimpleString& SimpleString::operator+=(SimpleString& ss)
{
  return operator+=(ss.s);
}

SimpleString& SimpleString::operator+=(char* ssptr)
{
  int sslen = strlen(ssptr);
  enlarge_alloclen(len + sslen + 1);
  char *sptr = s;
  while (*sptr)
    sptr++;
  while((*sptr++ = *ssptr++))
    ;
  len += sslen;
  return *this;
}

// ordering
int SimpleString::compare(SimpleString& ss)
{
  return compare(ss.s);
}

int SimpleString::compare(char* ssptr)
{
  char* sptr = s;
  while(*sptr && *ssptr)
    {
      if (*sptr == *ssptr)
	{
	  sptr++;
	  ssptr++;
	}
      else
	{
	  if (*sptr < *ssptr)
	    return -1;
	  else
	    return 1;
	}
    }
  if (*sptr == *ssptr) // both 0
    return 0;
  if (*ssptr == 0) // && *sptr != 0
    return 1; // this is greater
  else
    return -1;
}

// Subsstring
SimpleString SimpleString::operator()(int i, int j)
{
  if (j < i)
    {
      cerr << "SimpleString : operator() : empty substring \n";
      exit(-1);
    }
  if (i < 0 || j >= len)
    {
      cerr << "SimpleString : operator() : index out of range \n";
      exit(-1);
    }
  SimpleString substring;
  substring.enlarge_alloclen(j-i+1);
  substring.len = j-i+1;
  int subi = 0;
  while (i <= j)
    {
      substring.s[subi++] = s[i++];
    }
  substring.s[subi] = 0;
  return substring;
}

void SimpleString::set_substring(SimpleString& what,int i, int j)
{
  if (j < i)
    {
      cerr << "SimpleString : operator() : empty substring \n";
      exit(-1);
    }
  if (i < 0 || j >= what.len)
    {
      cerr << "SimpleString : operator() : index out of range \n";
      exit(-1);
    }  
  enlarge_alloclen(j-i+1);
  len = j-i+1;
  int subi = 0;
  while (i <= j)
    {
      s[subi++] = what[i++];
    }
  s[subi] = 0;
}

void SimpleString::del(int from, int to)
{
  if (from < 0)
    from = 0;
  if (to > len)
    to = len;
  if (from > to)
    return;
  for(int i=to+1;i<=len;i++) // include copying 0
    {
      s[i-(to-from+1)] = s[i];
    }
  len -= to-from+1;
}

void SimpleString::ins(SimpleString& ss, int i)
{
  ins(ss.s,i);
}

void SimpleString::ins(char* ss,int i)
{
  if (i > len)
    {
      operator+=(s);
      return;
    }
  if (i < 0)
    i = 0;
  int inslen = this->strlen(ss);
  if (inslen == 0)
    return;
  enlarge_alloclen(inslen+len);
  int j;
  for(j=len;j >= i;j--) // for all including 0x00
    {
      s[j+inslen] = s[j];
    }
  for(j=0;j<inslen;j++)
    {
      s[j+i] = *ss++;
    }
  len = inslen+len;
}

// searching
int SimpleString::find(SimpleString& ss,int start)
{
  return find(ss.s,start);
}

int SimpleString::find(char* f,int start)
{
  // simple algorithm (no KMP ... (its called SIMPLE!!String!!))
  if (start < 0)
    start = 0;
  if (f == 0) // empty string always matches
    return start;
  while(start < len)
    {
      if (*f == s[start]) // first character matches
	{
	  int i = 1;
	  while (f[i] && s[start+i] && (f[i] == s[start+i]))
	    i++;
	  if (f[i] == 0)
	    return start;
	  start++;
	}
      else
	start++;
    }
  return -1;
}

int SimpleString::skip_chars(SimpleString& ss,int start)
{
  return skip_chars(ss.s,start);
}


int SimpleString::skip_chars(char* skip,int start)
{
  while(start < len)
    {
      if (strmember(s[start],skip))
	start++;
      else
	break;
    }
  if (start < len)
    return start;
  else
    return -1;
}


int SimpleString::Read(istream& in,char* delims)
{ // mainly from LEDA
  char buf[512];
  char* q = buf+511;
  int go = 1;
  char *p;

  clear();
  while (in && go)
  { 
    for(p = buf; p < q && in.get(*p); p++)
      { 
	if (strmember(*p,delims))
	  { 
	    if (*p != '\n') 
	      in.putback(*p);
	    go = 0;
	    break;
	  }
      } 
    *p = '\0';
    operator+=(buf);
  }
  return 0;
}
 
void SimpleString::Print(ostream& out) const
{
  out << s;
}

ostream& operator<<(ostream& out,const SimpleString& ss)
{
  ss.Print(out);
  return out;
}

istream& operator>>(istream& in,SimpleString& ss)
{
  ss.Read(in);
  return in;
}

// helpfunctions

int SimpleString::alloclen_for(int tolen)
{
  int newalloclen = ALLOCMIN;

  while(newalloclen < tolen)
    newalloclen *= 2;
  return newalloclen;
}

int SimpleString::enlarge_alloclen(int tolen)
{
  if (alloclen > tolen)
    return 0;
  alloclen = alloclen_for(tolen);
  char* h = new char[alloclen];
  if (h == 0)
    {
      cerr << "SimpleString : enlarge_alloclen : no more memory \n";
      exit(-1);
    }
  char* hptr = h;
  char* sptr = s;
  while((*hptr++ = *sptr++))
    ;
  delete [] s;
  s = h;		     
  return 0;
}

int SimpleString::strlen(const char *sptr)
{
  int slen = 0;
  while(*sptr++)
    slen++;
  return slen;
}

int SimpleString::strmember(char c,char* sptr)
{
  while(*sptr)
    {
      if (*sptr++ == c)
	return 1;
    }
  return 0;
}
