#ifndef SimpleString_h
#define SimpleString_h

#include "streaminclude.h"

class SimpleString {
public:
  SimpleString();
  ~SimpleString();
  SimpleString(const SimpleString&);
  SimpleString(const char* );
  SimpleString& operator=(const SimpleString&);
  SimpleString& operator=(const char* );

  void clear(); 
  int length();
  int is_empty();
  char& operator[](int );
  SimpleString& operator+=(SimpleString&);
  SimpleString& operator+=(char* );

  // lexicographic ordering (empty string is smallest)
  int compare(SimpleString& ss);
  int compare(char* ss);
  int operator==(SimpleString& ss) { return compare(ss) == 0; }
  int operator==(char* ss) { return compare(ss) == 0; }
  int operator!=(SimpleString& ss) { return compare(ss) != 0; }
  int operator!=(char* ss) { return compare(ss) != 0; }
  int operator<(SimpleString& ss) { return compare(ss) < 0; }
  int operator<(char* ss) { return compare(ss) < 0; }
  int operator<=(SimpleString& ss) { return compare(ss) <= 0; }
  int operator<=(char* ss) { return compare(ss) <= 0; }
  int operator>(SimpleString& ss) { return compare(ss) > 0; }
  int operator>(char* ss) { return compare(ss) > 0; }
  int operator>=(SimpleString& ss) { return compare(ss) >= 0; }
  int operator>=(char* ss) { return compare(ss) >= 0; }
  
  // Substring; starting from index i to j 
  SimpleString operator()(int i, int j);
  void set_substring(SimpleString& ,int , int );

  // delete substring from index from to to
  // if from < 0 from beginning; if to >= len until end
  void del(int from, int to);

  // insert string beginning position i ; move rest backwards
  // if i < 0; i = 0; i>len; append
  void ins(SimpleString& ins, int i);
  void ins(char *ins, int i);

  // searching; optionally after a specific index
  // returns -1 if not found; the index of the found position otherwise
  int find(SimpleString& ,int start=0);
  int find(char* ,int start=0);
  // skips all chars that occur in the first argument
  // i.e. as along as such a char is found start is incremented
  // and finally returned (-1 if end of string is reached)
  int skip_chars(SimpleString& ,int start=0);
  int skip_chars(char* ,int start=0);

  int Read(istream& in=cin,char* delims = " \t\n");
  void Print(ostream& out=cout) const;
  friend ostream& operator<<(ostream& ,const SimpleString& );
  friend istream& operator>>(istream& ,SimpleString& );
private:
  char* s;
  int len;
  int alloclen;
  enum {ALLOCMIN = 32};

  // internal functions
  int alloclen_for(int );
  int enlarge_alloclen(int );
  int strlen(const char *);
  int strmember(char,char*);
};

inline int SimpleString::is_empty() { return len == 0; }
inline char& SimpleString::operator[](int i)
{
  if (i < 0 || i >= len)
    {
      cerr << "SimpleString : operator[] : index out of range \n";
      exit(-1);
    }
  return s[i];
}


#endif

