#include "lp_write.h"
#include "SimpleString.h"

void lp_products_write(ostream& out, Products& objf, Atoms* atoms, int solver)
{
  int count=0;
  Coefficient constant = objf.get_constant();
  objf.first();
  while(!objf.last())
    {
      count++;
      Product& p = objf.next();
      if (p.literal < 0)
	{
	  out << "- ";
	  out << p.coefficient;
	  constant += p.coefficient;
	}
      else
	{
	  out << "+ ";
	  out << p.coefficient;
	}
      out << " ";
      if (atoms == 0)
	{
	  out << "C";
	  out << abs(p.literal);
	}
      else
	{
	  out << (*atoms)(abs(p.literal));
	}
      out << " ";
      if (solver == CPLEX)
	{
	  if (count > 8)
            {
              out << "\n";
              count = 0;
            }
	}
    }  
  if (constant != 0)
    {
      out << " + ";
    }
  if (solver == CPLEX)
    out << "\n";
  else
    out << ";\n";    
}

void lp_lpbineq_write(ostream& out, LPBIneq& lpbi, Atoms* atoms,int solver)
{
  int count = 0;
  Coefficient rhs = lpbi.get_rhs();
  lpbi.first();
  while(!lpbi.last())
    {
      count++;
      Product& p = lpbi.next();
      if (p.literal < 0)
	{
	  out << "- ";
	  out << p.coefficient;
	  rhs -= p.coefficient;
	}
      else
	{
	  out << "+ ";
	  out << p.coefficient;
	}
      out << " ";
      if (atoms == 0)
	{
	  out << "C";
	  out << abs(p.literal);
	}
      else
	{
	  out << (*atoms)(abs(p.literal));
	}
      out << " ";
      if (solver == CPLEX)
	{
	  if (count >= 8)
	    {
	      out << "\n";
	      count = 0;
	    }
	}
    }  
  out << " >= ";
  out << rhs;
  if (solver == CPLEX)
    out << "\n";
  else
    out << ";\n";
}

void lp_bounds_write(ostream& out,OrdInt& av,Atoms* atoms,int solver)
{
  if (solver == OPBDP)
    return;
  if (solver == CPLEX)
    out << "Bounds\n";
  else
    out << "/* Bounds */\n";
  av.first();
  while(!av.last())
    {      
      if (solver == CPLEX)
	out << "0 <=";
      if (atoms == 0)
	out << "C" << av.next();
      else
	out << (*atoms)(av.next());
      if (solver == CPLEX)
	out << " <= 1\n";
      else
	out << " <= 1;\n";
    }
}

void lp_integers_write(ostream& out,OrdInt& av,Atoms* atoms,int solver)
{
  if (solver == OPBDP)
    return;
  if (solver == CPLEX)
    out << "Integers\n";
  else
    out << "int ";
  av.first();
  int i=0;
  while(!av.last())
    {      
      if (atoms == 0)
	out << "C" << av.next() << " ";
      else
	out << (*atoms)(av.next()) << " ";
      i++;
      if (solver == CPLEX)
	{
	  if (i%8 == 0)
	    out << "\n";
	}
    }
  if (solver == CPLEX)
    out << "\n";
  else
    out << ";\n";
}

void lp_write(ostream& out, LPBIneqSet& S, Products& objf, int max, Atoms* atoms, int solver)
{  
  if (max)
    if (solver == CPLEX)
      out << "Maximize\n";
    else
      out << "max: ";
  else
    if (solver == CPLEX)
      out << "Minimize\n";
    else
      out << "min: ";
  if (solver == OPBDP)
    {
      objf.Print(out,atoms);
      out << "\n";
    }
  else
    lp_products_write(out,objf,atoms,solver);
  if (solver == CPLEX)
    out << "Subject To\n";
  S.first();
  while(!S.last())
    {      
      LPBIneq& lpbi = S.next();
      if (solver == OPBDP)
	{
	  lpbi.Print(out,atoms);
	  out << "\n";
	}
      else
	lp_lpbineq_write(out,lpbi,atoms,solver);
    }
  OrdInt av = S.all_vars();
  lp_bounds_write(out,av,atoms,solver);
  lp_integers_write(out,av,atoms,solver);
  if (solver == CPLEX)
    out << "End\n";
}
