#include "opbdp_read.h"

void delete_constraint_name(SimpleString& s)
{
  int index = s.find(":");
  if (index == -1)
    return;
  else
    {
      s.del(0,index);
    }
}

void delete_punctation(SimpleString& s)
{
  int index=0;
  while ((index = s.find(";",index)) != -1)
    {
      s.del(index,index);
    }
}

// clears the string if empty line and handles comments
// * something at the beginning of a line is a comment
// /* something */ is a comment
int is_empty_line(SimpleString& s,int line_counter)
{
  if (s.is_empty())
    return 1;
  if (s[0] == '*')
    {
      s.clear();
      return 1;
    }
  // delete comments in a line
  int index;
  while ((index = s.find("/*")) != -1)
    {
      int endcomment = s.find("*/",index+2);
      if (endcomment == -1)
	{
	  cerr << "open_comment /* found in line " << line_counter << "\n";
	  return -1;
	}
      s.del(index,endcomment+1);
    }
  // just blanks and tabs remaining?
  int i;
  for(i=0;i<s.length();i++)
    {
      if ((s[i] != ' ') && (s[i] != '\t'))
	break;
    }
  if (i == s.length()) // an empty line
    {
      s.clear();
      return 1;
    }
  return 0;
}

int addS(LPBIneqSet& S, LPBIneq& lpbi, SimpleString& s, Atoms* atoms, int line_counter)
{
  if (lpbi.initialize(s,atoms))
    {
      cerr << "Problem reading constraint in line " << line_counter << "\n";
      return 1;
    }
  if (!lpbi.is_tautology())
    {
      if (lpbi.is_contradiction())
	cout << "Contradition added in line " << line_counter << " \n";
      S.add(lpbi);
      // cout << s << "\n";
      // lpbi.Print(cout,atoms); cout << "\n";
    }
  else
    cout << "Tautology deleted in line " << line_counter << " \n";
  return 0;
}

int opbdp_read(Atoms& atoms,LPBIneqSet& S,Products& objf,istream& in,int& max)
{
  int line_counter=0;
  int equation_counter=0;
  SimpleString s;
  max = 1;
    
  // skip comment lines
  do { 
    s.Read(in,"\n");
    delete_punctation(s);
    line_counter++;
    if (is_empty_line(s,line_counter) == -1)
      return 1; // a syntax error
    if (!in)
      break;
  } while (s.length() == 0);
  if (s.length() >= 4)
    {
      SimpleString h=s(0,3);
      if (h == "min:")
	{
	  max = 0;
	  s = s(4,s.length()-1);	  
	}
      else if (h == "max:")
	{
	  s = s(4,s.length()-1);	  
	}
    }
  if ((s.length() == 0) || (objf.initialize(s,&atoms)))
    {
      cerr << "Problem reading objective function\n";
      return 1;
    }
  LPBIneq lpbi;
  // read all the constraints
  while(in)
    {
      // skip comment lines
      do { 
	s.Read(in,"\n");    
	line_counter++;
	if (is_empty_line(s,line_counter) == -1)
	  return 1; // syntax error
	else
	  {
	    delete_constraint_name(s);
	    delete_punctation(s);
	  }
	if (!in)
	  break;
      } while (s.length() == 0);
      if (s.length() > 0)
	{
	  int eqi = s.find("=");
	  if ((eqi > 0) && ((s[eqi-1] != '<') && (s[eqi-1] != '>')))
	    {
	      equation_counter++;
	      SimpleString greater = ">";
	      s.ins(greater,eqi);
	      if (addS(S,lpbi,s,&atoms,line_counter) == 1)
		return 1;
	      s[eqi] = '<';
	      if (addS(S,lpbi,s,&atoms,line_counter) == 1)
		return 1;
	    }
	  else
	    {	      
	      if (addS(S,lpbi,s,&atoms,line_counter) == 1)
		return 1;
	    }
	}
    }
  if (equation_counter == 1)
    cout << "1 equation changed into 2 inequalities \n";
  else if (equation_counter > 1)
    cout << equation_counter << " equations changed into " << equation_counter*2 << " inequalities \n"; 
  cout << "problem consists of " << S.card() << " inequalities and ";
  cout << atoms.size() << " variables \n";
  return 0;
}
