// pbcsia.C
// Author : pb
// Date : 19.5.95
// A Pseudo-Boolean Constraint Solver; interactive

#include "streaminclude.h"
#include "fstream.h"
#include "timing.h"
#include "PBCS.h"
#include "MGetOpt.h"
#include "Atoms.h"
#include "SimpleString.h"
#include "OrdInt.h"
#include "opbdp_read.h"
#include "readnonlinear.h"
#include "lp_write.h"

// the data
PBCS pbcs; // the pseudo-Boolean constraints

Products objf;   // the objective function
Atoms atoms;     // The Atomstable
int max;         // whether maximizing or minimizing
OrdInt ones;     // the solution

// where the commands are stored
int command_no = 1;
SimpleString command;
SimpleString arguments[20];
SimpleString history[100];


int command_loop()
{
  for (int k=0;k < 20;k++)
    arguments[k] = "";
  SimpleString line;
  cout << "opbdp [" << command_no++ << "] : ";
  line.Read(cin,"\n");
  line.del(0,line.skip_chars(" \t")-1); // delete leading blanks
  
  // a very simple history mechanism
  int number = 0;
  int i=0;
  while ((i < line.length()) && ((line[i] >= '0') && (line[i] <= '9')))
    number = number*10 + (line[i++] - '0');	  
  if (number > 0)
    {
      // got a history expansion
      line = history[number % 100];
      cout << number % 100 << " : " << line << "\n";
      command_no--;
    }
  else
    {
      history[(command_no-1) % 100] = line;
    }

  int blank = line.find(" ");
  if (blank == -1)
    {
      command = line;
      return 0;
    }
  command = line(0,blank-1);
  while ((blank < line.length()) && (line[blank] == ' ' ))
    blank++;
  if ((command == "add") || (command == "lowerbound") ||
      (command == "min") || (command == "max"))
    { // functions which have a single string as argument
      arguments[0] = line(blank,line.length()-1);
      return 1;
    }
  int start = blank;
  i=0;
  while ((blank = line.find(" ",start)) != -1)
    {
      arguments[i] = line(start,blank-1);
      while ((blank < line.length()) && (line[blank] == ' ' ))
	blank++;
      start = blank;
      i++;
    }
  if (start <= line.length()-1)
    {
//      cout << line(start,line.length()-1) << "\n";
      arguments[i] = line(start,line.length()-1);
    }
  return 1;
}

int do_atoi(SimpleString& s)
{
  char which_c[80];
  int i;
  for(i=0;i < s.length();i++)
    which_c[i] = s[i];	      
  which_c[i] = 0;
  int which = atoi(which_c);
  return which;
}

int get_literal(SimpleString& s)
{
  // extract and handle '~' 
  int sign = 1;
  int position;
  int literal;
  if ((position = s.find("~")) != -1)
    {
      sign = -1;
      s = s(position+1,s.length()-1);
    }
  if (atoms.inserted(s))
    {
      literal = sign*atoms(s);
      return literal;
    }
  else
    {
      cout << s << " is not a literal in a constraint set\n";
      return 0;
    }
}

int cout_literal(int literal)
{
  if (literal > 0)
    cout << atoms(literal) << " ";
  else
    cout << "~" << atoms(-literal) << " ";  
  return 0;
}

int cout_literals(OrdInt literals)
{
  literals.first();
  while(!literals.last())
    cout_literal(literals.next());
  return 0;
}

float tmptime,usedtime;

void start_timing()
{
  tmptime = used_time();
}

void show_timing()
{
  usedtime = used_time() - tmptime;
  cout << "Time used (secs) : " << usedtime << "\n";	  
}

#ifdef QUANTIFIED
extern "C" {
quantify_start_recording_data();
quantify_stop_recording_data();
}
#endif

int show_lpbi_iid(PBCS& _pbcs, int iid)
{
  LPBIneq lpbi = _pbcs.get_lpbi(iid);
  cout << iid << " : "; 
  lpbi.Print(cout,&atoms); 
  cout << "\n";
  return 0;
}

int show_all_lpbis(PBCS& _pbcs)
{
  OrdInt alliids = _pbcs.get_lpbi_iids();
  alliids.first();	      
  while(!alliids.last())
    show_lpbi_iid(_pbcs,alliids.next());
  return 0;
}

int show_ranges_sol(Products term, OrdInt solution)
{
  if (term.empty())
    return 0;
  
  // apply equations 
  Equations Eq = pbcs.get_equations();
  Eq.first();
  while(!Eq.last())
    {
      int l = Eq.next();
      int r = Eq(l);
      term.equate(r,l); // unnatural ordering significant!
    }
  // apply fixings
  OrdInt Fixed = pbcs.get_fixed();
  Fixed.first();
  while(!Fixed.last())
    {
      term.fix(Fixed.next());
    }
  term.first();
  Coefficient Min,Max;
  Min=Max=term.get_constant();  
  while(!term.last())
    {
      Product p = term.next();
      if (solution.member(p.literal))
	{
	  Min += p.coefficient;
	  Max += p.coefficient;
	}
      else if (solution.member(p.literal))
	{
	  // nothing to do
	}
      else 
	{
	  Max += p.coefficient;
	}
    }
  cout << "[" << Min << ".." << Max << "]";
  return 0;
}

////////////////////////////////////////////////////////////
// the commands

void com_help();

void com_history()
{  
  int no = do_atoi(arguments[0]);
  if (no <= 0)
    no = 10;
  if (no >= 100)
    no = 100;
  command_no--;
  int cno = command_no-1;
  int i = cno%100;  
  int c = 0;

  while (c < no)
    {
      if (history[i].length() > 0)
	{
	  cout << i + (cno/100)*100 << " : " << history[i] << "\n";
	  c++;
	}
      if (i == 0)
	{
	  cno -= 100;
	  if (cno < 0)
	    return;
	  i = 99;
	}
      else
	i--;
    }  
}

void com_clear()
{
  pbcs.clear();
  objf.clear();
  atoms.clear();
}

void com_read()
{
  com_clear();
  int i;
  char filename[80];
  for(i=0;i < arguments[0].length();i++)
    filename[i] = arguments[0][i];
  filename[i] = 0;
  ifstream problem(filename);
  
  cout << "reading : " << filename << "\n";
  LPBIneqSet X;
  if (opbdp_read(atoms,X,objf,problem,max))
    { 
      cout << "Problem reading data\n";
      com_clear();
    }
  X.first();
  while(!X.last())
    {
      LPBIneq& lpbi = X.next();
      pbcs.add(lpbi);
    }
}

void com_readnl(int lin)
{
  com_clear();
  int i;
  char filename[80];
  for(i=0;i < arguments[0].length();i++)
    filename[i] = arguments[0][i];
  filename[i] = 0;
  ifstream problem(filename);
  
  cout << "nonlinear reading : " << filename << "\n";
  LPBIneqSet X;
  if (readnonlinear(problem,objf,X,max,atoms,lin))
    { 
      cout << "Problem reading data\n";
      com_clear();
    }
  X.first();
  while(!X.last())
    {
      LPBIneq& lpbi = X.next();
      pbcs.add(lpbi);
    }
}

void com_readnlf()
{
  com_readnl(0);
}

void com_readnlp()
{
  com_readnl(1);
}

void com_write()
{
  int i;
  char filename[80];
  for(i=0;i < arguments[1].length();i++)
    filename[i] = arguments[1][i];
  filename[i] = 0;
  ofstream lp(filename);
  int solver;
  if (arguments[0] == "cplex")
    solver = CPLEX;
  else if (arguments[0] == "lpsolve")
    solver = LPSOLVE;
  else
    solver = OPBDP;
  start_timing();
// lp_write(lp,pbcs.S,objf,max,0,solver);
  lp_write(lp,pbcs.S,objf,max,&atoms,solver);
}

void com_show()
{
  if (arguments[0] == "ineq")
    show_lpbi_iid(pbcs,do_atoi(arguments[1]));
  else if (arguments[0] == "ineqs")
    show_all_lpbis(pbcs); 
  else if (arguments[0] == "objf")
    {
      objf.Print(cout,&atoms); 
      cout << "\n";
    }
  else if (arguments[0] == "all")
    {
      objf.Print(cout,&atoms); 
      cout << "\n";
      show_all_lpbis(pbcs);
    }
  else if (arguments[0] == "size")
    {
      cout << "There are " << pbcs.S.card() << " inequalities \n";
    }
  else if (arguments[0] == "lastsolution")
    {
      cout << "Last Solution : ";
      cout_literals(pbcs.get_last_solution());
      cout << "\n";
      if (!objf.empty())
	{
	  cout << "Objective function range : ";
	  show_ranges_sol(objf,pbcs.get_last_solution());
	  cout << "\n";
	}
    }
  else    
    {
      cout << "don't know how to show '" << arguments[0] << "' \n";
    }
}

void com_del()
{
  int iid = do_atoi(arguments[0]);
  cout << "deleting inequality " << iid << "\n";
  pbcs.del_lpbi(iid);
}

void com_add()
{
  LPBIneq ineq;
  int eqi = arguments[0].find("=");
  if ((eqi > 0) && ((arguments[0][eqi-1] != '<') && (arguments[0][eqi-1] != '>')))
    {
      SimpleString s = arguments[0];
      SimpleString greater = ">";
      s.ins(greater,eqi);
      ineq.initialize(s,&atoms);
      pbcs.add(ineq);
      s[eqi] = '<';
      ineq.initialize(s,&atoms);	      
      pbcs.add(ineq);
      cout << "Equation changed into two inequalities\n";
    }
  else
    {
      ineq.initialize(arguments[0],&atoms);
      pbcs.add(ineq);
    }
}

void com_objf(int do_max)
{
  if (objf.initialize(arguments[0],&atoms))
    {
      cout << "Problem reading objective function \n";
      objf.clear();
    }
  if (do_max)
    max = 1;
  else
    max = 0;
}

void com_singlevars()
{
  OrdInt sv = pbcs.single_vars();
  if (!sv.empty())
    {
      cout << "Single Variables : ";
      sv.first();
      while(!sv.last())
	{
	  cout_literal(sv.next());
	  cout << " ";
	}
      cout << "\n";
    }
}

void com_fix()
{
  int literal;
  int k=0;
  while(arguments[k] != "")
    {
      if ((literal = get_literal(arguments[k])) != 0)
	{
	  pbcs.fix(literal);
	  objf.fix(literal);
	}
      else
	{
	  cout << "Argument is not a literal\n";
	}
      k++;
    } 
}

void com_fixing()
{
  OrdInt fixed_lits = pbcs.fixing();
  if (!fixed_lits.empty())
    {
      int lit;
      cout << "Fixing : ";
      fixed_lits.first();
      while(!fixed_lits.last())
	{
	  objf.fix(lit = fixed_lits.next());
	  cout_literal(lit);
	}
      if (pbcs.is_unsatisfiable())
	cout << "\n'fixing' detected inconsistency\n";		
      else if (fixed_lits.length() == 1)		
	cout << "\n1 Literal has been fixed\n";
      else
	cout << "\n" << fixed_lits.length() << " Literals have been fixed\n";
    }
  else
    cout << "No Literals need obviously be fixed \n";
}

void com_fixingcheck()
{
  pbcs.enum_stat_reset();
  OrdInt fixed_lits = pbcs.fixingcheck();
  if (!fixed_lits.empty())
    {
      int lit;
      cout << "Fixingcheck : ";
      fixed_lits.first();
      while(!fixed_lits.last())
	{		  
	  objf.fix(lit = fixed_lits.next());
	  pbcs.fix(lit);
	  cout_literal(lit);
	}
      if (pbcs.is_unsatisfiable())
	cout << "\n'fixingcheck' detected inconsistency\n";		
      if (fixed_lits.length() == 1)		
	cout << "\n1 Literal has been fixed\n";
      else
	cout << "\n" << fixed_lits.length() << " Literals have been fixed\n";
    }
  else
    cout << "No Literals are fixed by fixingcheck \n";
}

void com_equate()
{
  int lit1=0;
  int lit2=0;

  if (((lit1 = get_literal(arguments[0])) != 0) && 
      ((lit2 = get_literal(arguments[1])) != 0))
    {
      start_timing();
      pbcs.equate(lit1,lit2);
      objf.equate(lit1,lit2);
    }	      
  else
    {
      if (lit1 == 0)
	cout << "First ";
      else
	cout << "Second ";
      cout << "argument is not a literal\n";
    }	  
}

void com_eqcheck()
{
  pbcs.eq_check();
  Equations& eq = pbcs.Eq;
  eq.first();
  if (!eq.last())
    {
      int eqs=0;
      int l,r;
      cout << "Implications_equations : ";
      while(!eq.last())
	{	
	  l = eq.next();
	  r = eq(l);
	  objf.equate(l,r);
	  cout_literal(l); cout << "= "; cout_literal(r);
	  pbcs.equate(l,r);
	  cout.flush();
	  eqs++;
	}
      if (eqs == 1)		
	cout << "\n1 Equation has been applied\n";
      else
	cout << "\n" << eqs << " Equations have been applied\n";
    }
  else
    cout << "No Equations are detected eqcheck\n";
}

void com_gencover()
{
  OrdInt cover;
  pbcs.enum_stat_reset();
  int res = pbcs.GenerateCover(cover);
  if (res == 1)
    cout << "Solution found : ";
  else
    cout << "Cover is : ";
  cout_literals(cover);
  cout << "\n";
}

void com_genmincover()
{
  OrdInt cover;
  pbcs.enum_stat_reset();
  int res = pbcs.GenerateMinimizedCover(cover);
  if (res == 1)
    cout << "Solution found : ";
  else
    cout << "Cover is : ";
  cout_literals(cover);
  cout << "\n";
}

void com_addcover()
{
  LPBIneq lpbi;
  switch (pbcs.GenerateMinimizedCoverIneq(lpbi)) {
  case -1 : 
    cout << "Adding CoverIneq : "; 
    lpbi.Print(cout,&atoms);
    cout << "\n";
    pbcs.add_delete_implied(lpbi);
    break;
  case 1 :
    cout << "\nSolution found : ";
    cout_literals(pbcs.get_last_solution());
    cout << "\n";
    if (!objf.empty())
      {
	cout << "Objective function range : ";
	show_ranges_sol(objf,pbcs.get_last_solution());
      }
    cout << "\n";
    break;
  }
}

void com_addcovers()
{
  int no_covers;
  if (arguments[0].length() == 0)
    no_covers = 20;
  else
    no_covers = do_atoi(arguments[0]);
  int i=0;
  int sol_found = 0;
  int clen=0;
  while ((i < no_covers) && !sol_found)
    {
      switch (pbcs.addcover(clen)) {
      case 0 : 
	cout << "[" << clen << "]";
	cout.flush();
	break;
      case 1 :
	if (i)
	  cout << "\n";
	cout << "Solution found : ";
	cout_literals(pbcs.get_last_solution());
	cout << "\n";
	if (!objf.empty())
	  {
	    cout << "Objective function range : ";
	    show_ranges_sol(objf,pbcs.get_last_solution());
	    cout << "\n";
	  }
	sol_found = 1;
	break;
      case -1 :
	break;  // handled elsewhere
      }
      i++;
      if (pbcs.is_unsatisfiable())
	break;
    }
  if (sol_found)
    {
      if (i == 1)
	cout << "Solution found after generating 1 cover\n";
      else
	cout << "Solution found after generating " << i << " covers\n";
    }
  else
    {
      if (i == 1)
	cout << "Generated 1 cover\n";
      else
	cout << "Generated " << i << " covers\n";
    }
  if (pbcs.is_unsatisfiable())
    cout << "Inconsistency detected!\n";
  cout << "There are " << pbcs.S.card() << " inequalities \n";
}

void com_lowerbound()
{
  Products p;
  p.initialize(arguments[0],&atoms);
  pbcs.Enum.init_enumeration();
  pbcs.enum_stat_reset();
  Coefficient lb = pbcs.Enum.LowerBound(p);
  cout << "lower bound is " << lb << "\n";
}

void com_coeffredM()
{
  pbcs.Enum.init_enumeration();
  pbcs.enum_stat_reset();
  int changed = pbcs.coefficient_reduction_bigM();
  if (changed > 1)	    
    cout << "coeffredM changed " << changed << " coefficients\n";
  else if (changed == 1)
    cout << "coeffredM changed " << changed << " coefficient\n";
  else
    cout << "coeffredM changed no coefficients\n";
}

void com_coeffred()
{
  pbcs.Enum.init_enumeration();
  pbcs.enum_stat_reset();
  int changed = pbcs.coefficient_reduction();
  if (changed > 1)	    
    cout << "coeffred changed " << changed << " coefficients\n";
  else if (changed == 1)
    cout << "coeffred changed " << changed << " coefficient\n";
  else
    cout << "coeffred changed no coefficients\n";
}

void com_fixinglit()
{
  int unsat = 0;
  int lit = get_literal(arguments[0]);
  if (lit != 0)
    {
      OrdInt fixinglits = pbcs.fixed_by(lit,unsat);
      if (unsat)
	{
	  cout << "Fixing Literal ";
	  cout_literal(lit);
	  cout << "yields inconsistent problem\n";
	}
      else 
	{
	  cout << "Fixing Literal ";
	  cout_literal(lit);
	  if (!fixinglits.empty())
	    {
	      cout << "forces the following literals to be fixed : \n";
	      fixinglits.first();
	      while(!fixinglits.last())
		{
		  cout_literal(fixinglits.next());
		}
	      cout << "\n";
	    }
	  else
	    {
	      cout << "forces no literals to be fixed : \n";
	    }
	}
    }
}
 
void com_delimp()
{
  int no_del = pbcs.delete_implied();
  if (no_del == 1)
    cout << "Deleted 1 implied inequality\n";
  else if (no_del > 1)
    cout << "Deleted "<< no_del << " implied inequalities\n";
}

void com_solve()
{
  if (pbcs.init_enumeration() == -1)
    {
      cout << "Constraint set is inconsistent\n";
      return;
    }
  if (pbcs.S.card() == 0)
    {
      cout << "empty constraint set\n";
      return;
    }
  pbcs.enum_stat_reset();
  int consistent = 0;
  int node_limit = 0;
  if (arguments[0] != "")
    {
      node_limit = do_atoi(arguments[0]);
      consistent = pbcs.Enum.solve_limit(node_limit);
    }
  else
    {
      consistent = pbcs.Enum.solve();
    }
  if (consistent == pbcs.Enum.OPEN)
    cout << "No solution found in " << node_limit << " nodes\n";
  else if (consistent == pbcs.Enum.SAT)
    cout << "Constraint set is satisfiable\n";
  else
    cout << "Constraint set is inconsistent\n";
  cout << "Explored Nodes : " << pbcs.Enum.stat_nodes();
  cout << "     Fixed Literals : ";
  cout << pbcs.Enum.stat_fixedlits() << "\n"; 
}

void com_coversolve()
{
  if (pbcs.S.card() == 0)
    {
      cout << "empty constraint set\n";
      return;
    }
  pbcs.enum_stat_reset();
  pbcs.reset_cover_statistics();
  if (pbcs.coversolve() == -1)
    cout << "Constraint set is inconsistent\n";
  else
    {
      cout << "Constraint is consistent\n";
      if (!objf.empty())
	{
	  OrdInt sol = pbcs.get_last_solution();
	  cout << "Solution is : ";
	  cout_literals(sol);
	  cout << "\n";

	  Coefficient last_optimum,remaining;
	  objf.get_value(sol,last_optimum,remaining);
	  cout << "Objective function range with this solution : [";
	  cout << last_optimum;
	  cout << "..";
	  cout << last_optimum+remaining;
	  cout << "]\n";
	}
    }
  cout << "Generated Covers : " << pbcs.get_no_covers();
  cout << "     Explored Nodes : " << pbcs.Enum.stat_nodes();
  cout << "     Fixed Literals : ";
  cout << pbcs.Enum.stat_fixedlits() << "\n"; 
}

void com_opt()
{
  if (objf.empty())
    {
      cout << "no objective function \n";
      return;
    }
  if (pbcs.init_enumeration() == -1)
    {
      cout << "Constraint set is inconsistent\n";
      return;
    }
  pbcs.enum_stat_reset();
  int consistent;
  cout << "                       Value        Nodes /    Fixed /       Time \n";
  cout.flush();
  start_timing();
  Coefficient value;
  if (pbcs.is_unsatisfiable())
    {
      value = 0;
      consistent = 0;
    }
  else if (pbcs.is_empty())
    {
      value = max ? objf.get_upperbound() : objf.get_lowerbound();
      consistent = 1;
    }
  else
    {
      //////////////////
      value = pbcs.optimize(objf,consistent,ones,max);
      //////////////////
    }
  if (consistent)
    {
      cout << "Solution is : ";
      cout_literals(ones);
      cout << "\n";
      if (max)
	cout << "Global Maximum:   " << value << "\n";
      else
	cout << "Global Minimum:   " << value << "\n";
    }
  else
    {
      cout << "Constraint set is inconsistent \n";
    }		
  cout << "Explored Nodes : " << pbcs.Enum.stat_nodes();
  cout << "   Max. Depth: " << pbcs.Enum.stat_maxdepth();
  cout << "   Fixed Literals : ";
  cout << pbcs.Enum.stat_fixedlits() << "\n"; 
}

void com_coveropt()
{
  if (objf.empty())
    {
      cout << "no objective function \n";
      return;
    }
  if (pbcs.init_enumeration() == -1)
    {
      cout << "Constraint set is inconsistent\n";
      return;
    }
  pbcs.enum_stat_reset();
  pbcs.reset_cover_statistics();
  int consistent;
  cout << "                       Value       Covers /    Nodes /    Fixed /       Time \n";
  cout.flush();
  start_timing();
  // copy constraint set
  LPBIneqSet Copy;
  OrdInt alliids = pbcs.get_lpbi_iids();
  LPBIneq clpbi;
  alliids.first();
  while (!alliids.last())
    {
      clpbi = pbcs.get_lpbi(alliids.next());
      Copy.add(clpbi);
    }
  OrdInt Fixed = pbcs.get_fixed();
  Equations Eq = pbcs.get_equations();
  // JUST DO IT
  Coefficient value = pbcs.coveroptimize(objf,consistent,ones,max);      
  // DID IT
  // restore constraint set
  pbcs.clear();
  Copy.first();
  while(!Copy.last())
    pbcs.add(Copy.next());
  Fixed.first();
  while(!Fixed.last())
    pbcs.fix(Fixed.next());
  Eq.first();
  while(!Eq.last())
    {
      int l = Eq.next();
      pbcs.equate(l,Eq(l));
    }
  if (consistent)
    {
      cout << "Solution is : ";
      cout_literals(ones);
      cout << "\n";
      if (max)
	cout << "Global Maximum:   " << value << "\n";
      else
	cout << "Global Minimum:   " << value << "\n";
    }
  else
    {
      cout << "Constraint set is inconsistent \n";
    }		
  cout << "Generated Covers : " << pbcs.get_no_covers();
  cout << "     Explored Nodes : " << pbcs.Enum.stat_nodes();
  cout << "     Max. Depth: " << pbcs.Enum.stat_maxdepth();
  cout << "     Fixed Literals : ";
  cout << pbcs.Enum.stat_fixedlits() << "\n"; 
}

void com_set()
{
  if (arguments[0] == "")
    {
      // show settings
      cout << "Settings: \n";
      cout << "  Heuristics     : " << pbcs.get_enum_heuristic() << "\n";
      cout << "  Verbosity      : " << pbcs.get_verbosity() << "\n";
      cout << "  EnumVerbosity  : " << pbcs.get_enum_verbosity() << "\n";
      cout << "  MaxCutNodes    : " << pbcs.get_max_cut_nodes() << "\n";
      cout << "  DelLitHeu      : " << pbcs.get_dellitheu() << "\n";
    }
  else if (arguments[0] == "heuristic")
    pbcs.set_enum_heuristic(do_atoi(arguments[1]));
  else if (arguments[0] == "enumverbosity")
    pbcs.set_enum_verbosity(do_atoi(arguments[1]));
  else if (arguments[0] == "verbosity")
    pbcs.set_verbosity(do_atoi(arguments[1]));
  else if (arguments[0] == "maxcutnodes")
    pbcs.set_max_cut_nodes(do_atoi(arguments[1]));
  else if (arguments[0] == "dellitheu")
    pbcs.set_dellitheu(do_atoi(arguments[1]));
  else if ((arguments[0] == "lowerbound") || (arguments[0] == "upperbound"))
    {
      if (!objf.empty())
	{
	  int bound = do_atoi(arguments[1]);
	  
	  if (arguments[0] == "upperbound")
	    {
	      Products objfcopy(objf);
	      objfcopy.mult(-1);
	      LPBIneq lpbi(objfcopy,-bound);
	      pbcs.add(lpbi);
	    }
	  else
	    {
	      LPBIneq lpbi(objf,bound);
	      pbcs.add(lpbi);
	    }
	}
      else
	cout << "No objective function\n";
    }	  
  else
    cout << "don't know how to set " << arguments[0] << "\n";
}

void com_check()
{
  if (pbcs.datastructure_corrupted())
    cout << "Problems with the internal data_structure\n";
  else
    cout << "Internal data_structure OK!\n";
}

void header(char* );

int pbcsia_main(int , char **argv)
{
#ifdef QUANTIFIED
  quantify_stop_recording_data();
#endif
  pbcs.set_enum_verbosity(3);
  pbcs.set_enum_heuristic(1);
  pbcs.set_max_cut_nodes(500);
  pbcs.set_atoms(&atoms);

  header(argv[0]);
  cout << "interaction mode \n";
  while ((command_loop() != -1) && cin)
    {
      start_timing();
      if (command == "") command_no--;
      else if (command == "help") com_help();
      else if (command == "h") com_history();
      else if (command == "history") com_history();
      else if (command == "clear") com_clear();
      else if (command == "read") com_read();
      else if (command == "readnlf") com_readnlf();
      else if (command == "readnlp") com_readnlp();
      else if (command == "write") com_write();
      else if (command == "show") com_show();
      else if (command == "del") com_del();
      else if (command == "add") com_add();
      else if (command == "min") com_objf(0);
      else if (command == "max") com_objf(1);
      else if (command == "singlevars") com_singlevars();
      else if (command == "fix") com_fix();
      else if (command == "fixing") com_fixing();
      else if (command == "fixingcheck") com_fixingcheck();
      else if (command == "equate") com_equate();
      else if (command == "eqcheck") com_eqcheck();
      else if (command == "gencover") com_gencover();
      else if (command == "genmincover") com_genmincover();
      else if (command == "addcover") com_addcover();
      else if (command == "addcovers") com_addcovers();
      else if (command == "lowerbound") com_lowerbound();
      else if (command == "coeffredM") com_coeffredM();
      else if (command == "coeffred") com_coeffred();
      else if (command == "fixinglit") com_fixinglit();
      else if (command == "delimp") com_delimp();
      else if (command == "solve") com_solve();
      else if (command == "coversolve") com_coversolve();
      else if (command == "opt") com_opt();
      else if (command == "coveropt") com_coveropt();
      else if (command == "set") com_set();
      else if (command == "check") com_check();
      else if ((command == "quit") || (command == "end") || (command == "bye") || (command == "exit"))
	break; // leave command loop
      else
	{
	  cout << "command '" << command << "' not understood\n";
	  command_no--;
	}
      show_timing();
    }
  cout << "Bye ... \n";
  return 0;
}



