#include "readnonlinear.h"
#include "Linearize.h"
#include "opbdp_read.h"

static int line_counter;
static int equation_counter;

int read_nlpbc(SimpleString& s,LPBIneqSet& S,Atoms& atoms,int lin)
{
  SimpleString ls,rs;
  PolPBTerm l;
  int oppos; 
  int eq=0;

  // handle a syntax error produced by CPLEX
  // something like + -12 A I don't like to parse
  // i.e., + - something without left term
  // (- + something is an error anyway!)
  // I handle it by always deleting a leading plus :-)
  int i=0;
  while (i < s.length())
    {
      if ((s[i] != ' ') && (s[i] != '\t'))
	{
	  if (s[i] == '+')
	    s.del(0,i);
	  else
	    break;
	}
      i++;
    }
  // search relop and split strings
  oppos = s.find("<");
  if (oppos < 0)
    oppos = s.find(">");
  if (oppos < 0)
    oppos = s.find("=");
  if (oppos < 0)
    {
      cerr << "read_nlpbc : no relation symbol in \n";
      cerr << s << "\n";
      return 1;
    }
  if (oppos == 0)
    {
      cerr << "read_nlpbc : no left term\n";
      return 1;
    }
  if (s.length() <= oppos+1)
    {
      int err = 0;
      if (s.length() <= oppos)
	err = 1;
      else if ((s[oppos+1] == '=') || (s[oppos+1] == '<') || (s[oppos+1] == '>'))
	err = 1;
      if (err)
	{
	  cerr << "read_nlpbc : no right term\n";
	  return 1;
	}
    }
  // merge strings to l >= 0
  switch (s[oppos]) {
  case '=' :
    ls = s(0,oppos-1);
    rs = s(oppos+1,s.length()-1);
    eq = 1;
    equation_counter++;
    break;
  case '<':    
    if (s[oppos+1] == '=')
      {
	rs = s(0,oppos-1);
	ls = s(oppos+2,s.length()-1);
      }
    else
      {
	rs = s(0,oppos-1);
	rs += " + 1";
	ls = s(oppos+2,s.length()-1);
      }
    break;
  case '>':
    if (s[oppos+1] == '=')
      {
	ls = s(0,oppos-1);
	rs = s(oppos+2,s.length()-1);
      }
    else
      {
	ls = s(0,oppos-1);
	ls += " + 1";
	rs = s(oppos+2,s.length()-1);
      }
    break;
  default:
    cerr << "read_nlpbc : cannot happen\n";
    return 1;
  }
  ls += " - (";
  ls += rs;
  ls += ")";
  int ret;
  if ((ret = l.initialize(ls,&atoms)))
    return ret;
  PolPBIneq pbi(l,0);

  if (lin == 0) // fortet
    {
      linearize_fortet(pbi, S, &atoms); 
      if (eq)
	{
	  l.mul_constant(-1);
	  PolPBIneq pbi(l,0);
	  linearize_fortet(pbi, S, &atoms); 	  
	}
    }
  else if (lin == 1) // mine
    {
      linearize(pbi, S); 
      if (eq)
	{
	  l.mul_constant(-1);
	  PolPBIneq pbi(l,0);
	  linearize(pbi, S);
	}
    }  
  else // show ineq
    {
      pbi.Print(cout,&atoms);cout << "\n";
    }
  return 0;
}

int next_string(istream& in,SimpleString& s, int delname = 0)
{
  // skip comment lines
  do { 
    s.Read(in,"\n");
    delete_punctation(s);
    line_counter++;
    if (is_empty_line(s,line_counter) == -1)
      return 1; // a syntax error
    delete_punctation(s);
    if (delname)
      delete_constraint_name(s);
    if (!in)
      break;
  } while (s.length() == 0);
  return 0;
}

int readnonlinear(istream& in, Products& objf, LPBIneqSet& S, int& max, Atoms& atoms, int lin, int verb)
{
  line_counter = 0;
  equation_counter = 0;
  SimpleString s;
  max = 1;
  PolPBTerm nlobjf;
  int ret;

  if ((ret = next_string(in,s)))
    return ret;
  if (s.length() >= 4)
    {
      SimpleString h=s(0,3);
      if (h == "min:")
	{
	  max = 0;
	  s = s(4,s.length()-1);	  
	}
      else if (h == "max:")
	{
	  s = s(4,s.length()-1);	  
	}
    }
  if ((s.length() == 0) || (nlobjf.initialize(s,&atoms)))
    {
      cerr << "Problem reading objective function\n";
      return 1;
    }
  linearize_fortet(nlobjf, objf, S, &atoms);
  while(in)
    {
      if ((ret = next_string(in,s,1)))
	return ret;
      if (s.length() == 0)
	break;
      if ((ret = read_nlpbc(s,S,atoms,lin)))
	return ret;
    }
  // delete simply implied inequalities (quadratic cost)
  // S.delete_implied(); // one can also do it by hand...
  if (verb)
    {
      if (equation_counter == 1)
	cout << "1 equation changed into 2 inequalities \n";
      else if (equation_counter > 1)
	cout << equation_counter << " equations changed into " << equation_counter*2 << " inequalities \n"; 
      cout << "problem consists of " << S.card() << " inequalities and ";
      cout << atoms.size() << " variables \n";
    }
  return 0;
}
